
#include "LzmaDecode.h"
#include "stdio.h"
#include "malloc.h"
#include "lzma.h"

#define printf(...)

char properties[16*1024];

int decompress_lzma_7z( unsigned char *inBuffer, 
                        unsigned       inAvail, 
                        unsigned char *g_OutBuffer, 
                        unsigned       outAvail,
                        unsigned      *outProcessed )
{
    CLzmaDecoderState   state;  // it's about 24-80 bytes structure, if int is 32-bit
    int                 res;
    SizeT               inProcessed;
    SizeT               compressedSize;
    int                 propSize;

    // Read properties bytes.
    if ( LzmaDecodeProperties( &state.Properties, inBuffer, LZMA_PROPERTIES_SIZE) != LZMA_RESULT_OK )
    {
        printf( "Incorrect stream properties\n" );
        return 1;
    }

    propSize = LzmaGetNumProbs(&state.Properties) * sizeof(CProb);
    printf( "Properties size = %d\n", propSize );

    // We'd prefer to use our statically allocated property area, which should
    // always be less than 16k.  But just in case it's over 16k, handle it.
    if ( propSize > 16*1024 )
        return 10;
        // Allocate area for private data - around 16k.
//        state.Probs = (CProb *) malloc( propSize );
    else
        state.Probs = (CProb *) properties;

    // Skip properties
    inBuffer       = inBuffer + LZMA_PROPERTIES_SIZE;
    compressedSize = inAvail  - LZMA_PROPERTIES_SIZE;

    res = LzmaDecode( &state,
                       inBuffer, 
                       compressedSize, 
                      &inProcessed,
                       g_OutBuffer, 
                       outAvail, 
                       outProcessed );
    if ( res != 0 )
    {
        printf( "\nDecoding error = %d\n", res );
        if ( inProcessed == compressedSize )
        {
            res = 0;
        }
    }
    printf( "inAvail = %u, inProcessed = %u, outProcessed = %u\n", 
                       inAvail,          inProcessed,       *outProcessed );

//    if ( propSize > 16*1024 )
//        free( (char *) state.Probs );

    return res;
}
