/***************************************************************************
*
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: asreq.c
*    Creation Date: 27 Oct 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:53:56 $
*
****************************************************************************
*    Description:
*
*      Implementation of packetCable Kereberos AS request and AS reply
*
****************************************************************************/


#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <time.h>
#include <net.h>
#ifdef PSOS
#include <mibAccessSec.h>
#endif

#include "seckrb.h"
#include "krb5_err.h"
#include "pkinit.h"
#include "pkinitReply.h"
#include "krb5client.h"
#include "krbbuf.h"
#include "des_int.h"


const unsigned char PROV_SERVICE_COMP1[] = "ProvSrvr"; //"mtaprovsrvr";
const unsigned char CMS_SERVICE_COMP1[] =  "cms";
const unsigned char MTA_SERVICE_COMP1[] =  "PSElement"; //"mta"; 

/* OID needed for PKINIT reply */
const int        SIGNED_DATA_OID[7] = {1,2,840,113549,1,7,2};
const int        SIGNED_DATA_OID_LEN = 7;
const int        SHA1_OID[6] = {1,3,14,3,2,26};
const int        SHA1_OID_LEN = 6;
const int        ECONTENT_OID[8] = {1,3,6,1,5,2,3,2};
const int        ECONTENT_OID_LEN = 8;
const int        RSA_ENCRYPITON_OID[7] = {1,2,840,113549,1,1,1};
const int        RSA_ENCRYPITON_OID_LEN = 7;



/* flags to allow files to be opened once */
#ifdef 0//DEMO
int fopened_once1 = 0;
int fopened_once2 = 0;
int fopened_once3 = 0;
int fopened_once4 = 0;
int fopened_once5 = 0;
int fopened_once6 = 0;
#endif



#if 0
static unsigned char as_rep_txt[] = {
#include "as_rep.txt"
};
#endif


/* defined in asn.1/krb5_encode.c */
krb5_error_code encode_krb5_as_req(const krb5_kdc_req *rep, krb5_data **code);
krb5_error_code decode_krb5_error(const krb5_data *code, krb5_error **rep);
krb5_error_code decode_krb5_as_rep(const krb5_data * code, krb5_kdc_rep ** rep);
krb5_error_code decode_krb5_padata_sequence(const krb5_data * code, krb5_pa_data ***rep);
/*
 * key-producing procedure for use by krb5_get_in_tkt_with_password.
 */
krb5_error_code pwd_keyproc(const krb5_enctype enctype, krb5_data *data,
      krb5_const_pointer keyseed, krb5_keyblock **key);

krb5_error_code krb5_c_string_to_key(krb5_enctype enctype,
      const krb5_data *string, const krb5_data *salt,
      krb5_keyblock *key);

typedef krb5_error_code (*git_key_proc) (const krb5_enctype enctype,
      krb5_data *data, krb5_const_pointer keyseed, krb5_keyblock **key);

typedef krb5_error_code (*git_decrypt_proc) (const krb5_keyblock *key,
      krb5_const_pointer keyseed, krb5_kdc_rep *as_rep);

krb5_error_code krb5_principal2salt( register krb5_const_principal pr,
      krb5_data *ret);

krb5_error_code krb5_kdc_rep_decrypt_proc(const krb5_keyblock *key,
      krb5_const_pointer decryptarg, krb5_kdc_rep *as_rep);

krb5_boolean krb5_principal_compare( krb5_const_principal princ1,
      krb5_const_principal princ2);

krb5_error_code krb5_copy_principal(krb5_const_principal inprinc,
      krb5_principal *outprinc);

krb5_error_code krb5_copy_keyblock_contents( const krb5_keyblock *from,
      krb5_keyblock *to);

krb5_error_code encode_krb5_ticket(const krb5_ticket * rep, krb5_data ** code);

krb5_error_code krb5_obtain_padata(krb5_pa_data **preauth_to_use,
      git_key_proc key_proc,
      krb5_const_pointer key_seed,
      krb5_creds *creds,
      krb5_kdc_req *request);


/*
*****************************************************************************
** FUNCTION:    make_preauth_list
**
** PURPOSE:     makes padata list for preauthentication
**
** PARAMETERS:  ptypes - list of preauthentication types
**              nptypes - number of types (entries)
**              ret_list - out: list will be constructed and returned
**
** RETURNS:     success or error code
**
** NOTE:
**        The only type supported is ENC-TIMESTAMP, which will be used later
**        on for I02
**
*****************************************************************************
*/
static krb5_error_code make_preauth_list(krb5_preauthtype *ptypes,
      int			     nptypes,
      krb5_pa_data    ***ret_list)
{
   krb5_preauthtype *ptypep;
   krb5_pa_data **preauthp;
   krb5_pa_data **preauth_to_use;
   int				i;

   if (nptypes < 0)
   {
      for (nptypes=0, ptypep = ptypes; *ptypep; ptypep++, nptypes++)
         ;
   }

   /* allocate space for a NULL to terminate the list */

   if ((preauthp =
            (krb5_pa_data **) krbAllocBuf((nptypes+1)*sizeof(krb5_pa_data *))) == NULL)
   {
      return(ENOMEM);
   }

   for (i=0; i<nptypes; i++)
   {
      if ((preauthp[i] =
               (krb5_pa_data *) krbAllocBuf(sizeof(krb5_pa_data))) == NULL)
      {
         for (; i>=0; i++)
            krbFreeBuf((char *)preauthp[i]);
         krbFreeBuf((char *)preauthp);
         return (ENOMEM);
      }
      preauthp[i]->magic = KV5M_PA_DATA;
      preauthp[i]->pa_type = ptypes[i];
      preauthp[i]->length = 0;
      preauthp[i]->contents = 0;
   }

   /* fill in the terminating NULL */

   preauthp[nptypes] = NULL;

   *ret_list = preauthp;
   return 0;
}



/*
*****************************************************************************
** FUNCTION:      pwd_keyproc
**
** PURPOSE:       password key derivation function
**
** PARAMETERS:    type - encryption type 3des-cbc-md5 for packetcable
**                salt - salt for password
**                keyseed - seed for password - null
**                key - out: key derived and returned
**
** RETURNS:       success or error code
**
** NOTE:
**        the password here is hardcoded "broadcom", this is a temporary
**        function used for testing purposes only.
**
*****************************************************************************
*/
krb5_error_code pwd_keyproc(const krb5_enctype type,
      krb5_data *salt,
      krb5_const_pointer keyseed,
      krb5_keyblock **key)
{
   krb5_error_code retval;
   krb5_data *password;

   /* password is "broadcom" */
   password = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
   if (password == NULL)
   {
      return ENOMEM;
   }
   password->data = "broadcom";
   password->length = strlen("broadcom");

   if (!password->length)
   {
      printf("No password provided to decrypt.\n");
   }

   if (!(*key = (krb5_keyblock *)krbAllocBuf(sizeof(**key))))
   {
      return ENOMEM;
   }

   if ((retval = krb5_c_string_to_key(type, password, salt, *key)))
      printf("string_to_key failed.\n");

   krbFreeBuf(password->data);
   krbFreeBuf((char *)password);
   return(retval);
}


/*
*****************************************************************************
** FUNCTION:     dh_keyproc
**
** PURPOSE:      Key derivation from a diffie-hellman secret
**
** PARAMETERS:   type - type of encryption - 3des-cbc-md5
**               salt - salt for string to key
**               keyseed - seed on the key, DH shared secret in this case
**               key - out: the key produced (24 bytes for 3des)
**
** RETURNS:      success or error code
**
** NOTE:
**
*****************************************************************************
*/
krb5_error_code dh_keyproc(const krb5_enctype type,
      krb5_data *salt,
      krb5_const_pointer keyseed,
      krb5_keyblock **key)
{
   krb5_error_code retval;
   krb5_data *tkey;
   krb5_data *tmpseed;
   mit_des3_cblock pkey;

   tmpseed = (krb5_data *)keyseed;

   tkey = (krb5_data *)krbAllocBuf(sizeof(krb5_data));
   if (tkey == NULL)
   {
      return ENOMEM;
   }
   tkey->data = (unsigned char *)krbAllocBuf(tmpseed->length);
   if ( tkey->data == NULL )
   {
	   return ENOMEM;
   }
   memcpy(tkey->data, tmpseed->data, tmpseed->length);
   tkey->length = tmpseed->length;

   if (!tkey->length)
   {
      printf("No dhkey provided to decrypt.\n");
   }

   if (!(*key = (krb5_keyblock *)krbAllocBuf(sizeof(**key))))
   {
      return ENOMEM;
   }

   /* copy the first 24 bytes into des3 key and fix the parity */
   memcpy(pkey, tmpseed->data, 24);

   /* Force odd parity, according to pkinit 13 (3.1.1) */
   mit_des3_fixup_key_parity(pkey);
   retval = 0;

   (*key)->contents = (krb5_octet *)krbAllocBuf(24);
   memcpy((*key)->contents, pkey, 24);

   (*key)->enctype = ENCTYPE_DES3_CBC_MD5;
   (*key)->length = 24;

   /*
   if ((retval = krb5_c_string_to_key(type, tkey, salt, *key)))
      printf("string_to_key failed.\n");
   */

   krbFreeBuf(tkey->data);
   krbFreeBuf(tkey);
   return(retval);
}





/*
*****************************************************************************
** FUNCTION:    decrypt_as_reply
**
** PURPOSE:     decrypt the enc-part of AS_REP message
**
** PARAMETERS:  request - request which was sent
**              as_reply - corresponding as_reply that was received
**              key_proc - encryption function to be used
**              keyseed - seed for the key, ignore
**              key - the key itself to use to decrypt
**              decrypt_proc - decryption function
**              decryptarg - decrypt arg, ignore
**
** RETURNS:     success or error code
**
** NOTE:
**
**
*****************************************************************************
*/
static krb5_error_code decrypt_as_reply(
      krb5_kdc_req        *request,
      krb5_kdc_rep        *as_reply,
      git_key_proc        key_proc,
      krb5_const_pointer  keyseed,
      krb5_keyblock       *key,
      git_decrypt_proc    decrypt_proc,
      krb5_const_pointer  decryptarg)
{
   krb5_error_code		retval;
   krb5_keyblock *		decrypt_key = 0;
   krb5_data 			   salt;

   if (as_reply->enc_part2)
   {
      printf("empyt enc_part2.\n");
      return 0;
   }

   if (key)
      decrypt_key = key;
   else
   {
      /*
      if ((retval = krb5_principal2salt(request->client, &salt)))
      {
         printf("principal to salt falied.\n");
         return(retval);
      }
      */

      /* pass the DH shared seccret as keyseed */
      retval = (*key_proc)(as_reply->enc_part.enctype,
            NULL, keyseed, &decrypt_key);
      /* krb5_xfree(salt.data); */
      if (retval)
         goto cleanup;
   }

   if ((retval = (*decrypt_proc)(decrypt_key, decryptarg, as_reply)))
   {
      printf("decrypt failed erro code %d\n", retval);
      goto cleanup;
   }

cleanup:
   if (!key && decrypt_key)
      krb5_free_keyblock(decrypt_key);
   return (retval);
}


/*
*****************************************************************************
** FUNCTION:     setAsreqClientPrincipal
**
** PURPOSE:      set the client principal name and realm in a given AS_REQ
**
** PARAMETERS:   request - request structure
**               mtafqdn - FQDN for the MTA
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
int setAsreqClientPrincipal(
      krb5_kdc_req *request,
      char *krbrealm,
      char *mtafqdn)
{
   char *comp1 = MTA_SERVICE_COMP1;

   /* set the realm */
   request->client->magic=KV5M_PRINCIPAL;
   request->client->realm.magic=KV5M_DATA;
   request->client->realm.data = (char *)krbAllocBuf(strlen(krbrealm)+1);
   if (request->client->realm.data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->client->realm.data, krbrealm, strlen(krbrealm));
   request->client->realm.length=strlen(krbrealm);
   request->client->realm.data[request->client->realm.length] = '\0';

   /* set the principal names comp1=mta, comp2=fqdn */
   request->client->data = (krb5_data *)krbAllocBuf(2*sizeof(krb5_data));
   if (request->client->data == NULL)
   {
      return ENOMEM;
   }

   /* set the first component "mta" */
   request->client->data[0].magic=KV5M_DATA;
   request->client->data[0].data = (char *)krbAllocBuf(strlen(comp1)+1);
   if (request->client->data[0].data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->client->data[0].data, comp1, strlen(comp1));
   request->client->data[0].length=strlen(comp1);
   request->client->data[0].data[request->client->data[0].length] = '\0';

   /* set the second component <fqdn> */
   request->client->data[1].magic=KV5M_DATA;
   request->client->data[1].data = (char *)krbAllocBuf(strlen(mtafqdn)+1);
   if (request->client->data[1].data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->client->data[1].data, mtafqdn, strlen(mtafqdn));
   request->client->data[1].length=strlen(mtafqdn);
   request->client->data[1].data[request->client->data[1].length] = '\0';

   request->client->length= 2;
   request->client->type = KRB5_NT_SRV_HST;

   return 0;
}

/*
*****************************************************************************
** FUNCTION:     setAsreqServerPrincipal
**
** PURPOSE:      set the server principal name and realm in a given AS_REQ
**
** PARAMETERS:   request - request structure
**               srv_princ_name - principal name of service(CA, PROV or krbtgt)
**               srv_realm - realm of the service
**               app_flag - flag - IPSec or SNMPv3
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
int setAsreqServerPrincipal(
      krb5_kdc_req *request,
      const char *comp1,
      const char *comp2,
      const char *srv_realm,
      const int app_flag)
{
   /* Set the realm now */
   request->server->magic=KV5M_PRINCIPAL;
   request->server->realm.magic=KV5M_DATA;
   request->server->realm.data = (char *)krbAllocBuf(strlen(srv_realm)+1);
   if (request->server->realm.data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->server->realm.data, srv_realm, strlen(srv_realm));
   request->server->realm.length=strlen(srv_realm);
   request->server->realm.data[request->server->realm.length] = '\0';

   /* Set the name CMS/<FQDN>@realm,krbtgt/<FQDN>@realm,snmpuser/<FQDN>@realm */
   request->server->data = (krb5_data *)krbAllocBuf(2*sizeof(krb5_data));
   if (request->server->data == NULL)
   {
      return ENOMEM;
   }

   /* set the first component */
   request->server->data[0].magic=KV5M_DATA;
   request->server->data[0].data = (char *)krbAllocBuf(strlen(comp1)+1);
   if (request->server->data[0].data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->server->data[0].data, comp1, strlen(comp1));
   request->server->data[0].length=strlen(comp1);
   request->server->data[0].data[request->server->data[0].length] = '\0';

   /* set the second component */
   request->server->data[1].magic=KV5M_DATA;
   request->server->data[1].data = (char *)krbAllocBuf(strlen(comp2)+1);
   if (request->server->data[1].data == NULL)
   {
      return ENOMEM;
   }
   memcpy(request->server->data[1].data, comp2, strlen(comp2));
   request->server->data[1].length=strlen(comp2);
   request->server->data[1].data[request->server->data[1].length] = '\0';

   request->server->length=2;
   request->server->type = KRB5_NT_SRV_HST;

   return 0;
}


/*
*****************************************************************************
** FUNCTION:     compare_oid
**
** PURPOSE:      comapre 2 oids
**
** PARAMETERS:   oids and their length
**
** RETURNS:      success or failure
**
** NOTE:
*****************************************************************************
*/
int compare_oid(int *oid1, int oid1_len, int *oid2, int oid2_len)
{
   int i;

   if ( oid1_len != oid2_len )
      return 1;
   for (i=0;i<oid1_len;i++)
   {
      if (oid1[i] != oid2[i])
         return 1;
   }
   return 0;
}


/*
*****************************************************************************
** FUNCTION:     verify_pkinit_reply
**
** PURPOSE:      verify PKINIT Reply message (time, nonce etc.)
**
** PARAMETERS:   pkinit_rep
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
static krb5_error_code verify_pkinit_reply(
      pkinit_reply *pkinit_rep)
{
   if (pkinit_rep->enc_key_pack)
   {
      printf("PKINIT Reply optinal field encKeyPack present.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ((pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.signed_attrs) ||
         ( pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.unsigned_attrs))
   {
      printf("PKINIT Reply signerInfo optional field (un)signedAttrs present.\n");
      return KRB5_KDCREP_MODIFIED;
   }

   if ( compare_oid(SIGNED_DATA_OID, SIGNED_DATA_OID_LEN,
            pkinit_rep->dh_signed_data.content_type.oid_array,
            pkinit_rep->dh_signed_data.content_type.length) )
   {
      printf("PKINIT Reply incorrect signed_data oid.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ( compare_oid(SHA1_OID, SHA1_OID_LEN,
            pkinit_rep->dh_signed_data.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array,
            pkinit_rep->dh_signed_data.signed_data.digest_alg_ids.digest_alg_id.alg_oid.length) )
   {
      printf("PKINIT Reply incorrect signed_data digestAlg oid.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ( compare_oid(SHA1_OID, SHA1_OID_LEN,
            pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array,
            pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.length) )
   {
      printf("PKINIT Reply incorrect signed_data signer_info digestAlg oid.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ( compare_oid(RSA_ENCRYPITON_OID, RSA_ENCRYPITON_OID_LEN,
            pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.oid_array,
            pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.signature_alg_id.alg_oid.length) )
   {
      printf("PKINIT Reply incorrect signed_data signer_info signatureAlg oid.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ( compare_oid(ECONTENT_OID, ECONTENT_OID_LEN,
            pkinit_rep->dh_signed_data.signed_data.encap_content_info.econtent_type.oid_array,
            pkinit_rep->dh_signed_data.signed_data.encap_content_info.econtent_type.length) )
   {
      printf("PKINIT Reply incorrect signed_data econtent type oid.\n");
      return KRB5_KDCREP_MODIFIED;
   }



   if ( pkinit_rep->dh_signed_data.signed_data.cms_version != 3 )
   {
      printf("PKINIT Reply:signed_data:cm_version(%d) not 3.\n",
            pkinit_rep->dh_signed_data.signed_data.cms_version);
      return KRB5_KDCREP_MODIFIED;
   }
   if (!( ( pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.cms_version == 1) ||
         ( pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.cms_version == 3) ))
   {
      printf("PKINIT Reply:signed_data:signer_info:cm_version(%d) not 1 or 3.\n",
            pkinit_rep->dh_signed_data.signed_data.signer_infos.signer_info.cms_version);
      return KRB5_KDCREP_MODIFIED;
   }
   if (!( ( pkinit_rep->dh_signed_data.signed_data.certificates.length == 2 ) ||
         ( pkinit_rep->dh_signed_data.signed_data.certificates.length == 3 ) ))
   {
      printf("PKINIT Reply number of certificates (%d) must be 2 or 3\n",
            pkinit_rep->dh_signed_data.signed_data.certificates.length);
      return KRB5_KDCREP_MODIFIED;
   }

   return 0;
}

/*
*****************************************************************************
** FUNCTION:     verify_as_reply
**
** PURPOSE:      verify AS_REP message (time, nonce etc.)
**
** PARAMETERS:   time_now - current time
**               request - AS_REQ message
**               as_reply - AS_REP message
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
static krb5_error_code verify_as_reply(
      krb5_timestamp 		time_now,
      krb5_kdc_req		*request,
      krb5_kdc_rep		*as_reply)
{
   if ( as_reply->msg_type != KRB5_AS_REP )
   {
      printf("AS Reply msg_type(=%d) not KRB5_AS_REP.\n", as_reply->msg_type);
      return KRB5_KDCREP_MODIFIED;
   }
   if( as_reply->client->length != 2 )
   {
      printf("AS Reply cname does not have 2 components (%d).\n",
            as_reply->client->length);
      return KRB5_KDCREP_MODIFIED;
   }
   if( as_reply->client->type != KRB5_NT_SRV_HST )
   {
      printf("AS Reply cname does not have the right type (%d).\n",
            as_reply->client->type);
      return KRB5_KDCREP_MODIFIED;
   }

   if( as_reply->ticket->server->length != 2 )
   {
      printf("AS Reply ticket:sname does not have 2 components (%d).\n",
            as_reply->ticket->server->length);
      return KRB5_KDCREP_MODIFIED;
   }
   if( as_reply->ticket->server->type != KRB5_NT_SRV_HST )
   {
      printf("AS Reply ticket:sname does not have the right type (%d).\n",
            as_reply->ticket->server->type);
      return KRB5_KDCREP_MODIFIED;
   }

   if ( as_reply->ticket->enc_part.enctype != ENCTYPE_DES3_CBC_MD5 )
   {
      printf("AS Reply ticket enctype(=%d) is not DES3_CBC_MD5.\n",
            as_reply->ticket->enc_part.enctype);
      return KRB5_KDCREP_MODIFIED;
   }

   if ( as_reply->enc_part2->times.starttime )
   {
      printf("AS Reply ticket start time Must not be present.\n");
      return KRB5_KDCREP_MODIFIED;
   }

   if ( as_reply->enc_part2->times.renew_till )
   {
      printf("AS Reply ticket renew till Must not be present.\n");
      return KRB5_KDCREP_MODIFIED;
   }

   if ( as_reply->enc_part2->key_exp )
   {
      printf("AS Reply ticket Key Expiration Must not be present.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ( as_reply->enc_part2->caddrs )
   {
      printf("AS Reply caddrs Must not be present.\n");
      return KRB5_KDCREP_MODIFIED;
   }

#if 0
   /* check the contents for sanity: */
   if (!as_reply->enc_part2->times.starttime)
      as_reply->enc_part2->times.starttime =
         as_reply->enc_part2->times.authtime;
#endif

   if (!krb5_principal_compare(as_reply->client, request->client))
   {
      printf("AS Reply and Request client principal names do not match:\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if (!krb5_principal_compare(as_reply->enc_part2->server, request->server))
   {
      printf("AS Reply and Request server principal names do not match:\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if (!krb5_principal_compare(as_reply->ticket->server, request->server))
   {
      printf("AS Ticket server does not match AS request.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if (request->nonce != as_reply->enc_part2->nonce)
   {
      printf("AS Kerberos nonces don't match.\n");
      return KRB5_KDCREP_MODIFIED;
   }

   if ((request->kdc_options & KDC_OPT_POSTDATED) ||
         (request->kdc_options & KDC_OPT_RENEWABLE) ||
         (request->kdc_options & KDC_OPT_RENEWABLE_OK)||
         (request->kdc_options & KDC_OPT_ENC_TKT_IN_SKEY)||
         (request->kdc_options & KDC_OPT_VALIDATE)||
         (request->kdc_options & KDC_OPT_RENEW) )
   {
      printf("PacketCable doe s not support any ticket flags.\n");
      return KRB5_KDCREP_MODIFIED;
   }


   /* NOTE: We'll leave these checks, although no ticket flag is supposed
    * to be supported in packetCable. If we fall into one of these cases,
    * even better. There is something wrong with the ticket.
    * */
   if ((request->kdc_options & KDC_OPT_POSTDATED) &&
            (request->from != 0) &&
            (request->from != as_reply->enc_part2->times.starttime))
   {
      printf("AS KDC_OPT_POSTDATED but from time not the same as starttime.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ((request->till != 0) &&
            (as_reply->enc_part2->times.endtime > request->till))
   {
      printf("AS ticket endtime greater then AS request till time.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ((request->kdc_options & KDC_OPT_RENEWABLE) &&
            (request->rtime != 0) &&
            (as_reply->enc_part2->times.renew_till > request->rtime))
   {
      printf("ticket renew_till greater than AS request rtime.\n");
      return KRB5_KDCREP_MODIFIED;
   }
   if ((request->kdc_options & KDC_OPT_RENEWABLE_OK) &&
            (as_reply->enc_part2->flags & KDC_OPT_RENEWABLE) &&
            (request->till != 0) &&
            (as_reply->enc_part2->times.renew_till > request->till))
   {
      printf("KDC_OPT_RENEWABLE_OK and renew_till greater than AS request till time.\n");
      return KRB5_KDCREP_MODIFIED;
   }

#if 0
   /* Check the ticket starttime and make sure it's not far off */
   if ((request->from == 0) &&
         (labs(as_reply->enc_part2->times.starttime - time_now)
		 > MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS))
   {
      printf("SKEW ERROR IN TKT starttime, time_now = %d, kdc time = %d\n",
            time_now,
            as_reply->enc_part2->times.starttime);
      return (KRB5_KDCREP_SKEW);
   }
#endif

   return 0;
}


/*
*****************************************************************************
** FUNCTION:     store_creds
**
** PURPOSE:      store the credentials received after AS exchange to struct
**               and to NVRAM
**
** PARAMETERS:   as_reply - AS_REP message
**               mtacredentials - credetials (ticket, session key, etc.)
**               app_flag - flag for service type
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
static krb5_error_code
store_creds(krb5_kdc_rep *as_reply, MtaCredentials *rcreds, const int app_flag)
{
   krb5_error_code     retval;
   krb5_data           *packet;
   int                 tflag;


   /* Write the service info into cerdentails */
   rcreds->service.realm_len = as_reply->enc_part2->server->realm.length;
   rcreds->service.realm = krbAllocBuf(rcreds->service.realm_len+1);
   if ( rcreds->service.realm == NULL )
   {
	   return ENOMEM;
   }
   memcpy(rcreds->service.realm, as_reply->enc_part2->server->realm.data,
         rcreds->service.realm_len);

   rcreds->service.name_len = as_reply->enc_part2->server->data[1].length;
   rcreds->service.name = krbAllocBuf(rcreds->service.name_len+1);
   if ( rcreds->service.name == NULL )
   {
	   return ENOMEM;
   }
   memcpy(rcreds->service.name, as_reply->enc_part2->server->data[1].data,
         rcreds->service.name_len);

   /* Write the session key information to the credentials  */
   rcreds->session_key.session_key_len = as_reply->enc_part2->session->length;
   rcreds->session_key.session_key_type = as_reply->enc_part2->session->enctype;
   rcreds->session_key.session_key_val = krbAllocBuf(rcreds->session_key.session_key_len+1);
   memcpy(rcreds->session_key.session_key_val, as_reply->enc_part2->session->contents,
         rcreds->session_key.session_key_len);


   /* Write ticket times to the credentials */
   rcreds->tkt_times.auth_time = as_reply->enc_part2->times.authtime;
   rcreds->tkt_times.start_time = as_reply->enc_part2->times.starttime;
   rcreds->tkt_times.end_time = as_reply->enc_part2->times.endtime;
   rcreds->tkt_times.renew_till = as_reply->enc_part2->times.renew_till;

   /* Write the ticket information to the credentials */
   if ((retval = encode_krb5_ticket(as_reply->ticket, &packet)))
      goto cleanup;

   rcreds->tkt.tkt_len = packet->length;
   rcreds->tkt.tkt_val = krbAllocBuf(rcreds->tkt.tkt_len);
   memcpy(rcreds->tkt.tkt_val, packet->data, rcreds->tkt.tkt_len);

   /* write ticket, session key and times to NVRAM */
   if ( app_flag == APPLICATION_PROV )
   {
      tflag = AS_PROV_FLAG;
   }
   else
   {
      tflag = AS_CMS_FLAG;
   }

   LOG(("Writing Ticket to NVRAM ... "));
   SecLibGetNVRAMIface()->secWriteNVRAMTkt(rcreds->tkt.tkt_val,
         rcreds->tkt.tkt_len,
         tflag,
         as_reply->enc_part2->server->data[1].data,
         as_reply->enc_part2->server->data[1].length);
   LOG(("Writing Session Key and Ticket times to NVRAM ... "));
   SecLibGetNVRAMIface()->secWriteNVRAMSessionKeyAndTktTimes(
         as_reply->enc_part2->session->contents,
         as_reply->enc_part2->session->length,
         as_reply->enc_part2->session->enctype,
         tflag,
         as_reply->enc_part2->server->data[1].data,
         as_reply->enc_part2->server->data[1].length,
         as_reply->enc_part2->times.authtime,
         as_reply->enc_part2->times.starttime,
         as_reply->enc_part2->times.endtime,
         as_reply->enc_part2->times.renew_till);

   krbFreeBuf(packet->data);
   krbFreeBuf(packet);

cleanup:
   if (retval)
   {
         rcreds->session_key.session_key_len = 0;
         rcreds->tkt.tkt_len = 0;
   }
   return (retval);
}

/*
*****************************************************************************
** FUNCTION:     getASExpAlgInfo
**
** PURPOSE:      get necessary information for exp backoff algorithm
**
** PARAMETERS:
**               realm - realm
**               max_retries - out -
**               nom_timeout - out -
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
void getASExpAlgInfo(
      unsigned char    *realm,
      int              *max_retries,
      int              *nom_timeout)
{
   int         tmp;
   int         mt;
   int         nt;

#ifdef DEMO
   *max_retries = 2;   //4
   *nom_timeout = 10;
#else
   if ( mibAccessSec(
         MIBITEM(SEC, realmUnsolicitedKeyMaxRetries),
         NULL,
         realm,
         NULL,
         &tmp,
         (void*)&mt,
         MIB_ITEM_GET))
   { 
      LOG(("ERROR: MIB GET realmUnsolicitedKeyMaxRetries Failed for realm %s, set to default 10.\n", realm));
   }
   if ( mibAccessSec(
         MIBITEM(SEC, realmUnsolicitedKeyNomTimeOut),
         NULL,
         realm,
         NULL,
         &tmp,
         (void*)&nt,
         MIB_ITEM_GET) )
   { 
      LOG(("ERROR: MIB GET realmUnsolicitedKeyNomTimeOut Failed for realm %s, set to default 10.\n", realm));
   }
#endif

   LOG(("MIB VALUES mt = %d, nt = %d.", mt, nt));
   *max_retries = mt;
   *nom_timeout = nt;

   /* defaults if MIBS are empty */
   if ( *max_retries <= 0)
      *max_retries = 2;//4;
   if ( *nom_timeout <= 0)
      *nom_timeout = 10;

   LOG(("VALUES set to max retries = %d, nom timeout = %d.",
            *max_retries, *nom_timeout));
}


/*
*****************************************************************************
** FUNCTION:     send_as_request
**
** PURPOSE:      sends AS_REQ, received AS_REP, processes and returns credentials
**
** PARAMETERS:   mtafqdn - MTA FQDN
**               tgsfqdn -
**               app_flag - for IPSec or SNMPv3
**               krbServiceName - call agent principal name
**               krbServiceRealm - call agent realm
**               krbProvServFQDN - prov server fqdn - for preauth
**               krbto - timeout for sending AS_REQ (every 3 seconds)
**               credentials out - ticket, session key, etc. to be returned
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code send_as_request(
      const char   *mtafqdn,
      const char   *tgsfqdn,
      const int    app_flag,
      const char   *krbServiceName,
      const char   *krbServiceRealm,
      const char   *krbProvServFQDN,
      const int    krbto,
      MtaCredentials   *credentials)
{
   krb5_kdc_req        *request=0;
   krb5_data           *packet=0;
   krb5_data           *cert_data=0;
   krb5_data           reply;
   krb5_error          *err_reply=0;
   krb5_error_code     retval=0;
   krb5_kdc_rep        *as_reply = 0;
   krb5_timestamp      time_now;
   krb5_preauthtype    ptypes[1];
   krb5_pa_data        **preauth_to_use = 0;
   krb5_data           salt;
   pkinit_reply        *tpadata2=0;
   int                 i;
   int                 pkinit_nonce;
   int                 ret;
   priv_key_t          mtaprivkey;
	unsigned char       *pSecretKey;
	unsigned long       secretKeyLen = 0;
	unsigned long       status;
   krb5_data           dh_shared_key;
   char                *comp1;
   char                *comp2;
   int                 clock_skew;
   int                 orig_clock_skew;
   time_t              tmp_time;
   int                 tmp;
   krb5_typed_data     **e_data;
   krb5_data           scratch;
   int                 rep_nonce;
   int                 retries = 0;
   int                 max_retries = 0;
   int                 nomTO;
   int                 tafter;
   int                 tbefore;
   int                 rtt;
   int                 exp_flag;
   int                 utc_time_offset;
   unsigned char       dummy;
   
   int newSock;
   extern void NewKSend2(int sd);
   extern int NewKSock();

   

#ifdef WIN32
   FILE *doc_as_rep_bin, *as_req_bin, *my_secret_bin, *saved_priv_key_bin, *f1, *f2, *as_rep_bin;
   /* interop with Doc - read a pkinit reply part only */
#if 1
//   krb5_octet   treply[5120];
//   int flen;
//   krb5_octet   saved_priv_key[512];
//   krb5_octet   treq[5120];
//   int saved_pkey_len;


//   if (!fopened_once2)
//   {
//      doc_as_rep_bin = fopen("Asrep.bin","rb");
//   }
//   if (!fopened_once3)
//   {
//      as_req_bin = fopen("c:/doc_dh_stuff/as_req.bin","rb");
//   }
//   if (!fopened_once4)
//   {
// /     my_secret_bin = fopen("my_secret.bin","wb");
//   }
//   if (!fopened_once5)
//   {
//      saved_priv_key_bin = fopen("c:/doc_dh_stuff/privkey.bin","rb");
//   }
#endif

//   if (!fopened_once1)
//   {
//      f1 = fopen("pkinit_req.bin", "wb");
//      f2 = fopen("as_req.bin", "wb");
//   }

           
//   if (!fopened_once6)
//   {
//      as_rep_bin = fopen("as_rep.bin","wb");
 //  }
#endif

  // printf("\nCALLING NEWKSOCK\n\n");
  //newSock = NewKSock();
  //NewKSend2(newSock);

   request = (krb5_kdc_req *)krbAllocBuf(sizeof(krb5_kdc_req));
   if (request == NULL)
   {
      return ENOMEM;
   }
   memset(request, 0, sizeof(krb5_kdc_req));
   request->magic = KV5M_KDC_REQ;
   request->msg_type = KRB5_AS_REQ;
   request->kdc_options = 0;
   request->client = (krb5_principal)krbAllocBuf(sizeof(krb5_principal_data));
   if (request->client == NULL)
   {
      return ENOMEM;
   }
   request->server = (krb5_principal)krbAllocBuf(sizeof(krb5_principal_data));
   if (request->server == NULL)
   {
      return ENOMEM;
   }
   retval = setAsreqClientPrincipal(request, krbServiceRealm, mtafqdn);
   if ( retval )
   {
      printf("\tERROR Could not set AS_REQ client code %d.\n", retval);
      goto cleanup;
   }

      /* Figure out what the components are for service name */
   if (app_flag == APPLICATION_CMS)
   {
      comp1 = (char *)krbAllocBuf(strlen(CMS_SERVICE_COMP1)+1);
	   memcpy(comp1, CMS_SERVICE_COMP1, strlen(CMS_SERVICE_COMP1));
	   comp1[strlen(CMS_SERVICE_COMP1)] = '\0';
      comp2 = krbAllocBuf(strlen(krbServiceName)+1);
      memcpy(comp2, krbServiceName, strlen(krbServiceName));
      comp2[strlen(krbServiceName)] = '\0';
   }
   else if ( app_flag == APPLICATION_PROV )
   {
      /* for SNMPv3 */
      comp1 = (char *)krbAllocBuf(strlen(PROV_SERVICE_COMP1)+1);
      memcpy(comp1, PROV_SERVICE_COMP1, strlen(PROV_SERVICE_COMP1));
      comp1[strlen(PROV_SERVICE_COMP1)] = '\0';
      comp2 = krbAllocBuf(strlen(krbProvServFQDN)+1);
      memcpy(comp2, krbProvServFQDN, strlen(krbProvServFQDN));
      comp2[strlen(krbProvServFQDN)] = '\0';
   }
   else
   {
      printf("Bad service flag = %d.\n", app_flag);
      goto cleanup;
   }
   retval = setAsreqServerPrincipal(request,
         comp1,
         comp2,
         krbServiceRealm,
         app_flag);
   krbFreeBuf(comp1);
   krbFreeBuf(comp2);
   if ( retval )
   {
      printf("\tERROR Could not set AS_REQ server code %d.\n", retval);
      goto cleanup;
   }
   request->from = 0;   /* only for postdated tickets */
   time(&tmp_time);
#ifdef DEMO
   utc_time_offset = 0;
#else
   utc_time_offset = cmMiscGetUTCTimeOffset();
#endif
   LOG(("CM UTC Time Offset is %d.", utc_time_offset));
   tmp_time -= utc_time_offset;

   if ( app_flag == APPLICATION_CMS )
   {
	   clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(
            AS_CMS_FLAG, krbServiceRealm,
            strlen(krbServiceRealm));
   }
   else
   {
	   clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(
            AS_PROV_FLAG, krbServiceRealm,
            strlen(krbServiceRealm));
   }

   request->till = tmp_time + (FOUR_WEEK_IN_SECONDS) - clock_skew;
   printf("AS request till time (%d) = time(%d) + 4 weeks - clock_skew(%d).\n", request->till, tmp_time, clock_skew);
   request->nonce = rand();
   printf("as_req nonce is %d\n", request->nonce);
   request->ktype = (krb5_enctype *)krbCallocBuf(1, sizeof(krb5_enctype));
   request->nktypes  = 1;
   request->ktype[0] = ENCTYPE_DES3_CBC_MD5;    /* 5 is the value */


   /* make preauth to be salt */
#if 0
   ptypes[0] = KRB5_PADATA_ENC_TIMESTAMP;
   ptypes[1] = -1;
   retval = make_preauth_list(ptypes, -1, &preauth_to_use);
   printf("make_preauth_list retval = %d.\n", retval);
   if (retval)
      goto cleanup;

   if ((retval = krb5_obtain_padata(preauth_to_use, pwd_keyproc,
               NULL, NULL, request)) != 0)
   {
      printf("obtain_padata retval = %d.\n", retval);
      goto cleanup;
   }
#endif


   /* allocate 2 pre-auth fields (pkinit and provserv), plus 1 for delimiter */
   request->padata = (krb5_pa_data **)krbAllocBuf(3*sizeof(krb5_pa_data));
   if (request->padata == NULL)
   {
      return ENOMEM;
   }

   /* make pkinit and provserv preauth fields */
   status =
      make_krb5_preauth(request, krbProvServFQDN, &pkinit_nonce, &mtaprivkey, app_flag);

   if ( status )
   {
      printf("Problems making pre-auth. error code = %d\n", status);
      retval = status;
      goto cleanup;
   }


   /* encode in asn.1 */
   if ((retval = encode_krb5_as_req(request, &packet)) != 0)
   {
      printf("encode failed, retval=%d, length=%d\n", retval, packet->length);
      goto cleanup;
   }

#ifdef DEMO
#if 0
   if (!fopened_once1)
   {
      fwrite(packet->data, sizeof(char), packet->length, f2);
      fclose(f2);
      fopened_once1 = 1;
   }
#endif
#endif

   reply.length = 0;

#ifdef DEMO
#if 0
   if (!fopened_once3)
   {
      packet->length = fread(treq, 1, 5120, as_req_bin);
      packet->data = treq;
      fclose(as_req_bin);
      fopened_once3 = 1;
   }
#endif
#endif

#if 1
   /* send and apply exponential backoff retry if time out */
   /* get the maxRetries from MIB */
   getASExpAlgInfo(krbServiceRealm, &max_retries, &nomTO);
   if ( app_flag == APPLICATION_CMS )
   {
      exp_flag = AS_CMS_FLAG;
   }
   else
   {
      exp_flag = AS_PROV_FLAG;
   }

   while ( retries <= max_retries )
   {
      time(&tbefore);
      printf(".k  send_as_request: - CALLING  krb5_sendto_server()\n");
      /* send  to server */
      //-trc
      DumpBuf("SENDING AS REQUEST", (unsigned char *) packet->data, packet->length, 'x'); 

      if ((retval = krb5_sendto_server(packet, tgsfqdn,
                  AS_UDP_PORT_NUMBER, nomTO, &reply)) != 0)
      {
         time(&tafter);
         rtt = tafter - tbefore;
         retries++;
         LOG(("Elapsed time is %d.", rtt));
         printf("AS REQ timed out, retval=%d, retries=%d\n", retval, retries);
         if ( retries > max_retries )
         {
            LOG(("Reached max number of retries."));
            secUpdateExpBackoffRetryTable(
                  exp_flag,
                  rtt,
                  1,
                  0,   /* endpoint ? */
                  krbServiceRealm);
            goto cleanup;
         }
         else
         {
            secUpdateExpBackoffRetryTable(
                  exp_flag,
                  rtt,
                  1,
                  0,   /* endpoint ? */
                  krbServiceRealm);
         }
      }
      else
      { 
         printf(".k  send_as_request: - Back from krb5_sendto_server() - break\n");
         break;
      }
      nomTO = secGetRTT(exp_flag, krbServiceRealm);
   }
#endif
   printf(".k  send_as_request: - CALLING  secUpdateExpBackoffRetryTable()\n");

   time(&tafter);
   rtt = tafter - tbefore;
   secUpdateExpBackoffRetryTable(
         exp_flag,
         rtt,
         0,
         0,   /* endpoint ? */
         krbServiceRealm);


#if 0
   reply.length = sizeof(as_rep_txt);
   reply.data = as_rep_txt;
#endif

   /* now decode the reply...could be error or as_rep */
#ifdef DEMO
   /* read an external reply (doc's) */
#if 0
   if (!fopened_once2)
   {
      flen = fread(treply, 1, 5120, doc_as_rep_bin);
      fclose(doc_as_rep_bin);
      fopened_once3 = 1;
   }
   reply.length = flen;
   reply.data = treply;
#endif
#endif

#ifdef DEMO
#if 0
   if (!fopened_once6)
   {
      fwrite(reply.data, sizeof(char), reply.length, as_rep_bin);
      fclose(as_rep_bin);
      fopened_once6 = 1;
   }
#endif
#endif

   printf(".k  send_as_request: - Now decode the reply\n");

   /* now decode the reply...could be error or as_rep */
   if (krb5_is_krb_error(&reply))
   {
      printf(".k  send_as_request: - Decoding.....\n");

      if ((retval = decode_krb5_error(&reply, &err_reply)))
      {
         /* some other error code--??? */
         printf("err_reply bad error code %d\n", err_reply->error);
         return 1;
      }
      if (err_reply)
      {
         printf("KRB-ERROR received and decoded, errcode = %d\n",
               err_reply->error);
         /* verify error message is type REQ-NONCE and nonce matches */
         if (( retval =
               decode_krb5_typeddata_sequence(&(err_reply->e_data), &e_data)))
         {
            printf("KRB-ERROR e_data (REQ_NONCE) can not be decoded.\n");
            goto cleanup;
         }
         scratch.data = (*e_data)->contents;
         scratch.length = (*e_data)->length;
         /* Now decode the req-nonce inside e-data */
         if (( retval = decode_krb5_typeddata_req_seq(&scratch, &rep_nonce)))
         {
            printf("KRB-ERROR e_data can not be decoded.\n");
            goto cleanup;
         }

         /* Make sure the nonce is the same as AS-REQ */
         if ( ((*e_data)->data_type != KRB5_TD_REQ_NONCE) ||
               (rep_nonce != request->nonce) )
         {
            printf("invalid KRB_ERROR reply.\n");
            retval =  KRB5KRB_AP_ERR_BAD_INTEGRITY;
            goto cleanup;
         }
         else
         {
            LOG(("KRB-ERROR: e_data AS-REQ Nonce matched (%d).", rep_nonce));
         }
         /* free e_data */
         krbFreeBuf((*e_data)->contents);
         krbFreeBuf(*e_data);
         krbFreeBuf(e_data);

         if (err_reply->error == KDC_ERR_PREAUTH_REQUIRED &&
               err_reply->e_data.length > 0)
         {
            printf("padate error %d\n", err_reply->error);
            retval = decode_krb5_padata_sequence(&err_reply->e_data,
                  &preauth_to_use);
            if (retval)
               printf("as_reply padata error %d\n", retval);
         }
         else
         {
            retval = err_reply->error + ERROR_TABLE_BASE_krb5;
         }

         /* decode and print the text part of keb-error */
         if ( err_reply->text.length > 0 )
         {

            err_reply->text.data[err_reply->text.length] = '\0';
            printf("Text: %s.\n", err_reply->text.data);
         }
         if ( retval == KRB5KRB_AP_ERR_SKEW )
         {
            clock_skew = err_reply->ctime - err_reply->stime;
            printf("Clock Skew error ctime=%d, stime=%d, skew=%d.\n",
                  err_reply->ctime, err_reply->stime, clock_skew);
#if 0
//#ifndef DEMO
//            if ( ! cmNvIsStandaloneEnabled() )
//            {
//#endif
            if ( labs(clock_skew) > MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS )
            {
                  printf("SKEW GREATER THAN MAXIMUM ALLOWED (1 HR).\n");
               retval = KRB5_KDCREP_SKEW;
                  goto cleanup;
               }
            else
            {
               /* Write to NVRAM */
               if ( app_flag == APPLICATION_CMS )
               {
                  /* add to the original skew */
                     orig_clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(
                           AS_CMS_FLAG,
                        krbServiceRealm,
                        strlen(krbServiceRealm));
                  clock_skew += orig_clock_skew;

                     SecLibGetNVRAMIface()->secWriteNVRAMSkew(
                           clock_skew, AS_CMS_FLAG, krbServiceRealm,
                        strlen(krbServiceRealm));
               }
               else  /* bad flag has already been taken care of */
               {
                  /* add to the original skew */
                     orig_clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(
                           AS_PROV_FLAG,
                        krbServiceRealm,
                        strlen(krbServiceRealm));
                  clock_skew += orig_clock_skew;

                     SecLibGetNVRAMIface()->secWriteNVRAMSkew(
                           clock_skew, AS_PROV_FLAG, krbServiceRealm,
                        strlen(krbServiceRealm));
               }
            }
#endif
//#ifndef DEMO
//            }
//            else
//            {
            /* Use this code for now, the real code is the above #if0'd code */
            if ( labs(clock_skew) > MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS )
            {
                  printf("AS SKEW GREATER THAN MAXIMUM ALLOWED (1 hr), Ignored in Standalone.\n");
               }
            /* Write to NVRAM */
            if ( app_flag == APPLICATION_CMS )
            {
               /* add to the original skew */
                  orig_clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(
                        AS_CMS_FLAG,
                     krbServiceRealm,
                     strlen(krbServiceRealm));
               clock_skew += orig_clock_skew;

                  SecLibGetNVRAMIface()->secWriteNVRAMSkew(
                        clock_skew, AS_CMS_FLAG, krbServiceRealm,
                     strlen(krbServiceRealm));
            }
            else  /* bad flag has already been taken care of */
            {
                  orig_clock_skew =  SecLibGetNVRAMIface()->secReadNVRAMSkew(AS_PROV_FLAG,
                     krbServiceRealm,
                     strlen(krbServiceRealm));
               clock_skew += orig_clock_skew;

                  SecLibGetNVRAMIface()->secWriteNVRAMSkew(
                        clock_skew, AS_PROV_FLAG, krbServiceRealm,
                     strlen(krbServiceRealm));
            }
            goto cleanup;
         }
//#endif
//         }
         goto cleanup;
      }
      goto cleanup;
   }
   
   printf(".k  send_as_request: - Not Decoding err - Must be KRB_AS_REP or bad ret packet\n");

   /* It must be a KRB_AS_REP message, or an bad returned packet */
   if ((retval = decode_krb5_as_rep(&reply, &as_reply)))
   {
      /* some other error code ??? */
      printf("as_reply bad error code %d\n", retval);
      return 1;
   }

   if (as_reply->msg_type != KRB5_AS_REP)
   {
      retval = KRB5KRB_AP_ERR_MSG_TYPE;
      printf("as_reply bad msg code %d\n", retval);
      return 1;
   }

   if ( as_reply->padata )
   {
      tpadata2 = (pkinit_reply *)krbAllocBuf(sizeof(pkinit_reply));
      memset(tpadata2, 0, sizeof(pkinit_reply));
      if (tpadata2 == NULL)
      {
         return ENOMEM;
      }

      retval = get_krb5_pkinit_rep(*(as_reply->padata), &tpadata2);
      if ( retval )
      {
         printf("Error decoding PKINIT. Error code %d.\n", retval);
         goto cleanup;
      }
   }
   else
   {
      printf("EMPTY padata.\n");
      goto cleanup;
   }

   if (! tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.signature.signature_len)
   {
      printf("Bad Signature.\n");
      retval = 1;
      goto cleanup;
   }

   printf("pkinit reply content_type: length = %d : ", tpadata2->dh_signed_data.content_type.length);
   for (i=0;i<tpadata2->dh_signed_data.content_type.length;i++)
      printf(".%d", tpadata2->dh_signed_data.content_type.oid_array[i]);
   printf("\n");
   printf("cms_version = %d .\n", tpadata2->dh_signed_data.signed_data.cms_version);
   printf("DigestAlgorithmId oid: length=%d : ", tpadata2->dh_signed_data.signed_data.digest_alg_ids.digest_alg_id.alg_oid.length);
   for ( i=0;
        i<tpadata2->dh_signed_data.signed_data.digest_alg_ids.digest_alg_id.alg_oid.length;
        i++)
      printf(".%d", tpadata2->dh_signed_data.signed_data.digest_alg_ids.digest_alg_id.alg_oid.oid_array[i]);
   printf("\n");

   printf("encapsulatedContentInfo oid length =%d: ", tpadata2->dh_signed_data.signed_data.encap_content_info.econtent_type.length);
   for ( i=0;
        i<tpadata2->dh_signed_data.signed_data.encap_content_info.econtent_type.length;
        i++)
      printf(".%d", tpadata2->dh_signed_data.signed_data.encap_content_info.econtent_type.oid_array[i]);
   printf("\n");


   /* check and decode auth_pack octet string */
   if ( tpadata2->dh_signed_data.signed_data.encap_content_info.length )
   {
      get_krb5_auth_pack_reply(&tpadata2->dh_signed_data.signed_data.encap_content_info);
   }
   else
   {
      printf("EMPTY econtent.\n");
      goto cleanup;
   }

   if (tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.nonce == pkinit_nonce)
   {
      printf("*** pkinit nonces match %d\n", pkinit_nonce);
   }

   /* re-encode the certificate into a sequence for openSSL */
   for (i=0;i<tpadata2->dh_signed_data.signed_data.certificates.length;i++)
   {
      re_encode_cert(&tpadata2->dh_signed_data.signed_data.certificates.certs[i], &cert_data);
      krbFreeBuf(tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert);
      tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert = (unsigned char *)krbAllocBuf(cert_data->length);
      memcpy(tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert, cert_data->data, cert_data->length);
      tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert_len = cert_data->length;
      krbFreeBuf(cert_data->data);
      cert_data->length = 0;
   }

   if ( (status = check_pkinit_reply(tpadata2, krbServiceRealm, app_flag)) )
   {
      printf("\n\tPKINIT REPLY Check failed code = %d\n", status);
      retval = status;
      goto cleanup;
   }


   printf("DH public key length = %d ",  tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key_len);
   printf("\n");

   printf("Certificates in the as_rep: %d\n", tpadata2->dh_signed_data.signed_data.certificates.length);
   for (i=0;i<tpadata2->dh_signed_data.signed_data.certificates.length;i++)
   {
      printf("certificate number %d length= %d.\n", i, tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert_len);
   }

   printf("signer_info cms_version= %d.\n", tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.cms_version);
   printf("signerInfo sid signer_name_len = %d. \n",
         tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name_len);
   printf("signerInfo sid serial_num = %d. \n",
         tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_sernum);


   printf("signerInfo digest_alg_id oid len = %d: ",
         tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.length);
   for (i=0;i<tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.length;i++)
      printf(".%d", tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.digest_alg_id.alg_oid.oid_array[i]);
   printf("\n");


   printf("signerInfo signature len = %d ",  tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.signature.signature_len);
   printf("\n");



#ifdef DEMO
#if 0
   /* read saved private key for doc's as_reply */
   if (!fopened_once5)
   {
      saved_pkey_len = fread(saved_priv_key, 1, 512, saved_priv_key_bin);
      fclose(saved_priv_key_bin);
      fopened_once5 = 1;
   }
   status = SecLibGetCryptoIface()->secCryptoDhComputeSecret(
         &pSecretKey,
         &secretKeyLen,
         saved_priv_key,
         saved_pkey_len,
         tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key,
         tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key_len,
         MTA_OAKLEY2_PRIME,
         MTA_OAKLEY2_PRIME_LEN,
         OAKLEY2_BASE );
#endif
#endif


#if 1
   status = SecLibGetCryptoIface()->secCryptoDhComputeSecret(
         &pSecretKey,
         &secretKeyLen,
         mtaprivkey.priv_key,
         mtaprivkey.priv_key_len,
         tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key,
         tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key_len,
         MTA_OAKLEY2_PRIME,
         MTA_OAKLEY2_PRIME_LEN,
         OAKLEY2_BASE );
#endif

#ifdef DEMO
#if 0
   if (!fopened_once4)
   {
      fwrite(pSecretKey, sizeof(char), secretKeyLen, my_secret_bin);
      fclose(my_secret_bin);
      fopened_once4 = 1;
   }
#endif
#endif

	if ( status )
	{
		printf("\tFailed to generate shared DH secret.\n");
      retval = status;
      goto cleanup;
	}

   dh_shared_key.length = secretKeyLen;
   dh_shared_key.data = pSecretKey;

   if ((retval = decrypt_as_reply(request, as_reply, dh_keyproc,
               &dh_shared_key, NULL, krb5_kdc_rep_decrypt_proc, 0)))
   {
      printf("decrypt failed errorcode %d\n", retval);
      goto cleanup;
   }

   time(&time_now);
   if ((retval = verify_as_reply(time_now, request, as_reply)))
   {
      printf("AS Reply Verify failed - bad as_reply.\n");
      goto cleanup;
   }
   if ((retval = verify_pkinit_reply(tpadata2)))
   {
      printf("PKINTI Reply Verify failed - bad as_reply.\n");
      goto cleanup;
   }

   if ( request->nonce == as_reply->enc_part2->nonce)
   {
      printf("*** Krb nonces match %d.\n", request->nonce);
   }

   /* store credentials to be returned */
   store_creds(as_reply, credentials, app_flag);

   printf("session key type=%d, length=%d \n",
         as_reply->enc_part2->session->enctype,
         as_reply->enc_part2->session->length);
   for ( i=0;i<as_reply->enc_part2->session->length;i++ )
      printf("%x ", as_reply->enc_part2->session->contents[i]);
   printf("\n");



cleanup:
   if (request->padata)
   {
      krb5_free_pa_data(request->padata);
   /*   printf("Cleaned request padata.\n"); */
   }
   if ( request->client )
   {
      krbFreeBuf(request->client->realm.data);
      krbFreeBuf(request->client->data->data);
      krbFreeBuf(request->client->data);
      krbFreeBuf(request->client);
    /*  printf("Cleaned request client.\n"); */
   }
   if (request->server)
   {
      krbFreeBuf(request->server->realm.data);
      krbFreeBuf(request->server->data->data);
      krbFreeBuf(request->server->data);
      krbFreeBuf(request->server);
      /* printf("Cleaned request server.\n"); */
   }
   if ( request->ktype )
   {
      krbFreeBuf(request->ktype);
      /* printf("Cleaned request ktype.\n"); */
   }
   if ( request )
   {
      krbFreeBuf(request);
      /* printf("Cleaned request.\n"); */
   }

   if ( packet->length )
   {
      krbFreeBuf(packet->data);
      krbFreeBuf((char *)packet);
      /* printf("cleaned packet.\n"); */
   }
   if (as_reply)
   {
      krb5_free_kdc_rep(as_reply);
      /* printf("cleaned as_reply.\n"); */
   }
   if ( err_reply )
   {
      krbFreeBuf((char *)err_reply);
      /* printf("cleaned err_reply.\n"); */
   }

   if ( reply.length )
      krbFreeBuf(reply.data);

   if ( tpadata2 == 0 )
   {
      return retval;
   }


   if ( tpadata2->dh_signed_data.signed_data.encap_content_info.length )
   {
      tpadata2->dh_signed_data.signed_data.encap_content_info.length = 0;
      krbFreeBuf((char *)tpadata2->dh_signed_data.signed_data.encap_content_info.contents);
   }

   if ( tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key_len )
   {
      tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key_len = 0;
      krbFreeBuf((char *)tpadata2->dh_signed_data.signed_data.encap_content_info.econtent.tgs_pub_value.tgs_pub_key);
   }

   for (i=0;i<tpadata2->dh_signed_data.signed_data.certificates.length;i++)
   {
      tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert_len = 0;
      krbFreeBuf((char *)tpadata2->dh_signed_data.signed_data.certificates.certs[i].cert);
   }


   if ( tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name_len )
   {
      tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name_len = 0;
      krbFreeBuf(tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.sid.issuer_sernum.signer_name);
   }

   if ( tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.signature.signature_len )
   {
      tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.signature.signature_len = 0;
      krbFreeBuf(tpadata2->dh_signed_data.signed_data.signer_infos.signer_info.signature.signature_val);
   }


   if ( tpadata2 )
      krbFreeBuf(tpadata2);

   if ( mtaprivkey.priv_key_len )
   {
      mtaprivkey.priv_key_len = 0;
      SEC_FREE(mtaprivkey.priv_key);
   }
    if ( secretKeyLen )
    {
       secretKeyLen = 0;
       SEC_FREE(pSecretKey);
    }

   return (retval);

}
/*
*****************************************************************************
** FUNCTION:     getCredsFromNVRAM
**
** PURPOSE:      get the credentials from NVRAM.
**
** PARAMETERS:   krbServiceName - service name (call agent or prov server)
**               application flag
**               credentials - credentials filled up if found in NVRAM
**
** RETURNS:      success or failue
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code getCredsFromNVRAM(
      const char      *krbServiceName,
      App_flag        app_flag,
      MtaCredentials  *credentials)
{
   int                tflag;
   unsigned char      nvram_tkt[MAX_TKT_LENGTH];
   krb5_data          tkt_data;
   unsigned char      skey[MAX_SESSION_KEY_LENGTH+16];
   int                nvram_tkt_len;
   int                dummy;
   krb5_ticket        **dec_tkt;
   krb5_error_code    retval;
   int                tmp_time;
   int                i;


   if ( app_flag == APPLICATION_CMS )
      tflag = AS_CMS_FLAG;
   else
      tflag = AS_PROV_FLAG;

   nvram_tkt_len = SecLibGetNVRAMIface()->secReadNVRAMTkt(
         nvram_tkt, MAX_TKT_LENGTH, tflag,
         krbServiceName,
         strlen(krbServiceName));
   if ( ! nvram_tkt_len )
   {
      printf("No NVRAM Ticket found for service %s.\n", krbServiceName);
      return 0;
   }

   tkt_data.data = krbAllocBuf(nvram_tkt_len);
   tkt_data.length = nvram_tkt_len;
   memcpy(tkt_data.data, nvram_tkt, nvram_tkt_len);
   dec_tkt = (krb5_ticket **)krbAllocBuf(sizeof(krb5_ticket *));
   if ((retval = decode_krb5_ticket(&tkt_data, dec_tkt)))
      printf("Unable to decode the ticket stored in NVRAM service %s.\n",
            krbServiceName);


   credentials->session_key.session_key_len =
      SecLibGetNVRAMIface()->secReadNVRAMSessionKeyAndTktTimes(
            skey,
            &credentials->session_key.session_key_type,
            tflag,
            krbServiceName,
            strlen(krbServiceName),
            &credentials->tkt_times.auth_time,
            &credentials->tkt_times.start_time,
            &credentials->tkt_times.end_time,
            &credentials->tkt_times.renew_till);

   if ( !credentials->session_key.session_key_len )
      printf("No NVRAM Ticket times found for service %s.\n", krbServiceName);
   else
      printf("Session Key Retrieved from NVRAM.\n");

   printf("NVRAM Ticket auth_time=%d, start_time=%d,end_time=%d,renew_till=%d \n",
         credentials->tkt_times.auth_time,
         credentials->tkt_times.start_time,
         credentials->tkt_times.end_time,
         credentials->tkt_times.renew_till);

   time(&tmp_time);
   if ( credentials->tkt_times.end_time > tmp_time )
   {
      printf("Valid ticket found for service %s retrieving ...\n", krbServiceName);
      /* copy cerdentials and return it */
      /* copy ticket */
      credentials->tkt.tkt_val = krbAllocBuf(nvram_tkt_len);
      memcpy(credentials->tkt.tkt_val, nvram_tkt, nvram_tkt_len);
      credentials->tkt.tkt_len = nvram_tkt_len;

      credentials->session_key.session_key_val =
         krbAllocBuf(credentials->session_key.session_key_len);
      memcpy(credentials->session_key.session_key_val, skey,
            credentials->session_key.session_key_len);

      printf("Session Key: ");
      for (i=0;i<credentials->session_key.session_key_len;i++)
         printf("%x ", credentials->session_key.session_key_val[i]);
      printf("\n");

      /* just return a non-zero value */
      retval = 1;
   }
   else
   {
      printf("Ticket expired, get new credentials.\n");
      retval = 0;
   }


cleanup:
   krb5_free_ticket(*dec_tkt);
   krbFreeBuf(dec_tkt);
   krbFreeBuf(tkt_data.data);
   return(retval);
}
/*
*****************************************************************************
** FUNCTION:     secGetKrbCredentials
**
** PURPOSE:      get the credentials by sending an AS-REQ.
**
** PARAMETERS:   mtafqdn - MTA FQDN
**               tgsfqdn -
**               app_flag - for IPSec or SNMPv3
**               krbServiceName - service name (call agent or prov server)
**               krbServiceRealm - realm of the service (CA or provServer)
**               krbProvServFQDN - provisioning server location (for pre-auth)
**               krbto - timeout for sending AS_REQ (every 3 seconds)
**               krbreason - bootup, ticket expiry, wakeup/rekey recieved, etc.
**               credentials - credentials gathered
**
** RETURNS:      success or failue
**
** NOTE:
**
**
*****************************************************************************
*/
UINT32 secGetKrbCredentials(
      const char *mtafqdn,
      const char *tgsfqdn,
      App_flag   app_flag,
      const char *krbServiceName,
      const char *krbServiceRealm,
      const char *krbProvServFQDN,
      const int  krbto,
      int        krbreason,
      MtaCredentials  *credentials)
{
   krb5_error_code    retval;
   printf(".k  secGetKrbCredentials: -start\n");

   krbInitBufs();

   /* if ticket has expired or within grace period, or if we got a BADKEYVER error, 
    * renew */
   if ( ( krbreason != KRB_TICKET_EXPIRED ) && ( krbreason != KRB_BADKEYVER_ERROR ) )
   {
      if ( app_flag == APPLICATION_CMS )
         retval =  getCredsFromNVRAM(krbServiceName, app_flag, credentials );
      else
         retval =  getCredsFromNVRAM(krbProvServFQDN, app_flag, credentials );


      /* see if a ticket is already issued and stored in NVRAM */
      if ( retval )
      {
         return 0;
      }
   }
   printf(".k  secGetKrbCredentials: -calling send_as_request\n");
   retval = send_as_request(
         mtafqdn,
         tgsfqdn,
         app_flag,
         krbServiceName,
         krbServiceRealm,
         krbProvServFQDN,
         krbto,
         credentials);
   printf(".k  secGetKrbCredentials: -returned from send_as_request with(%d)\n",retval);
   
   /* krbPrintBufCounts(); */

   /* if time skew, resend immediately with adjusted time */
   if ( retval == KRB5KRB_AP_ERR_SKEW )
   {
      printf("Resending AS-REQ with adjusted time.\n");
      krbInitBufs();
      retval = send_as_request(
            mtafqdn,
            tgsfqdn,
            app_flag,
            krbServiceName,
            krbServiceRealm,
            krbProvServFQDN,
            krbto,
            credentials);
	  /* krbPrintBufCounts(); */
   }
   else if ( retval == KRB5_KDCREP_SKEW )
   {
      LOG(("CAN'T TRUST KDC, SKEW GREATER THAN 1 HR."));
      return retval;
   }

   if ( !retval )
   {
      credentials->service.realm[credentials->service.realm_len] = '\0';
      credentials->service.name[credentials->service.name_len] = '\0';

	   printf("\t*** Ticket Granted from %s %s ***\n",
		   credentials->service.realm, credentials->service.name);
      return 0;
	}
   else
   {
      return retval;
   }
}




/*
******************************************************************************
** FUNCTION:   secFreeCredentials
**
** PURPOSE:    Frees allocated name, realm, ticket and key in credentials
**
** PARAMETERS:
**             credentials - in
**
** RETURNS:    nothing
**
** NOTE:
**
******************************************************************************
*/
void secFreeCredsAndAppkey(MtaCredentials *credentials,
      MtaApplicationKey *app_key)
{
   /* cleanup allocated memory */
   if ( credentials->service.realm_len )
   {
      krbFreeBuf(credentials->service.realm);
      credentials->service.realm_len = 0;
   }
   if ( credentials->service.name_len )
   {
      krbFreeBuf(credentials->service.name);
      credentials->service.name_len = 0;
   }
   if ( credentials->session_key.session_key_len )
   {
      krbFreeBuf(credentials->session_key.session_key_val);
      credentials->session_key.session_key_len = 0;
   }
   if ( credentials->tkt.tkt_len )
   {
      krbFreeBuf(credentials->tkt.tkt_val);
      credentials->tkt.tkt_len = 0;
   }
   if ( app_key->sub_key.sub_key_len )
   {
      free(app_key->sub_key.sub_key_val);
      app_key->sub_key.sub_key_len = 0;
   }
}
