/***************************************************************************
*    Copyright 2001  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: expBackoffRetry.c
*    Creation Date: 27 November 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.2 $
*        $Date: 2005/02/17 21:07:14 $
*
****************************************************************************
*    Description:
*
*      Header file for Implementation of exponential backoff and retry
*      algorithm for security
*
****************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <hausenv.h>
#include <ChKerbNamespace.h>
#ifdef PSOS
#include <MibAccessSec.h>
#endif
#include "timeTable.h"
///#include "krbtasks.h"

#ifdef DEMO
#define GLOBAL_CFG_MAX_ENDPT   8
#else
#include <pktcMib.h>
#endif

#define KRB_MIN_TIMEOUT 3

///#include "MibAccessSec.h"

/* total of 18 entries needed:
 * 1 for PROV AS - index 0
 * 1 for PROV AP - index 1
 * 8 for CMS AS - index 2 to 9
 * 8 for CMS AP - index 10 to 17
 * */
#define MAX_EXP_TABLE (2+(2*GLOBAL_CFG_MAX_ENDPT))



typedef struct
{
   int              expRTT;
   int              expRTO;
   unsigned char    fqdn[255];
}  exp_table_entry;

/* Tables and entries to keep the exponential backoff retry data */
static exp_table_entry         gExpTable[MAX_EXP_TABLE];


/*
******************************************************************************
** FUNCTION:  secUpdateExpBackoffRetryTable
**
** PURPOSE:   Update/get timers from back off retry table
**
** PARAMETERS:   flag - AS or AP, prove or CMS
**               endpt - end point ID in case of ant CMS timers
**               name - realm, cms fqdn, or any other name IDs
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
void secUpdateExpBackoffRetryTable(
      int                  flag,
      int                  rtt,
      int                  retransmission,
      int                  endpt,
      unsigned char        *name)
{
   int                maxTO = 0;
   int                ix;
   int                i;
   int                tmp;
   unsigned char      lname[255];
   unsigned char      *p;

   
   strcpy(lname, name);
   p = lname;
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }

   /* determine the entry by looking at the flag */
   switch (flag)
   {
      case AS_PROV_FLAG:
         ix = 0;
#ifdef DEMO
         maxTO = 60;
#else
         if ( mibAccessSec(
               MIBITEM(SEC, realmUnsolicitedKeyMaxTimeOut),
               NULL,
               name,
               NULL,
               &tmp,
               (void*)&maxTO,
               MIB_ITEM_GET) )
         { 
            LOG(("ERROR: MIB GET realmUnsolicitedKeyMaxTimeOut Failed for realm %s.", 
                     name));
         }
#endif
         break;
      case AP_PROV_FLAG:
         ix = 1;
#ifdef DEMO
         maxTO = 30;
#else
         if ( mibAccessSec(
               MIBITEM(SEC, provUnsolicitedKeyMaxTO),
               NULL,
               NULL,
               NULL,
               &tmp,
               (void*)&maxTO,
               MIB_ITEM_GET))
         { 
            LOG(("ERROR: MIB GET provUnsolicitedKeyMaxTO Failed."));
         }

#endif         
		 break;
      case AS_CMS_FLAG:
#ifdef DEMO
         maxTO = 30;
#else
         if ( mibAccessSec(
               MIBITEM(SEC, realmUnsolicitedKeyMaxTimeOut),
               NULL,
               name,
               NULL,
               &tmp,
               (void*)&maxTO,
               MIB_ITEM_GET))
         { 
            LOG(("ERROR: MIB GET realmUnsolicitedKeyMaxTimeOut Failed, for realm %s.", 
                     name));
         }
#endif
         LOG(("Update EXP table: MAX TO from MIB for %s is %d.", name, maxTO));
         for ( i=2;i<(MAX_EXP_TABLE-GLOBAL_CFG_MAX_ENDPT);i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lname, gExpTable[i].fqdn,strlen(lname)))
            {
               ix = i;
               LOG(("Matched lname=%s, gExpTable[%d].fqdn = %s.", 
                        lname, i, gExpTable[i].fqdn));
               break;
            }
         }
         if ( i == (MAX_EXP_TABLE-GLOBAL_CFG_MAX_ENDPT) )
         {
            for ( i=2;i<(MAX_EXP_TABLE-GLOBAL_CFG_MAX_ENDPT);i++ )
            {
               if ( gExpTable[i].fqdn[0] == '\0' )
               {
                  memcpy(gExpTable[i].fqdn, lname, strlen(lname));
                  gExpTable[i].fqdn[strlen(lname)] = '\0';
                  gExpTable[i].expRTO = maxTO;
                  ix = i;
                  LOG(("Write to lname=%s, gExpTable[%d].fqdn = %s.", 
                           lname, i, gExpTable[i].fqdn));
                  break;
               }
            }
         }
         if ( i == (MAX_EXP_TABLE-GLOBAL_CFG_MAX_ENDPT) )
         {
            LOG(("PANIC: No more gExpTable Entries left for AS CMS."));
            return;
         }
         break;
      case AP_CMS_FLAG:
         /* Find the endpt corresponding to this call agent */
#ifdef DEMO
         maxTO = 30;
#else
         if ( mibAccessSec(
               MIBITEM(SEC, cmsMaxUnsolicitedKeyMaxTimeOut),
               NULL,
               name,
               NULL,
               &tmp,
               (void*)&maxTO,
               MIB_ITEM_GET))
         {
            LOG(("ERROR: MIB GET cmsMaxUnsolicitedKeyMaxTimeOut Failed, for realm %s.", 
                     name));
         }

#endif
         LOG(("Update EXP table: MAX TO from MIB for %s is %d.", name, maxTO));
         for ( i=(2+GLOBAL_CFG_MAX_ENDPT);i<MAX_EXP_TABLE;i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lname, gExpTable[i].fqdn,strlen(lname)))
            {
               ix = i;
			   break;
            }
         }
         if ( i == MAX_EXP_TABLE )
         {
            for ( i=(2+GLOBAL_CFG_MAX_ENDPT);i<MAX_EXP_TABLE;i++ )
            {
               if ( gExpTable[i].fqdn[0] == '\0')
               {
                  memcpy(gExpTable[i].fqdn, lname, strlen(lname));
                  gExpTable[i].fqdn[strlen(lname)] = '\0';
                  gExpTable[i].expRTO = maxTO;
                  ix = i;
                  break;
               }
            }
         }
         if ( i == MAX_EXP_TABLE )
         {
            LOG(("PANIC: No more gExpTable Entries left for AP CMS."));
            return;
         }
         break;
      default:
         printf("EXPONENTIAL BACKOFF BAD FLAG %d.\n", flag);
         return;
         break;
   }

   if ( !maxTO )
      maxTO = 30;
      
   if (retransmission)
   {
      /*
      ** Update RTT, RTO after a retransmission:
      ** 1) double the estimated value of the average delay
      ** 2) rnd = rand value between (0.5 * RTT) -- (1.0 * RTT)
      ** 3) RTO = min( rnd + N*RTT, maxTO )
      */

      /* Update RTT estimate: double the average delay (cap to 2^16) */
      if ( ! gExpTable[ix].expRTT )
         gExpTable[ix].expRTT = rtt;
      gExpTable[ix].expRTT = min( gExpTable[ix].expRTT * 2, (1<<16) );

      /* rand(0.5*RTT,1.0*RTT) = ( (0-255)*RTT + 256*RTT ) / 512 */
      gExpTable[ix].expRTO  = (((rand() & 0xff) * gExpTable[ix].expRTT + 
               (gExpTable[ix].expRTT<<8)) >>9);

      /* RTO = rnd + N*RTT
       * RTO is bounded by min RTO value:KRB_MIN_TIMEOUT,and maximum RTO value: TOLIMIT */
      gExpTable[ix].expRTO = max( min(gExpTable[ix].expRTO, maxTO), KRB_MIN_TIMEOUT );
   }

   /* Ack received */
   else
   {
      /* Update values RTT and RTO after receiving a response */

      /* Update RTT estimate: 7/8 calc from TCP */
      gExpTable[ix].expRTT = (((gExpTable[ix].expRTT<<3) - 
               gExpTable[ix].expRTT) + rtt) >>3;

      /* Update RTO: (average delay) + N * (avg. deviation)
       * RTO is bounded by min RTO value:KRB_MIN_TIMEOUT,and maximum RTO value: TOLIMIT */
      gExpTable[ix].expRTO = gExpTable[ix].expRTT ;
      gExpTable[ix].expRTO = max( min(gExpTable[ix].expRTO, maxTO), KRB_MIN_TIMEOUT );
   }

   LOG(("Security exponenatial backoff rtt=%d, RTT=%d, RTO=%d", 
            rtt, 
            gExpTable[ix].expRTT, 
            gExpTable[ix].expRTO));
}

/*
******************************************************************************
** FUNCTION: secGetRTT 
**
** PURPOSE:   get timers from back off retry table
**
** PARAMETERS:   flag - AS or AP, prov or CMS
**               name - realm or RQDN name
**
** RETURNS:
**
** NOTE:
**
******************************************************************************
*/
unsigned int secGetRTT(int flag, unsigned char *name)
{
   int    i;
   unsigned char   lname[255];
   unsigned char   *p;

   
   strcpy(lname, name);
   p = lname;
   while(*p)
   {
      *p++ = (char)tolower((int)*p);
   }


   /* determine the entry by looking at the flag */
   switch (flag)
   {
      case AS_PROV_FLAG:
         return gExpTable[0].expRTO;
         break;
      case AP_PROV_FLAG:
         return gExpTable[1].expRTO;
         break;
      case AS_CMS_FLAG:
         /* Find the endpt corresponding to this call agent */
         for ( i=2;i<(MAX_EXP_TABLE-GLOBAL_CFG_MAX_ENDPT);i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lname, gExpTable[i].fqdn,strlen(lname)))
            {
               return gExpTable[i].expRTO;
            }
         }
         printf("EXPONENTIAL BACKOFF NAME %s NOT FOUND IN AS CMS TABLE.\n", name);
         return 0;
         break;
      case AP_CMS_FLAG:
         /* Find the endpt corresponding to this call agent */
         for ( i=(2+GLOBAL_CFG_MAX_ENDPT);i<MAX_EXP_TABLE;i++ )
         {
            /* if there is an existing entry, overwrite the values */
            if ( !memcmp(lname, gExpTable[i].fqdn,strlen(lname)))
            {
               return gExpTable[i].expRTO;
            }
         }
         printf("EXPONENTIAL BACKOFF NAME %s NOT FOUND IN AP CMS TABLE.\n", name);
         return 0;
         break;
      default:
         printf("EXPONENTIAL BACKOFF BAD FLAG %d.\n", flag);
         return 0;
         break;
   }
}
