/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: udp.c
*    Creation Date: 29 June 1999 (v0.00)
*    VSS Info:
*        $Revision: 1.6 $
*        $Date: 2005/08/15 15:03:32 $
*
****************************************************************************
*    Description:
*
*      Provides a set of functions for UDP communication using sockets.
*
****************************************************************************/

#include <xbcmos.h>   /* OS specific definitions */
#include <net.h>
#include <netcfg.h>
//#include <cmmtaint.h>
#include <string.h>
#include "bcmosHelper.h"

/*
*****************************************************************************
** TYPEDEFS AND CONSTANTS
*****************************************************************************
*/

/*
*****************************************************************************
** LOCAL VARIABLES
*****************************************************************************
*/

/*
*****************************************************************************
** FUNCTION:   bcmNetUdpOpen
**
** PURPOSE:    Open UDP socket
**
** PARAMETERS: localPort  - local port of socket
**             bBlockMode - blocking or non-blocking mode of the socket
**             sd         - socket descriptor (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
#if CH_NEEDED
NET_STATUS bcmNetUdpOpen( UINT16 localPort, int bBlockMode, int *sd )
{
   OS_SOCKADDR    sin = {0};  /* socket address information */
   int            optval;
   OS_STATUS      status;

   status = bcmOsSocketCreate( OS_AF_INET, OS_SOCK_DGRAM, 0, sd );

   if (status != OS_STATUS_OK)
   {
      LOG(("ERROR: bcmNetUdpOpen() Failed to create socket.\n" ));
      return NET_STATUS_SOCKET_ERROR;
   }

   /* set socket options so that its addr and port can be re-used */
   optval = 1;
   status = bcmOsSocketSetOptions( *sd, OS_SOL_SOCKET, OS_SO_REUSEADDR, (char *)&optval, sizeof( optval ));
   if (status != OS_STATUS_OK)
   {
      LOG(( "WARNING: bcmNetUdpOpen() Failed to set SO_REUSEADDR option.\n" ));
   }

   if (!bBlockMode) /* Let's make the socket non-blocking. */
   {
      if (NET_STATUS_SOCKET_ERROR == *sd)
      {
         bcmOsSocketClose(*sd);
         return NET_STATUS_SOCKET_ERROR;
      }
   }

   sin.sin_family = OS_AF_INET;
   sin.sin_addr.S_addr = OS_INADDR_ANY;
   sin.sin_port = htons( localPort );

   status = bcmOsSocketBind( *sd, &sin, sizeof( sin ));
   if (status != OS_STATUS_OK)
   {
      bcmOsSocketClose( *sd );
      LOG(( "ERROR: bcmNetUdpOpen() Failed to bind the socket.\n" ));
      return NET_STATUS_SOCKET_ERROR;
   }

   return ( NET_STATUS_OK );
}
#endif
/*
*****************************************************************************
** FUNCTION:   bcmNetUdpRecv
**
** PURPOSE:    Receive UDP socket
**
** PARAMETERS: udpSock  - udp socket port
**             buf      - pointer to receive buffer
**             length   - length of receive buffer
**             rAddr    - pointer to remote udp socket address (in host-byte order)
**             rPort    - pointer to remote udp socket port (in host-byte order)
**             nRxBytes - number of received bytes
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpRecv( int             udpSock,
                          char            *buf,
                          int             length,
                          unsigned long   *raddr,
                          UINT16          *rport,
                          int             *nRxBytes )
{
    //#if (!NET_CFG_MULTI_IP_STACKS)
   long           ifno     = 0;
//#else
//   long           ifno     = cmGetIpStackIndex();
//#endif
   
    int            flags    = 0;
   OS_SOCKADDR    fromAddr = {0};
   int            fromLen  = sizeof( fromAddr );
   OS_STATUS      status;
   
   //printf("bcmNetUdpSend -no mulit ip stack trc-\n");

   /* Receive the packet */
   *nRxBytes = 0;
   status = bcmOsSocketReceiveFrom( udpSock,
                                    buf,
                                    length,
                                    flags,
                                    ifno,
                                    &fromAddr,
                                    &fromLen,
                                    nRxBytes );

   if ( status != OS_STATUS_OK )
   {
      LOG(( "ERROR: bcmNetUdpRecv(): error returned by bcmOsSocketReceiveFrom.\n" ));
      return ( NET_STATUS_SOCKET_ERROR );
   }

   *rport = ntohs( fromAddr.sin_port );
   *raddr = ntohl( fromAddr.sin_addr.S_addr );

   return ( NET_STATUS_OK );
}

/*
*****************************************************************************
** FUNCTION:   bcmNetUdpSend
**
** PURPOSE:    Send UDP socket
**
** PARAMETERS: udpSock  - udp socket port
**             buf      - pointer to transmit buffer
**             length   - length of transmit buffer
**             remAddr  - remote udp socket address (in host-byte order)
**             remPort  - remote udp socket port (in host-byte order)
**             nTxBytes - number of bytes transmitted (output)
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpSend( int             udpSock,
                          char            *buf,
                          int             length,
                          unsigned long   remAddr,
                          UINT16          remPort,
                          int             *nTxBytes )
{
   OS_STATUS            status;
   int                  flags = 0;
   OS_SOCKADDR          sin   = {0};
 //trc-
   //#if (!NET_CFG_MULTI_IP_STACKS)
   long                 ifno  = 0;
//#else
//   long                 ifno  = cmGetIpStackIndex();
//#endif

   //printf("bcmNetUdpSend -no mulit ip stack trc-\n");
   //printf("remAddr =0x<%x>\n",remAddr);
   //printf("ifno    = <%d>\n\n",ifno);

   sin.sin_family = OS_AF_INET;
   sin.sin_port = htons( remPort );
   sin.sin_addr.S_addr = htonl( remAddr );
 //DumpBuf("SENDING BUF", (unsigned char *) buf, length, 'x'); 
   status = bcmOsSocketSendTo( udpSock,
                               (char *)buf,
                               length,
                               flags,
                               ifno,
                               &sin,
                               sizeof(sin),
                               nTxBytes );

   if (( status != OS_STATUS_OK ) || ( *nTxBytes != length ))
   {
      LOG(( "ERROR: bcmNetUdpSend() - failed to send packet.\n" ));
      LOG(( "OS_STATUS: %i and nTxBytes: %i\n", status, nTxBytes ));
      return ( NET_STATUS_SOCKET_ERROR );
   }

   return ( NET_STATUS_OK );
}


/*
*****************************************************************************
** FUNCTION:   bcmNetUdpClose
**
** PURPOSE:    Close UDP socket
**
** PARAMETERS: sockDesc - udp socket description
**
** RETURNS:    NET_STATUS_OK or error code
**
*****************************************************************************
*/
NET_STATUS bcmNetUdpClose( int sockDesc )
{
   OS_STATUS status;

   status = bcmOsSocketClose(sockDesc);
   if (status != OS_STATUS_OK)
   {
      LOG(( "ERROR: bcmNetUdpClose()\n" ));
      return ( NET_STATUS_SOCKET_ERROR );
   }

   return ( NET_STATUS_OK );
}
