/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: seckrb.h
*    Creation Date: 27 Oct 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.3 $
*        $Date: 2004/01/05 21:00:43 $
*
****************************************************************************
*    Description:
*
*       Implementation of packetCable PKINIT reply processing
*
****************************************************************************/

/*
 *
 * Copyright 1994 by the Massachusetts Institute of Technology.
 * All Rights Reserved.
 *
 * Export of this software from the United States of America may
 *   require a specific license from the United States Government.
 *   It is the responsibility of any person or organization contemplating
 *   export to obtain such a license before exporting.
 *
 * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
 * distribute this software and its documentation for any purpose and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of M.I.T. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior
 * permission.  Furthermore if you modify this software you must label
 * your software as modified software and not distribute it in such a
 * fashion that it might be confused with the original M.I.T. software.
 * M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 */



# ifndef __SECKRB_H__
#define __SECKRB_H__

#include <stdlib.h>
#include <Errno.h>
//#include <memory.h>
#include "krbbuf.h"

/* clock skew related constants */
#define ONE_WEEK_IN_SECONDS   604800
#define FOUR_WEEK_IN_SECONDS  (4*604800)
#define MAX_ALLOWED_CLOCK_SKEW_IN_SECONDS    3600

/* Constant string seed for SNMPV3 key derivation function */
static const unsigned char SNMPv3_SEED[] = "SNMPv3 Keys";

/* Constant string seed for IPSec key derivation function */
static const unsigned char IPSEC_SEED[] = "IPsec Security Association";

/* maximum number of ciphersuites in AP req nd rep */
#define MAX_AP_CSUITES_LENGTH 5



/* #define FALSE 0 */
/* #define TRUE  1 */
/* #define SEC_MALLOC(x) malloc((x)) */
#define SEC_FREE(x) free((x))
/* #define SEC_CALLOC(x,y) calloc((x),(y)) */
/* #define SEC_REALLOC(x,y) realloc((x),(y)) */
/* to keep lint happy */
#define krb5_xfree(val) krbFreeBuf((char *)(val))

/* per Kerberos v5 protocol spec */
#define	ENCTYPE_NULL		0x0000
#define	ENCTYPE_DES_CBC_CRC	0x0001	/* DES cbc mode with CRC-32 */
#define	ENCTYPE_DES_CBC_MD4	0x0002	/* DES cbc mode with RSA-MD4 */
#define	ENCTYPE_DES_CBC_MD5	0x0003	/* DES cbc mode with RSA-MD5 */
#define	ENCTYPE_DES_CBC_RAW	0x0004	/* DES cbc mode raw */
/* XXX deprecated? */
#define	ENCTYPE_DES3_CBC_MD5	0x0005	/* DES cbc mode with RSA-MD5 */
/* #define	ENCTYPE_DES3_CBC_SHA	0x0005	*/ /* DES-3 cbc mode with NIST-SHA */
#define	ENCTYPE_DES3_CBC_RAW	0x0006	/* DES-3 cbc mode raw */
#define ENCTYPE_DES_HMAC_SHA1	0x0008
#define ENCTYPE_DES3_CBC_SHA1	0x0010
#define ENCTYPE_UNKNOWN		0x01ff
/* local crud */
/* marc's DES-3 with 32-bit length */
#define ENCTYPE_LOCAL_DES3_HMAC_SHA1 0x7007

#define	CKSUMTYPE_CRC32		0x0001
#define	CKSUMTYPE_RSA_MD4	0x0002
#define	CKSUMTYPE_RSA_MD4_DES	0x0003
#define	CKSUMTYPE_DESCBC	0x0004
/* des-mac-k */
/* rsa-md4-des-k */
#define	CKSUMTYPE_RSA_MD5	0x0007
#define	CKSUMTYPE_RSA_MD5_DES	0x0008
#define	CKSUMTYPE_RSA_MD5_DES3	0x0009
/* define CKSUMTYPE_NIST_SHA	0x0009 */
#define CKSUMTYPE_NIST_SHA	0x000d
#define CKSUMTYPE_HMAC_SHA1_DES3	0x000c




#define CLOCK_SKEW_ERROR 300    /* 5 minutes for krb/pkinit */


/* Protocol version number */
#define	KRB5_PVNO	5

/* Message types */

#define	KRB5_AS_REQ	((krb5_msgtype)10) /* Req for initial authentication */
#define	KRB5_AS_REP	((krb5_msgtype)11) /* Response to KRB_AS_REQ request */
#define	KRB5_TGS_REQ	((krb5_msgtype)12) /* TGS request to server */
#define	KRB5_TGS_REP	((krb5_msgtype)13) /* Response to KRB_TGS_REQ req */
#define	KRB5_AP_REQ	((krb5_msgtype)14) /* application request to server */
#define	KRB5_AP_REP	((krb5_msgtype)15) /* Response to KRB_AP_REQ_MUTUAL */
#define	KRB5_SAFE	((krb5_msgtype)20) /* Safe application message */
#define	KRB5_PRIV	((krb5_msgtype)21) /* Private application message */
#define	KRB5_CRED	((krb5_msgtype)22) /* Credential forwarding message */
#define	KRB5_ERROR	((krb5_msgtype)30) /* Error response */


#define	KDC_OPT_POSTDATED		0x02000000
#define	KDC_OPT_RENEWABLE		0x00800000



#define KV5M_NONE                                (-1760647424L)
#define KV5M_PRINCIPAL                           (-1760647423L)
#define KV5M_DATA                                (-1760647422L)
#define KV5M_KEYBLOCK                            (-1760647421L)
#define KV5M_CHECKSUM                            (-1760647420L)
#define KV5M_ENCRYPT_BLOCK                       (-1760647419L)
#define KV5M_ENC_DATA                            (-1760647418L)
#define KV5M_CRYPTOSYSTEM_ENTRY                  (-1760647417L)
#define KV5M_CS_TABLE_ENTRY                      (-1760647416L)
#define KV5M_CHECKSUM_ENTRY                      (-1760647415L)
#define KV5M_AUTHDATA                            (-1760647414L)
#define KV5M_TRANSITED                           (-1760647413L)
#define KV5M_ENC_TKT_PART                        (-1760647412L)
#define KV5M_TICKET                              (-1760647411L)
#define KV5M_AUTHENTICATOR                       (-1760647410L)
#define KV5M_TKT_AUTHENT                         (-1760647409L)
#define KV5M_CREDS                               (-1760647408L)
#define KV5M_LAST_REQ_ENTRY                      (-1760647407L)
#define KV5M_PA_DATA                             (-1760647406L)
#define KV5M_KDC_REQ                             (-1760647405L)
#define KV5M_ENC_KDC_REP_PART                    (-1760647404L)
#define KV5M_KDC_REP                             (-1760647403L)
#define KV5M_ERROR                               (-1760647402L)
#define KV5M_AP_REQ                              (-1760647401L)
#define KV5M_AP_REP                              (-1760647400L)
#define KV5M_AP_REP_ENC_PART                     (-1760647399L)
#define KV5M_RESPONSE                            (-1760647398L)
#define KV5M_SAFE                                (-1760647397L)
#define KV5M_PRIV                                (-1760647396L)
#define KV5M_PRIV_ENC_PART                       (-1760647395L)
#define KV5M_CRED                                (-1760647394L)
#define KV5M_CRED_INFO                           (-1760647393L)
#define KV5M_CRED_ENC_PART                       (-1760647392L)
#define KV5M_PWD_DATA                            (-1760647391L)
#define KV5M_ADDRESS                             (-1760647390L)
#define KV5M_KEYTAB_ENTRY                        (-1760647389L)
#define KV5M_CONTEXT                             (-1760647388L)
#define KV5M_OS_CONTEXT                          (-1760647387L)
#define KV5M_ALT_METHOD                          (-1760647386L)
#define KV5M_ETYPE_INFO_ENTRY                    (-1760647385L)
#define KV5M_DB_CONTEXT                          (-1760647384L)
#define KV5M_AUTH_CONTEXT                        (-1760647383L)
#define KV5M_KEYTAB                              (-1760647382L)
#define KV5M_RCACHE                              (-1760647381L)
#define KV5M_CCACHE                              (-1760647380L)
#define KV5M_PREAUTH_OPS                         (-1760647379L)
#define KV5M_SAM_CHALLENGE                       (-1760647378L)
#define KV5M_SAM_KEY                             (-1760647377L)
#define KV5M_ENC_SAM_RESPONSE_ENC                (-1760647376L)
#define KV5M_SAM_RESPONSE                        (-1760647375L)
#define KV5M_PREDICTED_SAM_RESPONSE              (-1760647374L)
#define KV5M_PASSWD_PHRASE_ELEMENT               (-1760647373L)
#define KV5M_GSS_OID                             (-1760647372L)
#define KV5M_GSS_QUEUE                           (-1760647371L)
#define ERROR_TABLE_BASE_kv5m (-1760647424L)

#define KRB5KDC_ERR_BAD_PVNO                     (-1765328381L)
#define KRB5_BADMSGTYPE                          (-1765328246L)
#define KRB5_BAD_KEYSIZE                         (-1765328195L)
#define KRB5_BAD_MSIZE                           (-1765328194L)
#define KRB5DES_BAD_KEYPAR                       (-1765328198L)
#define KRB5DES_WEAK_KEY                         (-1765328197L)
#define KRB5_BAD_ENCTYPE                         (-1765328196L)
#define KRB5_CRYPTO_INTERNAL                     (-1765328206L)
#define KRB5KRB_AP_ERR_BAD_INTEGRITY             (-1765328353L)
#define KRB5KRB_AP_ERR_INAPP_CKSUM               (-1765328334L)


#define	AP_OPTS_RESERVED		0x80000000
#define	AP_OPTS_USE_SESSION_KEY		0x40000000
#define	AP_OPTS_MUTUAL_REQUIRED		0x20000000
#define	AP_OPTS_FORWARD_CREDS           0x10000000
#define	AP_OPTS_FORWARDABLE_CREDS       0x08000000
#define AP_OPTS_USE_SUBKEY	0x00000001

#define AP_OPTS_WIRE_MASK	0xfffffff0


/* Ticket flags */
/* flags are 32 bits; each host is responsible to put the 4 bytes
   representing these bits into net order before transmission */
/* #define	TKT_FLG_RESERVED	0x80000000 */
#define	TKT_FLG_FORWARDABLE		0x40000000
#define	TKT_FLG_FORWARDED		0x20000000
#define	TKT_FLG_PROXIABLE		0x10000000
#define	TKT_FLG_PROXY			0x08000000
#define	TKT_FLG_MAY_POSTDATE		0x04000000
#define	TKT_FLG_POSTDATED		0x02000000
#define	TKT_FLG_INVALID			0x01000000
#define	TKT_FLG_RENEWABLE		0x00800000
#define	TKT_FLG_INITIAL			0x00400000
#define	TKT_FLG_PRE_AUTH		0x00200000
#define	TKT_FLG_HW_AUTH			0x00100000
#define	KDC_OPT_RENEWABLE_OK		0x00000010
#define	KDC_OPT_ENC_TKT_IN_SKEY		0x00000008
/* #define	KDC_OPT_UNUSED		0x00000004 */
#define	KDC_OPT_RENEW			0x00000002
#define	KDC_OPT_VALIDATE		0x00000001




/*
 * Per V5 spec on definition of principal types
 */

/* Name type not known */
#define KRB5_NT_UNKNOWN		0
/* Just the name of the principal as in DCE, or for users */
#define KRB5_NT_PRINCIPAL	1
/* Service and other unique instance (krbtgt) */
#define KRB5_NT_SRV_INST	2
/* Service with host name as instance (telnet, rcommands) */
#define KRB5_NT_SRV_HST		3
/* Service with host as remaining components */
#define KRB5_NT_SRV_XHST	4
/* Unique ID */
#define KRB5_NT_UID		5
/* X.500 name new for pkinit*/
#define KRB_NT_X500_PRINCIPAL 6

/* use UDP port number 1293 for AP exchange,
   and 88 for AS exchange */
#define AS_UDP_PORT_NUMBER 88
#define AP_UDP_PORT_NUMBER 1293

#ifndef krb5_roundup
/* round x up to nearest multiple of y */
#define krb5_roundup(x, y) ((((x) + (y) - 1)/(y))*(y))
#endif /* roundup */


/* ASN.1 encoding knowledge; KEEP IN SYNC WITH ASN.1 defs! */
/* here we use some knowledge of ASN.1 encodings */
/*
  Ticket is APPLICATION 1.
  Authenticator is APPLICATION 2.
  AS_REQ is APPLICATION 10.
  AS_REP is APPLICATION 11.
  TGS_REQ is APPLICATION 12.
  TGS_REP is APPLICATION 13.
  AP_REQ is APPLICATION 14.
  AP_REP is APPLICATION 15.
  KRB_SAFE is APPLICATION 20.
  KRB_PRIV is APPLICATION 21.
  KRB_CRED is APPLICATION 22.
  EncASRepPart is APPLICATION 25.
  EncTGSRepPart is APPLICATION 26.
  EncAPRepPart is APPLICATION 27.
  EncKrbPrivPart is APPLICATION 28.
  EnckredPart is APPLICATION 29.
  KRB_ERROR is APPLICATION 30.
 */
/* allow either constructed or primitive encoding, so check for bit 6
   set or reset */
#define krb5_is_krb_ticket(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x61 ||\
                               (dat)->data[0] == 0x41))
#define krb5_is_krb_authenticator(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x62 ||\
                               (dat)->data[0] == 0x42))
#define krb5_is_as_req(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6a ||\
                               (dat)->data[0] == 0x4a))
#define krb5_is_as_rep(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6b ||\
                               (dat)->data[0] == 0x4b))
#define krb5_is_tgs_req(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6c ||\
                               (dat)->data[0] == 0x4c))
#define krb5_is_tgs_rep(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6d ||\
                               (dat)->data[0] == 0x4d))
#define krb5_is_ap_req(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6e ||\
                               (dat)->data[0] == 0x4e))
#define krb5_is_ap_rep(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x6f ||\
                               (dat)->data[0] == 0x4f))
#define krb5_is_krb_safe(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x74 ||\
                               (dat)->data[0] == 0x54))
#define krb5_is_krb_priv(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x75 ||\
                               (dat)->data[0] == 0x55))
#define krb5_is_krb_cred(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x76 ||\
                               (dat)->data[0] == 0x56))
#define krb5_is_krb_enc_as_rep_part(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x79 ||\
                               (dat)->data[0] == 0x59))
#define krb5_is_krb_enc_tgs_rep_part(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x7a ||\
                               (dat)->data[0] == 0x5a))
#define krb5_is_krb_enc_ap_rep_part(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x7b ||\
                               (dat)->data[0] == 0x5b))
#define krb5_is_krb_enc_krb_priv_part(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x7c ||\
                               (dat)->data[0] == 0x5c))
#define krb5_is_krb_enc_krb_cred_part(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x7d ||\
                               (dat)->data[0] == 0x5d))
#define krb5_is_krb_error(dat)\
   ((dat) && (dat)->length && ((dat)->data[0] == 0x7e ||\
                               (dat)->data[0] == 0x5e))


#define krb5_princ_realm(princ) (&(princ)->realm)
#define krb5_princ_set_realm(princ,value) ((princ)->realm = *(value))
#define krb5_princ_set_realm_length(princ,value) (princ)->realm.length = (value)
#define krb5_princ_set_realm_data(princ,value) (princ)->realm.data = (value)
#define	krb5_princ_size(princ) (princ)->length
#define	krb5_princ_type(princ) (princ)->type
#define	krb5_princ_name(princ) (princ)->data
#define	krb5_princ_component(princ,i) ((princ)->data + i)

#define KRB5_KEYUSAGE_AS_REQ_PA_ENC_TS		1
#define KRB5_KEYUSAGE_KDC_REP_TICKET		2
#define KRB5_KEYUSAGE_AS_REP_ENCPART		3
#define KRB5_KEYUSAGE_TGS_REQ_AD_SESSKEY	4
#define KRB5_KEYUSAGE_TGS_REQ_AD_SUBKEY		5
#define KRB5_KEYUSAGE_TGS_REQ_AUTH_CKSUM	6
#define KRB5_KEYUSAGE_TGS_REQ_AUTH		7
#define KRB5_KEYUSAGE_TGS_REP_ENCPART_SESSKEY	8
#define KRB5_KEYUSAGE_TGS_REP_ENCPART_SUBKEY	9
#define KRB5_KEYUSAGE_AP_REQ_AUTH_CKSUM		10
#define KRB5_KEYUSAGE_AP_REQ_AUTH		11
#define KRB5_KEYUSAGE_AP_REP_ENCPART		12
#define KRB5_KEYUSAGE_KRB_PRIV_ENCPART		13
#define KRB5_KEYUSAGE_KRB_CRED_ENCPART		14
#define KRB5_KEYUSAGE_KRB_SAFE_CKSUM		15
#define KRB5_KEYUSAGE_APP_DATA_ENCRYPT		16
#define KRB5_KEYUSAGE_APP_DATA_CKSUM		17
#define KRB5_KEYUSAGE_KRB_ERROR_CKSUM		18
#define KRB5_KEYUSAGE_AD_KDCISSUED_CKSUM	19
#define KRB5_KEYUSAGE_AD_MTE			20
#define KRB5_KEYUSAGE_AD_ITE			21

/* XXX need to register these */

#define KRB5_KEYUSAGE_GSS_TOK_MIC		22
#define KRB5_KEYUSAGE_GSS_TOK_WRAP_INTEG	23
#define KRB5_KEYUSAGE_GSS_TOK_WRAP_PRIV		24


#define KRB5_PADATA_ENC_TIMESTAMP	2
#define	KRB5_PADATA_PW_SALT		3
#define KRB5_PADATA_ETYPE_INFO		11 /* Etype info for preauth */
#define KRB5_PREAUTH_BAD_TYPE                    (-1765328176L)

#define KRB5_PADATA_PA_PK_AS_REQ 14  /* pkinit msg type*/

#define KRB5_TD_APP_DEFINED_ERROR   106
#define KRB5_TD_REQ_NONCE           107
#define KRB5_TD_REQ_SEQ             108

/* per Kerberos v5 protocol spec */
#define	ADDRTYPE_INET		0x0002
#define	ADDRTYPE_CHAOS		0x0005
#define	ADDRTYPE_XNS		0x0006
#define	ADDRTYPE_ISO		0x0007
#define ADDRTYPE_DDP		0x0010
/* not yet in the spec... */
#define ADDRTYPE_ADDRPORT	0x0100
#define ADDRTYPE_IPPORT		0x0101



typedef const krb5_const;
typedef	void* krb5_pointer;
typedef void const *krb5_const_pointer;

typedef	signed long	krb5_int32;
typedef	unsigned char	krb5_octet;

typedef	unsigned int krb5_boolean;
typedef	unsigned int krb5_msgtype;
typedef	unsigned int krb5_kvno;

typedef	krb5_int32 krb5_addrtype;
typedef krb5_int32 krb5_enctype;
typedef krb5_int32 krb5_cksumtype;
typedef krb5_int32 krb5_authdatatype;
typedef krb5_int32 krb5_keyusage;

typedef krb5_int32	krb5_preauthtype; /* This may change, later on */
typedef	krb5_int32	krb5_flags;
typedef krb5_int32	krb5_timestamp;
typedef	krb5_int32	krb5_error_code;
typedef krb5_int32	krb5_deltat;

/* Assuming sizeof int is 2. */
typedef	int	krb5_int16;
typedef	unsigned int	krb5_ui_2;
typedef	unsigned long	krb5_ui_4;



typedef krb5_error_code	krb5_magic;

typedef struct _krb5_data
{
	krb5_magic magic;
	int length;
	char *data;
} krb5_data;

typedef struct krb5_principal_data
{
    krb5_magic magic;
    krb5_data realm;
    krb5_data *data;		/* An array of strings */
    krb5_int32 length;
    krb5_int32 type;
} krb5_principal_data;

typedef	krb5_principal_data *krb5_principal;
typedef const krb5_principal_data *krb5_const_principal;

typedef struct
{
   krb5_int32          priv_key_len;  /* private key, generated localy */
   krb5_octet          *priv_key;     /* private key, generated localy */
} priv_key_t;


typedef struct _krb5_oid
{
   int length;
   unsigned long oid_array[16];      /* OID */
} krb5_oid;

typedef struct _Parameter
{
   krb5_int32 type;
} Parameter;

typedef struct _digest_alg_identifier
{
   krb5_oid    alg_oid;
   Parameter   parameter;
} digest_alg_identifier;


typedef struct _digest_alg_identifiers
{
   /* This is a SET OF according to CMS, but we only have 1 el */
   krb5_int32              length;     /* length of the set */
   digest_alg_identifier   digest_alg_id;
} digest_alg_identifiers;


typedef struct _krb5_checksum
{
    krb5_magic magic;
    krb5_cksumtype checksum_type;	/* checksum type */
    int length;
    krb5_octet *contents;
} krb5_checksum;


typedef struct _PKAuthenticator
{
   krb5_int32       cu_sec;             /* tag [0] */
   krb5_timestamp   c_time;             /* tag [1] */
   krb5_int32       nonce;              /* tag [2] */
   krb5_checksum    pa_checksum;        /* tag [3] */
} PKAuthenticator;

typedef struct _DHParameter
{
   krb5_int32     prime_len;
   krb5_octet     *prime;        /* oakley group 1 768 bit prime */
   krb5_int32     base;              /* integer 2  for Oakley group 1*/
   krb5_int32     private_len;       /* OPTIONAL */
   krb5_int32     q_len;
   krb5_octet     *q;
} DHParameter;

typedef struct _subject_algorithm
{
   krb5_oid      dhkey_oid;
   DHParameter   dh_parameter;
} subject_algorithm;

typedef struct _subj_pub_key_info
{
   subject_algorithm     subj_alg;
   krb5_int32            subj_pub_key_len;
   krb5_octet            *subj_pub_key;      /* DH public exponent BITSTRING */
} subj_pub_key_info;

typedef struct _auth_pack
{
   PKAuthenticator     pk_authenticator;     /* tag [0] */
   subj_pub_key_info   client_pub_value;     /* tag [1]diffie hellman values */
} auth_pack;

typedef struct _tgs_pub_key_info
{
   krb5_int32       tgs_pub_key_len;
   krb5_octet       *tgs_pub_key;     /* tag [2]diffie hellman public exp */
} tgs_pub_key_info;

typedef struct _auth_pack_reply
{
   krb5_int32         nonce;             /* tag [0] */
   tgs_pub_key_info   tgs_pub_value;     /* tag [2]diffie hellman values */
} auth_pack_reply;


typedef struct _encapsulated_content_info
{
   krb5_oid         econtent_type;
   auth_pack        econtent;          /* tag [0] Explicit */
   int              length;
   krb5_octet       *contents;         /* octet string encoded auth_pack */
} encapsulated_content_info;

typedef struct _encapsulated_content_info_reply
{
   krb5_oid               econtent_type;
   auth_pack_reply        econtent;        /* tag [0] Explicit */
   int                    length;
   krb5_octet             *contents;       /* octet string encoded auth_pack */
} encapsulated_content_info_reply;


typedef digest_alg_identifier signature_alg_identifier;

typedef struct _issuer_serialnumber
{
   krb5_int32   signer_name_len;
   krb5_octet   *signer_name;
   krb5_octet   *signer_sernum;
   //krb5_int32   signer_sernum;
   krb5_int32   signer_sernum_len;
} issuer_serialnumber;

typedef struct _signer_identifier
{
   issuer_serialnumber    issuer_sernum;
} signer_identifier;

typedef struct _signature_t
{
   krb5_int32     signature_len;
   krb5_octet      *signature_val;
} signature_t;

typedef struct _signer_info_t
{
   krb5_int32                 cms_version;
   signer_identifier          sid;
   digest_alg_identifier      digest_alg_id;
   char                       *signed_attrs;
   signature_alg_identifier   signature_alg_id;
   signature_t                signature;
   char                       *unsigned_attrs;
} signer_info_t;

typedef struct _signer_infos_t
{
   krb5_int32              length;     /* length of the set is 1 */
   signer_info_t           signer_info;
} signer_infos_t;

typedef struct _cert_t
{
   krb5_int32    cert_len;
   krb5_octet    *cert;
} cert_t;

typedef struct _cert_set
{
   krb5_int32    length;        /* number of elements in the set */
   cert_t        certs[5];      /* assuming maximum of 5 */
} cert_set;

typedef struct _pkinit_signed_data
{
   krb5_int32                  cms_version;
   digest_alg_identifiers      digest_alg_ids;
   encapsulated_content_info   encap_content_info;
   cert_set                    certificates;
   signer_infos_t              signer_infos;
} pkinit_signed_data;

typedef struct _pkinit_signed_authpack
{
   krb5_oid             content_type;     /* no tag */
   pkinit_signed_data   signed_data;      /* [0] tag 0 */
} pkinit_signed_authpack;


typedef struct _pkinit_signed_data_reply
{
   krb5_int32                        cms_version;
   digest_alg_identifiers            digest_alg_ids;
   encapsulated_content_info_reply   encap_content_info;
   cert_set                          certificates;
   signer_infos_t                    signer_infos;
} pkinit_signed_data_reply;


typedef struct _pkinit_request
{
   pkinit_signed_authpack   signed_authpack;
} pkinit_request;

typedef struct _pkinit_dh_signed_data
{
   krb5_oid                   content_type;     /* no tag */
   pkinit_signed_data_reply   signed_data;      /* [0] tag 0 */
} pkinit_dh_signed_data;

typedef struct _pkinit_reply
{
   pkinit_dh_signed_data     dh_signed_data;
   char                      *enc_key_pack;
} pkinit_reply;



/* pre-authentication data */
typedef struct _krb5_pa_data
{
    krb5_magic         magic;
    krb5_preauthtype   pa_type;
    int                length;
    krb5_octet         *contents;
} krb5_pa_data;

/* typed data */
typedef struct _krb5_typed_data
{
    krb5_magic         magic;
    krb5_int32         data_type;
	krb5_int32         req_seq_val;   /* sequence number in case of an authenticator */
    int                length;
    krb5_octet         *contents;
} krb5_typed_data;

typedef struct _app_specific_data
{
   krb5_oid         app_oid;
   int              length;
   krb5_octet       *contents;
} app_specific_data;

typedef struct _app_specific_data_pktc_krb_error
{
   krb5_int32      e_code;
   krb5_data       *e_text;
   krb5_data       *pktc_edata;
} app_specific_data_pktc_krb_error;


/* structure for address */
typedef struct _krb5_address
{
    krb5_magic magic;
    krb5_addrtype addrtype;
    int length;
    krb5_octet *contents;
} krb5_address;

typedef struct _krb5_enc_data
{
    krb5_magic magic;
    krb5_enctype enctype;
    krb5_kvno kvno;
    krb5_data ciphertext;
} krb5_enc_data;

/* structure for auth data */
typedef struct _krb5_authdata
{
    krb5_magic magic;
    krb5_authdatatype ad_type;
    int length;
    krb5_octet *contents;
} krb5_authdata;

typedef struct _krb5_keyblock
{
    krb5_magic magic;
    krb5_enctype enctype;
    int length;
    krb5_octet *contents;
} krb5_keyblock;

/* structure for transited encoding */
typedef struct _krb5_transited
{
    krb5_magic magic;
    krb5_octet tr_type;
    krb5_data tr_contents;
} krb5_transited;

/* Time set */
typedef struct _krb5_ticket_times {
    krb5_timestamp authtime; /* XXX ? should ktime in KDC_REP == authtime
				in ticket? otherwise client can't get this */
    krb5_timestamp starttime;		/* optional in ticket, if not present,
					   use authtime */
    krb5_timestamp endtime;
    krb5_timestamp renew_till;
} krb5_ticket_times;

typedef struct _krb5_enc_tkt_part
{
    krb5_magic magic;
    /* to-be-encrypted portion */
    krb5_flags flags;			   /* flags */
    krb5_keyblock *session;		/* session key: includes enctype */
    krb5_principal client;		   /* client name/realm */
    krb5_transited transited;	   /* list of transited realms */
    krb5_ticket_times times;	   /* auth, start, end, renew_till */
    krb5_address  **caddrs;	   /* array of ptrs to addresses */
    krb5_authdata  **authorization_data; /* auth data */
} krb5_enc_tkt_part;


typedef struct _krb5_ticket
{
    krb5_magic magic;
    /* cleartext portion */
    krb5_principal server;		     /* server name/realm */
    krb5_enc_data enc_part;		  /* encryption type,kvno,encrypted encoding */
    krb5_enc_tkt_part *enc_part2;  /* ptr to decrypted version, if available */
} krb5_ticket;


typedef struct _krb5_safe
{
    krb5_magic magic;
    krb5_data user_data;		   /* user data */
    krb5_timestamp timestamp;		/* client time, optional */
    krb5_int32 usec;			      /* microsecond portion of time,
					   optional */
    krb5_int32 seq_number;		   /* sequence #, optional */
    krb5_address *s_address;	   /* sender address */
    krb5_address *r_address;	   /* recipient address, optional */
    krb5_checksum *checksum;	   /* data integrity checksum */
} krb5_safe;


/* Last request fields */
typedef struct _krb5_last_req_entry
{
    krb5_magic magic;
    krb5_octet lr_type;
    krb5_timestamp value;
} krb5_last_req_entry;

typedef struct _krb5_enc_kdc_rep_part
{
    krb5_magic magic;
    /* encrypted part: */
    krb5_msgtype msg_type;		         /* krb5 message type */
    krb5_keyblock *session;		      /* session key */
    krb5_last_req_entry  **last_req;   /* array of ptrs to entries */
    krb5_int32 nonce;			         /* nonce from request */
    krb5_timestamp key_exp;		      /* expiration date */
    krb5_flags flags;			         /* ticket flags */
    krb5_ticket_times times;		      /* lifetime info */
    krb5_principal server;		         /* server's principal identifier */
    krb5_address **caddrs;	            /* array of ptrs to addresses,optional */
} krb5_enc_kdc_rep_part;


typedef struct _krb5_kdc_rep
{
    krb5_magic magic;
    /* cleartext part: */
    krb5_msgtype msg_type;		/* AS_REP or KDC_REP? */
    krb5_pa_data **padata;	   /* preauthentication data from KDC */
    krb5_principal client;		/* client's principal identifier */
    krb5_ticket *ticket;		/* ticket */
    krb5_enc_data enc_part;	/* encryption type, kvno, encrypted encoding */
    krb5_enc_kdc_rep_part *enc_part2;    /* unencrypted version, if available */
} krb5_kdc_rep;


typedef struct _krb5_cred_info
{
    krb5_magic magic;
    krb5_keyblock *session;		 /* session key used to encrypt */
					/* ticket */
    krb5_principal client;		   /* client name/realm, optional */
    krb5_principal server;		   /* server name/realm, optional */
    krb5_flags flags;			   /* ticket flags, optional */
    krb5_ticket_times times;		/* auth, start, end, renew_till, */
					/* optional */
    krb5_address **caddrs;	      /* array of ptrs to addresses */
} krb5_cred_info;


/* Sandia password generation structures */
typedef struct _passwd_phrase_element
{
    krb5_magic magic;
    krb5_data *passwd;
    krb5_data *phrase;
} passwd_phrase_element;


/*
 * A null-terminated array of this structure is returned by the KDC as
 * the data part of the ETYPE_INFO preauth type.  It informs the
 * client which encryption types are supported.
 */
typedef struct _krb5_etype_info_entry
{
	krb5_magic	magic;
	krb5_enctype	etype;
	int		length;
	krb5_octet	*salt;
} krb5_etype_info_entry;

typedef krb5_etype_info_entry ** krb5_etype_info;

typedef struct _krb5_sam_challenge
{
	krb5_magic	magic;
	krb5_int32	sam_type;         /* information */
	krb5_flags	sam_flags;        /* KRB5_SAM_* values */
	krb5_data	sam_type_name;
	krb5_data	sam_track_id;
	krb5_data	sam_challenge_label;
	krb5_data	sam_challenge;
	krb5_data	sam_response_prompt;
	krb5_data	sam_pk_for_sad;
	krb5_int32	sam_nonce;
	krb5_checksum	sam_cksum;
} krb5_sam_challenge;


typedef struct _krb5_sam_key
{	/* reserved for future use */
	krb5_magic	magic;
	krb5_keyblock	sam_key;
} krb5_sam_key;


typedef struct _krb5_enc_sam_response_enc
{
	krb5_magic	magic;
	krb5_int32	sam_nonce;
	krb5_timestamp	sam_timestamp;
	krb5_int32	sam_usec;
	krb5_data	sam_passcode;
} krb5_enc_sam_response_enc;

typedef struct _krb5_sam_response
{
	krb5_magic	magic;
	krb5_int32	sam_type;                /* informational */
	krb5_flags	sam_flags;               /* KRB5_SAM_* values */
	krb5_data	sam_track_id;            /* copied */
	krb5_enc_data	sam_enc_key;          /* krb5_sam_key - future use */
	krb5_enc_data	sam_enc_nonce_or_ts;  /* krb5_enc_sam_response_enc */
	krb5_int32	sam_nonce;
	krb5_timestamp	sam_patimestamp;
} krb5_sam_response;


typedef struct _krb5_predicted_sam_response {
	krb5_magic	magic;
	krb5_keyblock	sam_key;
} krb5_predicted_sam_response;


/*
 * This structure is returned in the e-data field of the KRB-ERROR
 * message when the error calling for an alternative form of
 * authentication is returned, KRB_AP_METHOD.
 */
typedef struct _krb5_alt_method {
	krb5_magic	magic;
	krb5_int32	method;
	int		length;
	krb5_octet	*data;
} krb5_alt_method;

/* the unencrypted version */
typedef struct _krb5_authenticator
{
    krb5_magic magic;
    krb5_principal client;		/* client name/realm */
    krb5_checksum *checksum;	/* checksum, includes type, optional */
    krb5_int32 cusec;			/* client usec portion */
    krb5_timestamp ctime;		/* client sec portion */
    krb5_keyblock *subkey;		/* true session key, optional */
    krb5_int32 seq_number;		/* sequence #, optional */
    krb5_authdata **authorization_data; /* New add by Ari, auth data,optional*/
} krb5_authenticator;



typedef struct _krb5_ap_req
{
    krb5_magic magic;
    krb5_flags ap_options;		    /* requested options */
    krb5_ticket *ticket;		    /* ticket */
    krb5_enc_data authenticator;	 /* authenticator (already encrypted) */
} krb5_ap_req;


typedef struct _krb5_ap_rep
{
    krb5_magic magic;
    krb5_enc_data enc_part;
} krb5_ap_rep;


typedef struct _krb5_ap_rep_enc_part
{
    krb5_magic magic;
    krb5_timestamp ctime;		/* client time, seconds portion */
    krb5_int32 cusec;			/* client time, microseconds portion */
    krb5_keyblock *subkey;		/* true session key, optional */
    krb5_int32 seq_number;		/* sequence #, optional */
} krb5_ap_rep_enc_part;


typedef struct _krb5_priv
{
    krb5_magic magic;
    krb5_enc_data enc_part;		/* encrypted part */
} krb5_priv;


typedef struct _krb5_priv_enc_part
{
    krb5_magic magic;
    krb5_data user_data;		    /* user data */
    krb5_timestamp timestamp;		 /* client time, optional */
    krb5_int32 usec;			       /* microsecond portion of time, opt. */
    krb5_int32 seq_number;		    /* sequence #, optional */
    krb5_address *s_address;	    /* sender address */
    krb5_address *r_address;	    /* recipient address, optional */
} krb5_priv_enc_part;

typedef struct _krb5_cred_enc_part
{
    krb5_magic magic;
    krb5_int32 nonce;			    /* nonce, optional */
    krb5_timestamp timestamp;		 /* client time */
    krb5_int32 usec;			       /* microsecond portion of time */
    krb5_address *s_address;	    /* sender address, optional */
    krb5_address *r_address;	    /* recipient address, optional */
    krb5_cred_info **ticket_info;
} krb5_cred_enc_part;


typedef struct _krb5_cred
{
    krb5_magic magic;
    krb5_ticket  **tickets;	       /* tickets */
    krb5_enc_data enc_part;		    /* encrypted part */
    krb5_cred_enc_part *enc_part2;	 /* unencrypted version, if available*/
} krb5_cred;


/* error message structure */
typedef struct _krb5_error
{
    krb5_magic magic;
    /* some of these may be meaningless in certain contexts */
    krb5_timestamp ctime;		   /* client sec portion; optional */
    krb5_int32 cusec;			   /* client usec portion; optional */
    krb5_int32 susec;			   /* server usec portion */
    krb5_timestamp stime;		   /* server sec portion */
    krb5_ui_4 error;			      /* error code (protocol error #'s) */
    krb5_principal client;		   /* client's principal identifier;
					   optional */
    krb5_principal server;		   /* server's principal identifier */
    krb5_data text;			      /* descriptive text */
    krb5_data e_data;	         /* additional error-describing data */
    krb5_checksum e_cksum;       /* checksum */
} krb5_error;

typedef struct _krb5_pa_enc_ts
{
    krb5_timestamp	patimestamp;
    krb5_int32		pausec;
} krb5_pa_enc_ts;


typedef struct _krb5_pwd_data
{
    krb5_magic magic;
    int sequence_count;
    passwd_phrase_element  **element;
} krb5_pwd_data;


/* AS request message structure */
typedef struct _krb5_kdc_req
{
    krb5_magic magic;
    krb5_msgtype msg_type;		     /* AS_REQ or TGS_REQ? */
    krb5_pa_data **padata;	        /* e.g. encoded AP_REQ */
    /* real body */
    krb5_flags kdc_options;	     /* requested options */
    krb5_principal client;		     /* includes realm; optional */
    krb5_principal server;		     /* includes realm (only used if no client) */
    krb5_timestamp from;		     /* requested starttime */
    krb5_timestamp till;		     /* requested endtime */
    krb5_int32 nonce;			     /* nonce to match request/response */
    int nktypes;			           /* # of ktypes, must be positive */
    krb5_enctype *ktype;		     /* requested enctype(s) */
    /* following are not included in the as_req */
    krb5_address **addresses;	         /* requested addresses, optional */
    krb5_timestamp rtime;		         /* (optional) requested renew_till */
    krb5_enc_data authorization_data;	/* encrypted auth data; OPTIONAL */
    krb5_authdata **unenc_authdata;    /* unencrypted auth data, if available */
    krb5_ticket **second_ticket;       /* second ticket array; OPTIONAL */
} krb5_kdc_req;


/* credentials:	 Ticket, session key, etc. */
typedef struct _krb5_creds
{
    krb5_magic magic;
    krb5_principal client;		   /* client's principal identifier */
    krb5_principal server;		   /* server's principal identifier */
    krb5_keyblock keyblock;		/* session encryption key info */
    krb5_ticket_times times;		/* lifetime info */
    krb5_data ticket;			   /* ticket string itself */
    krb5_authdata **authdata;	   /* authorization data */
} krb5_creds;




/*********************************************************************/


/* new encryption provider api */
typedef struct krb5_enc_provider {
    void (*block_size) (size_t *output);

    /* keybytes is the input size to make_key;
       keylength is the output size */
    void (*keysize) (size_t *keybytes, size_t *keylength);

    /* ivec == 0 is an all-zeros ivec */
    krb5_error_code (*encrypt)
    (const krb5_keyblock *key, const krb5_data *ivec,
      const krb5_data *input, krb5_data *output);

    krb5_error_code (*decrypt)
    (const krb5_keyblock *key, const krb5_data *ivec,
      const krb5_data *input, krb5_data *output);

    krb5_error_code (*make_key)
    (const krb5_data *randombits, krb5_keyblock *key);
} krb5_enc_provider;

typedef struct krb5_hash_provider {
    void (*hash_size)
    (size_t *output);

    void (*block_size)
    (size_t *output);

    /* this takes multiple inputs to avoid lots of copying. */
    krb5_error_code (*hash)
    (unsigned int icount, const krb5_data *input, krb5_data *output);
} krb5_hash_provider;

typedef struct krb5_keyhash_provider {
    void (*hash_size) (size_t *output);

    krb5_error_code (*hash)
    (const krb5_keyblock *key, const krb5_data *ivec,
      const krb5_data *input, krb5_data *output);

    krb5_error_code (*verify)
    (const krb5_keyblock *key, const krb5_data *ivec,
      const krb5_data *input, const krb5_data *hash,
      krb5_boolean *valid);
} krb5_keyhash_provider;

typedef void (*krb5_encrypt_length_func)
(const struct krb5_enc_provider *enc,
  const struct krb5_hash_provider *hash,
  size_t inputlen, size_t *length);

typedef krb5_error_code (*krb5_crypt_func)
(const struct krb5_enc_provider *enc,
  const struct krb5_hash_provider *hash,
  const krb5_keyblock *key, krb5_keyusage usage,
  const krb5_data *ivec,
  const krb5_data *input, krb5_data *output);

typedef krb5_error_code (*krb5_str2key_func)
(const struct krb5_enc_provider *enc, const krb5_data *string,
  const krb5_data *salt, krb5_keyblock *key);

typedef struct krb5_keytypes {
    krb5_enctype etype;
    char *in_string;
    char *out_string;
    const struct krb5_enc_provider *enc;
    const struct krb5_hash_provider *hash;
    krb5_encrypt_length_func encrypt_len;
    krb5_crypt_func encrypt;
    krb5_crypt_func decrypt;
    krb5_str2key_func str2key;
} krb5_keytypes;

typedef struct krb5_cksumtypes {
    krb5_cksumtype ctype;
    unsigned int flags;
    char *in_string;
    char *out_string;
    /* if the hash is keyed, this is the etype it is keyed with.
       Actually, it can be keyed by any etype which has the same
       enc_provider as the specified etype.  DERIVE checksums can
       be keyed with any valid etype. */
    krb5_enctype keyed_etype;
    /* I can't statically initialize a union, so I'm just going to use
       two pointers here.  The keyhash is used if non-NULL.  If NULL,
       then HMAC/hash with derived keys is used if the relevant flag
       is set.  Otherwise, a non-keyed hash is computed.  This is all
       kind of messy, but so is the krb5 api. */
    const struct krb5_keyhash_provider *keyhash;
    const struct krb5_hash_provider *hash;
} krb5_cksumtypes;

#define KRB5_CKSUMFLAG_DERIVE		0x0001
#define KRB5_CKSUMFLAG_NOT_COLL_PROOF	0x0002

/*
 * in here to deal with stuff from lib/crypto
 */

void krb5_nfold
(int inbits, const unsigned char *in,
		int outbits, unsigned char *out);

krb5_error_code krb5_hmac
(const struct krb5_hash_provider *hash,
		const krb5_keyblock *key, unsigned int icount,
		const krb5_data *input, krb5_data *output);


/* old provider api */
/*
 * begin "encryption.h"
 */

typedef struct _krb5_encrypt_block {
    krb5_magic magic;
    krb5_enctype crypto_entry;		/* to call krb5_encrypt_size, you need
					   this.  it was a pointer, but it
					   doesn't have to be.  gross. */
    krb5_keyblock *key;
} krb5_encrypt_block;



typedef struct _krb5_cryptosystem_entry {
    krb5_magic magic;
    krb5_error_code (*encrypt_func) ( krb5_const_pointer /* in */,
					       krb5_pointer /* out */,
					       const size_t,
					       krb5_encrypt_block *,
					       krb5_pointer);
    krb5_error_code (*decrypt_func) ( krb5_const_pointer /* in */,
					       krb5_pointer /* out */,
					       const size_t,
					       krb5_encrypt_block *,
					       krb5_pointer);
    krb5_error_code (*process_key) ( krb5_encrypt_block *,
					      const krb5_keyblock *);
    krb5_error_code (*finish_key) ( krb5_encrypt_block *);
    krb5_error_code (*string_to_key) (const krb5_encrypt_block *,
						krb5_keyblock *,
						const krb5_data *,
						const krb5_data *);
    krb5_error_code (*init_random_key) ( const krb5_encrypt_block *,
						const krb5_keyblock *,
						krb5_pointer *);
    krb5_error_code (*finish_random_key) ( const krb5_encrypt_block *,
						krb5_pointer *);
    krb5_error_code (*random_key) ( const krb5_encrypt_block *,
					      krb5_pointer,
					      krb5_keyblock **);
    int block_length;
    int pad_minimum;			/* needed for cksum size computation */
    int keysize;
    krb5_enctype proto_enctype;		/* key type,
					   (assigned protocol number AND
					    table index) */
} krb5_cryptosystem_entry;

typedef struct _krb5_cs_table_entry {
    krb5_magic magic;
    krb5_cryptosystem_entry * system;
    krb5_pointer random_sequence;	/* from init_random_key() */
} krb5_cs_table_entry;


/* could be used in a table to find a sumtype */
typedef krb5_error_code
	(*SUM_FUNC) (
		const krb5_pointer /* in */,
		const size_t /* in_length */,
		const krb5_pointer /* key/seed */,
		const size_t /* key/seed size */,
		krb5_checksum * /* out_cksum */);

typedef krb5_error_code
	(*SUM_VERF_FUNC) (
		const krb5_checksum * /* out_cksum */,
		const krb5_pointer /* in */,
		const size_t /* in_length */,
		const krb5_pointer /* key/seed */,
		const size_t /* key/seed size */);

typedef struct _krb5_checksum_entry {
    krb5_magic magic;
    SUM_FUNC sum_func;			/* Checksum generator */
    SUM_VERF_FUNC sum_verf_func;	/* Verifier of checksum */
    int checksum_length;	   	/* length returned by sum_func */
    unsigned int is_collision_proof:1;
    unsigned int uses_key:1;
} krb5_checksum_entry;

krb5_error_code krb5_crypto_os_localaddr
	(krb5_address ***);



typedef krb5_error_code (*krb5_preauth_obtain_proc)
   ( krb5_pa_data *,
     krb5_etype_info,
     krb5_keyblock *,
     krb5_error_code ( * )( const krb5_enctype,
        krb5_data *,
        krb5_const_pointer,
        krb5_keyblock **),
     krb5_const_pointer,
     krb5_creds *,
     krb5_kdc_req *,
     krb5_pa_data **);

typedef krb5_error_code (*krb5_preauth_process_proc)
    ( krb5_pa_data *,
		    krb5_kdc_req *,
		    krb5_kdc_rep *,
		    krb5_error_code ( * )( const krb5_enctype,
					  krb5_data *,
					  krb5_const_pointer,
					  krb5_keyblock **),
		    krb5_const_pointer,
		    krb5_error_code ( * )( const krb5_keyblock *,
					  krb5_const_pointer,
					  krb5_kdc_rep * ),
		    krb5_keyblock **,
		    krb5_creds *,
		    krb5_int32 *,
		    krb5_int32 *);


    krb5_error_code krb5_encrypt_helper (const krb5_keyblock *key,
          krb5_keyusage usage, const krb5_data *plain,
          krb5_enc_data *cipher);

    typedef struct _krb5_preauth_ops {
      krb5_magic magic;
      int     type;
      int	flags;
      krb5_preauth_obtain_proc	obtain;
      krb5_preauth_process_proc	process;
   } krb5_preauth_ops;

   krb5_error_code krb5_obtain_padata
   (krb5_pa_data **,
    krb5_error_code ( * )(
       const krb5_enctype,
       krb5_data *,
       krb5_const_pointer,
       krb5_keyblock **),
    krb5_const_pointer,
    krb5_creds *,
    krb5_kdc_req *);

krb5_error_code krb5_c_make_checksum
      (krb5_cksumtype cksumtype,
		const krb5_keyblock *key, krb5_keyusage usage,
		const krb5_data *input, krb5_checksum *cksum);

krb5_error_code krb5_c_verify_checksum
      (const krb5_keyblock *key, krb5_keyusage usage,
		const krb5_data *data,
		const krb5_checksum *cksum,
		krb5_boolean *valid);

krb5_error_code krb5_c_checksum_length
      (krb5_cksumtype cksumtype, size_t *length);

krb5_error_code krb5_c_random_make_octets
      (krb5_data *data);

krb5_error_code krb5_c_decrypt
      (const krb5_keyblock *key,
      krb5_keyusage usage,
      const krb5_data *ivec,
      const krb5_enc_data *input,
      krb5_data *output);

krb5_error_code krb5_decrypt_tkt_part
      (const krb5_keyblock *srv_key,
      register krb5_ticket *ticket);

/* krb5_free.c */
void krb5_free_principal
	(krb5_principal );
void krb5_free_authenticator
	(krb5_authenticator * );
void krb5_free_authenticator_contents
	(krb5_authenticator * );
void krb5_free_addresses
	(krb5_address ** );
void krb5_free_address
	(krb5_address * );
void krb5_free_authdata
	(krb5_authdata ** );
void krb5_free_enc_tkt_part
	(krb5_enc_tkt_part * );
void krb5_free_ticket
	(krb5_ticket * );
void krb5_free_tickets
	(krb5_ticket ** );
void krb5_free_kdc_req
	(krb5_kdc_req * );
void krb5_free_kdc_rep
	(krb5_kdc_rep * );
void krb5_free_last_req
	(krb5_last_req_entry ** );
void krb5_free_enc_kdc_rep_part
	(krb5_enc_kdc_rep_part * );
void krb5_free_error
	(krb5_error * );
void krb5_free_ap_req
	(krb5_ap_req * );
void krb5_free_ap_rep
	(krb5_ap_rep * );
void krb5_free_safe
	(krb5_safe * );
void krb5_free_priv
	(krb5_priv * );
void krb5_free_priv_enc_part
	(krb5_priv_enc_part * );
void krb5_free_cred_enc_part
	(krb5_cred_enc_part *);
void krb5_free_checksum
	(krb5_checksum *);
void krb5_free_checksum_contents
	(krb5_checksum *);
void krb5_free_keyblock
	(krb5_keyblock *);
void krb5_free_keyblock_contents
	(krb5_keyblock *);
void krb5_free_pa_data
	(krb5_pa_data **);
void krb5_free_ap_rep_enc_part
	(krb5_ap_rep_enc_part *);
void krb5_free_pwd_data
	(krb5_pwd_data *);
void krb5_free_pwd_sequences
	(passwd_phrase_element **);
void krb5_free_data
	(krb5_data *);
void krb5_free_data_contents
	(krb5_data *);
void krb5_free_unparsed_name
	(char *);
void krb5_free_cksumtypes
	(krb5_cksumtype *);

#endif
