/*
 * lib/krb5/free/f_addr.c
 *
 * Copyright 1990-1998 by the Massachusetts Institute of Technology.
 *
 * Export of this software from the United States of America may
 *   require a specific license from the United States Government.
 *   It is the responsibility of any person or organization contemplating
 *   export to obtain such a license before exporting.
 * 
 * WITHIN THAT CONSTRAINT, permission to use, copy, modify, and
 * distribute this software and its documentation for any purpose and
 * without fee is hereby granted, provided that the above copyright
 * notice appear in all copies and that both that copyright notice and
 * this permission notice appear in supporting documentation, and that
 * the name of M.I.T. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior
 * permission.  Furthermore if you modify this software you must label
 * your software as modified software and not distribute it in such a
 * fashion that it might be confused with the original M.I.T. software.
 * M.I.T. makes no representations about the suitability of
 * this software for any purpose.  It is provided "as is" without express
 * or implied warranty.
 * 
 *
 * krb5_free_address()
 */

#include "seckrb.h"

void krb5_free_address(krb5_address *val)
{
   if (val->contents)
      krb5_xfree(val->contents);
   krb5_xfree(val);
   return;
}

void krb5_free_addresses(krb5_address **val)
{
   register krb5_address **temp;

   for (temp = val; *temp; temp++) {
      if ((*temp)->contents)
         krb5_xfree((*temp)->contents);
      krb5_xfree(*temp);
   }
   krb5_xfree(val);
   return;
}


void krb5_free_ap_rep(register krb5_ap_rep *val)
{
   if (val->enc_part.ciphertext.data)
      krb5_xfree(val->enc_part.ciphertext.data);
   krb5_xfree(val);
   return;
}

void krb5_free_ap_req(register krb5_ap_req *val)
{
   if (val->ticket)
      krb5_free_ticket(val->ticket);
   if (val->authenticator.ciphertext.data)
      krb5_xfree(val->authenticator.ciphertext.data);
   krb5_xfree(val);
   return;
}

void krb5_free_ap_rep_enc_part(krb5_ap_rep_enc_part *val)
{
   if (val->subkey)
      krb5_free_keyblock(val->subkey);
   krb5_xfree(val);
   return;
}

void krb5_free_authenticator_contents(krb5_authenticator *val)
{
   if (val->checksum)
      krb5_free_checksum(val->checksum);
   if (val->client)
      krb5_free_principal(val->client);
   if (val->subkey)
      krb5_free_keyblock(val->subkey);
   if (val->authorization_data)        
      krb5_free_authdata(val->authorization_data);
   return;
}

void krb5_free_authdata(krb5_authdata **val)
{
   register krb5_authdata **temp;

   for (temp = val; *temp; temp++) {
      if ((*temp)->contents)
         krb5_xfree((*temp)->contents);
      krb5_xfree(*temp);
   }
   krb5_xfree(val);
   return;
}

void krb5_free_authenticator(krb5_authenticator *val)
{
   if (val->checksum)
      krb5_free_checksum(val->checksum);
   if (val->client)
      krb5_free_principal(val->client);
   if (val->subkey)
      krb5_free_keyblock(val->subkey);
   if (val->authorization_data)        
      krb5_free_authdata(val->authorization_data);
   krb5_xfree(val);
   return;
}

void krb5_free_checksum(register krb5_checksum *val)
{
   if (val->contents)
      krb5_xfree(val->contents);
   krb5_xfree(val);
   return;
}

void krb5_free_checksum_contents(register krb5_checksum *val)
{
   if (val->contents)
      krb5_xfree(val->contents);
   return;
}

void krb5_free_cred(register krb5_cred *val)
{
   if (val->tickets)
      krb5_free_tickets(val->tickets);
   if (val->enc_part.ciphertext.data)
      krb5_xfree(val->enc_part.ciphertext.data);
   krb5_xfree(val);
   return;
}




void krb5_free_data(krb5_data * val)
{
   if (val->data)
      krb5_xfree(val->data);
   krb5_xfree(val);
   return;
}

void krb5_free_data_contents(krb5_data * val)
{
   if (val->data)
      krb5_xfree(val->data);
   return;
}

void krb5_free_etype_info(krb5_etype_info info)
{
   int i;

   for(i=0; info[i] != NULL; i++) {
      if (info[i]->salt)
         krbFreeBuf(info[i]->salt);
      krbFreeBuf(info[i]);
   }
   krbFreeBuf(info);
}


void krb5_free_enc_kdc_rep_part(register krb5_enc_kdc_rep_part *val)
{
   if (val->session)
      krb5_free_keyblock( val->session);
   if (val->last_req)
      krb5_free_last_req(val->last_req);
   if (val->server)
      krb5_free_principal(val->server);
   if (val->caddrs)
      krb5_free_addresses(val->caddrs);
   krb5_xfree(val);
   return;
}

void krb5_free_enc_tkt_part(krb5_enc_tkt_part *val)
{
   if (val->session)
      krb5_free_keyblock(val->session);
   if (val->client)
      krb5_free_principal(val->client);
   if (val->transited.tr_contents.data)
      krb5_xfree(val->transited.tr_contents.data);
   if (val->caddrs)
      krb5_free_addresses(val->caddrs);
   if (val->authorization_data)
      krb5_free_authdata(val->authorization_data);
   krb5_xfree(val);
   return;
}


void krb5_free_error(register krb5_error *val)
{
   if (val->client)
      krb5_free_principal(val->client);
   if (val->server)
      krb5_free_principal(val->server);
   if (val->text.data)
      krb5_xfree(val->text.data);
   if (val->e_data.data)
      krb5_xfree(val->e_data.data);
   krb5_xfree(val);
   return;
}

void krb5_free_kdc_rep(krb5_kdc_rep *val)
{
   if (val->padata)
      krb5_free_pa_data(val->padata);
   if (val->client)
      krb5_free_principal(val->client);
   if (val->ticket)
      krb5_free_ticket(val->ticket);
   if (val->enc_part.ciphertext.data)
      krb5_xfree(val->enc_part.ciphertext.data);
   if (val->enc_part2)
      krb5_free_enc_kdc_rep_part(val->enc_part2);
   krb5_xfree(val);
   return;
}


void krb5_free_kdc_req(krb5_kdc_req *val)
{
   if (val->padata)
      krb5_free_pa_data(val->padata);
   if (val->client)
      krb5_free_principal(val->client);
   if (val->server)
      krb5_free_principal(val->server);
   if (val->ktype)
      krb5_xfree(val->ktype);
   if (val->addresses)
      krb5_free_addresses(val->addresses);
   if (val->authorization_data.ciphertext.data)
      krb5_xfree(val->authorization_data.ciphertext.data);
   if (val->unenc_authdata)
      krb5_free_authdata(val->unenc_authdata);
   if (val->second_ticket)
      krb5_free_tickets(val->second_ticket);
   krb5_xfree(val);
   return;
}

void krb5_free_keyblock_contents(register krb5_keyblock *key)
{
   if (key->contents) {
      memset(key->contents, 0, key->length);
      krb5_xfree(key->contents);
   }
   return;
}

void krb5_free_keyblock(register krb5_keyblock *val)
{
   krb5_free_keyblock_contents(val);
   krb5_xfree(val);
   return;
}



void krb5_free_last_req(krb5_last_req_entry **val)
{
   register krb5_last_req_entry **temp;

   for (temp = val; *temp; temp++)
      krb5_xfree(*temp);
   krb5_xfree(val);
   return;
}

void krb5_free_pa_data(krb5_pa_data **val)
{
   register krb5_pa_data **temp;

   for (temp = val; *temp; temp++) {
      if ((*temp)->contents)
         krb5_xfree((*temp)->contents);
      krb5_xfree(*temp);
   }
   krb5_xfree(val);
   return;
}

void krb5_free_principal(krb5_principal val)
{
   register krb5_int32 i;

   if (!val)
      return;

   if (val->data) {
      i = krb5_princ_size(val);
      while(--i >= 0)
      {
         memset(krb5_princ_component(val, i)->data, 0,
               krb5_princ_component(val, i)->length);
         krbFreeBuf(krb5_princ_component(val, i)->data);
      }
      krb5_xfree(val->data);
   }
   if (val->realm.data)
   {
      memset(val->realm.data, 0, val->realm.length);
      krb5_xfree(val->realm.data);
   }
   krb5_xfree(val);
   return;
}

void krb5_free_priv(register krb5_priv *val)
{
   if (val->enc_part.ciphertext.data)
      krb5_xfree(val->enc_part.ciphertext.data);
   krb5_xfree(val);
   return;
}

void krb5_free_priv_enc_part(register krb5_priv_enc_part *val)
{
   if (val->user_data.data)
      krb5_xfree(val->user_data.data);
   if (val->r_address)
      krb5_free_address(val->r_address);
   if (val->s_address)
      krb5_free_address(val->s_address);
   krb5_xfree(val);
   return;
}

void krb5_free_pwd_data(krb5_pwd_data *val)
{
   if (val->element)
      krb5_free_pwd_sequences(val->element);
   krb5_xfree(val);
   return;
}


void krb5_free_pwd_sequences(passwd_phrase_element **val)
{
   if ((*val)->passwd)
      krb5_xfree((*val)->passwd);
   if ((*val)->phrase)
      krb5_xfree((*val)->phrase);
   return;
}


void krb5_free_safe(register krb5_safe *val)
{
   if (val->user_data.data)
      krb5_xfree(val->user_data.data);
   if (val->r_address)
      krb5_free_address(val->r_address);
   if (val->s_address)
      krb5_free_address(val->s_address);
   if (val->checksum)
      krb5_free_checksum(val->checksum);
   krb5_xfree(val);
   return;
}


void krb5_free_ticket(krb5_ticket *val)
{
   if (val->server)
      krb5_free_principal(val->server);
   if (val->enc_part.ciphertext.data)
      krb5_xfree(val->enc_part.ciphertext.data);
   if (val->enc_part2)
      krb5_free_enc_tkt_part(val->enc_part2);
   krb5_xfree(val);
   return;
}

void krb5_free_tickets(krb5_ticket **val)
{
   register krb5_ticket **temp;

   for (temp = val; *temp; temp++)
      krb5_free_ticket(*temp);
   krb5_xfree(val);
   return;
}




void krb5_free_unparsed_name(char * val)
{
   if (val)
      krb5_xfree(val);
   return;
}

void krb5_free_pa_enc_ts(krb5_pa_enc_ts *pa_enc_ts)
{
   if (!pa_enc_ts)
      return;
   krb5_xfree(pa_enc_ts);
}

