//**************************************************************************
//
//    Copyright 1999  Broadcom Corporation
//    All Rights Reserved
//    No portions of this material may be reproduced in any form without the
//    written permission of:
//             Broadcom Corporation
//             16251 Laguna Canyon Road
//             Irvine, California  92618
//    All information contained in this document is Broadcom Corporation
//    company private, proprietary, and trade secret.
//
//
//
//**************************************************************************
//    Filename: bcmos.h
//    Author:   Dannie Gay
//    Creation Date: 21-feb-99
//
//**************************************************************************
//    Description:
//
//	Header file for os specific definitions/includes
//		
//**************************************************************************
//    Revision History:
//
//		0.1  initial revision
//
//**************************************************************************

#ifndef bcmos_h
#define bcmos_h

#if __cplusplus
extern "C" {
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "taskLib.h"
#include "vmexLib.h"
#include "semLib.h"
#include "msgQLib.h"

#include "bsp.h"
#include "math.h"
#include "vxWorks.h"
#include <sys/socket.h>
#include <sockLib.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "zone.h"
#include "BcmTypes.h"
#include "cblmodem.h"

#define	EBADS	0x5009		//in EventLog.cpp

#define HAL_THREAD_PRIORITY_TABLE_LENGTH  8
#define HAL_THREAD_PRIORITY               230
#define HAL_THREAD_STACK_SIZE             4096
#define HAL_THREAD_RX_STACK_SIZE          4096
#define HAL_THREAD_TX_STACK_SIZE          4096

// Note: Define out BSP logging if we are compiling to reduce the image size.
#if (BCM_REDUCED_IMAGE_SIZE)
#define DEBUGMSG(z, s, w...)
#define LogRawBuffer(s, p, b, w) 
#define HalPrint(s...)              
#define	HalDebugPrint(zn, s...)     
#define MacPrint(x...)	            
#else
void LogRawBuffer (char * str, UCHAR *pBuffer, int BufLen, BOOL wacky);
int  DEBUGMSG(int zone, char *FmtStr, ...);
#define HalPrint(s...)              DEBUGMSG(ZONE_INIT, s )
#define	HalDebugPrint(zn, s...)     DEBUGMSG(zn, s )
#define MacPrint(x...)	            printf(x)
#endif


int Print(char *,...);
void Sleep(int x);


//--------------------------------------------------------------------------
// Virt/Physical Macros
//--------------------------------------------------------------------------
#define  CacheToNonCacheVirtAddress    K0_TO_K1
#define  NonCacheToCacheVirtAddress    K1_TO_K0
#define  CacheVirtToPhyAddress         K0_TO_PHYS
#define  NonCacheVirtToPhyAddress      K1_TO_PHYS
#define  PhyToNonCacheVirtAddress      PHYS_TO_K1
#define  PhyToCacheVirtAddress         PHYS_TO_K0

// alignment macro
#define 	BCM_ALIGN(addr, boundary) ((addr + boundary - 1) & ~(boundary - 1))
#define  	BCM_ALIGN_ADJUST( boundary, objSize ) ((boundary/objSize) + ((boundary % objSize) ?  1 : 0))

void   WaitLoop(unsigned long msecs);
void   WaitLoopMicro(unsigned long usecs);
void   Delayms(unsigned short);
uint32 BcmMuCreate(BcmHandle * mu);
uint32 BcmMuLock(BcmHandle mu);
uint32 BcmMuUnlock(BcmHandle mu);
uint32 BcmMuDelete(BcmHandle mu);
uint32 BcmSemCreate(BcmHandle * sem);
uint32 BcmSemLock(BcmHandle sem);
uint32 BcmSemUnlock(BcmHandle sem);
uint32 BcmCreateEvent(void);
int32  BcmWaitForEvent(uint32 event, uint32 * rEvent, uint32 msecs);
int32  BcmSendEvent(uint32 event, uint32 tid);
int32  BcmCreateThread(char * tname, uint32 tprio, uint32 tstack, uint32 * tid);
int32  BcmStartThread(uint32 tid, uint32 (*tfunc)(void *), void * targs);
int32  BcmSuspendThread(uint32 tid);
int32  BcmResumeThread(uint32 tid);
uint32 BcmStartTimerPeriodic(int32 ticks, uint32 event);
uint32 BcmStartTimerOneShot(int32 ticks, uint32 event);
uint32 BcmCancelTimer(uint32 * t_handle);
uint32 TimeMs(void);
uint32 BcmMsgQCreate(char *MsgQName, uint32 MsgQSize, uint32 MaxMsgSize);
void   BcmMsgQDelete(uint32 MsgQid);
ULONG  BcmMsgQSend(uint32 msgQId, ULONG *Msg, UINT16 MsgLen);
ULONG  BcmMsgQReceive(uint32 msgQId, ULONG *AvailMsgBuffer, uint32 AvailMsgBufferLen, 
                                                           uint32 *RcvdMsgBufferLen,BOOL WaitFlag);

//interrupt.c
void   BcmHalInterruptDisable(uint32 interruptId);
void   BcmHalInterruptEnable(uint32 interruptId);
int    BcmHalCpuInterruptDisable(void);
void   BcmHalCpuInterruptEnable( ULONG );
uint32 BcmHalMapInterrupt(uint32 (*pfunc)(), uint32 interruptId);

#define BcmHalInterruptDone   BcmHalInterruptEnable

PVOID BcmAlloc(size_t n, BOOL cached);
void  BcmFree(PVOID pvMem);
int   BcmAllocInit(void);
BOOL HalMapInterrupt(IN ULONG InterruptNumber,IN ULONG InterruptVector, OUT DWORD *InterruptId);


//**********************Special Section*********************************
// This section is dedicated to pSOS defines that are used in the code
// but cannot be found in VxWorks 
//**********************************************************************
//redefine C preprocessor for packed, Vxworks support diff syntax
#define __packed__
#define GNU_PACKED		__attribute__ ((packed))

#ifndef	htonl
/*---------------------------------------------------------------------*/
/* Definitions for Network Order Conversions Routines                  */
/*---------------------------------------------------------------------*/
#if BSP_LITTLE_ENDIAN
    /*-----------------------------------------------------------------*/
    /* Change byte order so the least significant byte is stored at    */
    /* the lowest address. Unsigned to prevent integer overflow on     */
    /* i960. Big endian cast unsigned for constant handling of vars.   */
    /*-----------------------------------------------------------------*/
    #define htonl(addr)     ( (((unsigned long)(addr) & 0x000000FF)<<24) | \
                             (((unsigned long)(addr) & 0x0000FF00)<<8)  | \
                             (((unsigned long)(addr) & 0x00FF0000)>>8)  | \
                             (((unsigned long)(addr) & 0xFF000000)>>24))

    #define ntohl(addr)     htonl(addr)

    #define htons(addr)     ( (((unsigned short)(addr) & 0x000000FF)<<8)  | \
                             (((unsigned short)(addr) & 0x0000FF00)>>8))

    #define ntohs(addr)     htons(addr)
#else

    #define htonl(a)       ((unsigned long)(a))
    #define ntohl(a)       ((unsigned long)(a))
    #define htons(a)       ((unsigned short)(a))
    #define ntohs(a)       ((unsigned short)(a))
#endif
#endif


#define ERR_TIMEOUT  0x01     /* Timed out; returned only if timeout   */
                              /* was requested                         */


//#ifndef	_ASMLANGUAGE

typedef enum parm_type
    {
    FLAG = 1, IP = 2, HEX = 3, DECIMAL = 4, CHAR = 5, STRING = 6
    } PARM_TYPE;

/*---------------------------------------------------------------------*/
/* These parameters correspond to the SD_ parameters in sys_conf.h     */
/* They are the parameters which may be stored and retrieved in the    */
/* board's parameter storage area.  In addition, the optional startup  */
/* configuration dialog can be used to change them.                    */
/*---------------------------------------------------------------------*/
struct SD_parms
    {
    /*-----------------------------------------------------------------*/
    /* Debug Mode determines how pROBE+ wil operate:                   */
    /*   DBG_SA - Standalone mode                                      */
    /*   DBG_XS - Communicating with host debugger via serial link     */
    /*   DBG_XN - Communicating with host debugger via network         */
    /*   DBG_AP - Start the application immediately (using "silent     */
    /*            startup").  pROBE+ should be configured to run in    */
    /*            standalone mode.                                     */
    /*-----------------------------------------------------------------*/
    unsigned char DebugMode;

    /*-----------------------------------------------------------------*/
    /* RbugDrvType determines how pROBE+ will connect to host debugger.*/
    /* if the Debug Mode is DBG_AP or DBG_SA and more than one type    */
    /* of the drivers available, for exmaple, both serial or network   */
    /* connection are available for connecting to the host debugger.   */
    /*  RBUG_DRV_SERIAL: use the serial port if possible               */
    /*  RBUG_DRV_NET:    use the network connection if possible        */
    /*-----------------------------------------------------------------*/
    unsigned char RbugDrvType;

    unsigned char Pad1[2];

    /*-----------------------------------------------------------------*/
    /* These variables control miscellaneous basic characteristics     */
    /* of the system:                                                  */
    /*   StartupDelay - If the startup configuration dialog is         */
    /*                  configured, it will wait this many seconds     */
    /*                  for input from the serial channel              */
    /*   DefBaud - Initial baud rate to use for the serial channels.   */
    /*             (Applications may change the baud rate via          */
    /*             de_ctrl() calls)                                    */
    /*   VmeBaseAddr - The base address of the board's dual-ported     */
    /*                 memory, as seen from the VMEbus.  Not used      */
    /*                 on non-VME boards (BSP_VME = NO)                */
    /*-----------------------------------------------------------------*/
    unsigned long StartupDelay;
    unsigned long DefBaud;
    unsigned long VmeBaseAddr;

    /*-----------------------------------------------------------------*/
    /* DefGtwyIP is used when networking is enabled to specify a       */
    /* "default gateway" node where packets should be sent in the      */
    /* absence of other routing information.                           */
    /*-----------------------------------------------------------------*/
    unsigned long DefGtwyIP;

    /*-----------------------------------------------------------------*/
    /* These variables control the LAN interface.  If SC_PNA and/or    */
    /* BSP_LAN1 is FALSE, they are ignored.                            */
    /*   Lan1 - TRUE if LAN interface should be enabled                */
    /*   Lan1IP - IP address for the LAN interface.  0 indicates that  */
    /*            RARP should be used to obtain the IP address.        */
    /*   Lan1SubnetMask - Subnet mask to use for the LAN interface     */
    /*-----------------------------------------------------------------*/
    unsigned char Lan1;
    unsigned char Pad2[3];
    unsigned long Lan1IP;
    unsigned long Lan1SubnetMask;

    /*-----------------------------------------------------------------*/
    /* If a shared memory interface is to be used (either NI or KI or  */
    /* both), it must have a node number, which is a small integer.    */
    /* SmNode holds this number.                                       */
    /*-----------------------------------------------------------------*/
    unsigned long SmNode;  

    /*-----------------------------------------------------------------*/
    /* These variables control the shared memory network interface.    */
    /*   Nism - TRUE if SMNI should be enabled                         */
    /*   NismIP - IP address to use for the interface                  */
    /*   NismSubnetMask - Subnet mask to use for the interface         */
    /*   NismDiraddr - Bus address of the system-wide "directory"      */
    /*           structure for the interface.                          */
    /* Note that SmNode, above, is also relevant to the SMNI.          */
    /*-----------------------------------------------------------------*/
    unsigned char Nism;   
    unsigned char Pad3[3];
    unsigned long NismIP;  
    unsigned long NismSubnetMask;
    unsigned long NismDiraddr;

    /*-----------------------------------------------------------------*/
    /* These variables control the shared memory kernel interface.     */
    /*   Kism - 0 if KISM should not be enabled.  A non-zero value     */
    /*          indicates that the SMKI should be enabled.  In this    */
    /*          case "Kism" should be set to the total number of nodes */
    /*          in the system.  SmNode (defined above here) is set     */
    /*          to this node's node number, which must be in the range */
    /*          1 through "Kism".                                      */
    /*   KismDiraddr - Bus address of the system-wide "directory"      */
    /*          structure for the interface                            */
    /*-----------------------------------------------------------------*/
    unsigned long Kism;
    unsigned long KismDiraddr;
    }; 

typedef struct NV_Parms {
    unsigned long magic;
    unsigned long SysVars_CheckSum;
    struct SD_parms SysVars;
 
#if BSP_PARMS
    unsigned long BspVars_CheckSum;
    UCHAR BspVars[BSP_PARMS];
#endif
 
#if SC_APP_PARMS
    unsigned long AppVars_CheckSum;
    UCHAR AppVars[SC_APP_PARMS];
#endif
} NV_Parms;

extern NV_Parms NVVars;

#define SysVars NVVars.SysVars
#define BspVars NVVars.BspVars
#define AppVars NVVars.AppVars


/*---------------------------------------------------------------------*/
/* mu_create() Definitions                                             */
/*---------------------------------------------------------------------*/
#define MU_GLOBAL       0x00000001  /* 1 = Global */
#define MU_LOCAL        0x00000000  /* 0 = Local */
#define MU_PRIOR        0x00000002  /* Queue by priority */
#define MU_FIFO         0x00000000  /* Queue by FIFO order */
#define MU_RECURSIVE    0x00000004  /* Recursive acquisition Allowed */
#define MU_NORECURSIVE  0x00000000  /* Recursive acquisition Not Allowed */
#define MU_PRIO_NONE    0x00000000  /* No unbounded priority inversion */
                                    /* protection */
#define MU_PRIO_INHERIT 0x00000028  /* Priority Inheritance protocol */
                                    /* enabled */
#define MU_PRIO_PROTECT 0x00000008  /* Priority Protection protocol */
                                    /* enabled */
/*---------------------------------------------------------------------*/
/* mu_lock() Definitions                                               */
/*---------------------------------------------------------------------*/
#define MU_NOWAIT       0x00000001  /* Don't wait for lock */
#define MU_WAIT         0x00000000  /* Wait for lock */
 
//#endif





//---------------------------- extracted from pSOS --------------------------
//The following defines are required for building the exisiting pSOS code 
//under VxWorks

//extracted from psos.h
/*---------------------------------------------------------------------*/
/* sm_p() Definitions                                                  */
/*---------------------------------------------------------------------*/
#define SM_WAIT         0x00000000  /* Wait for semaphore */
/*---------------------------------------------------------------------*/
/* ev_receive() Definitions                                            */
/*---------------------------------------------------------------------*/
#define EV_WAIT         0x00000000  /* Wait for events */
/*---------------------------------------------------------------------*/
/* as_catch(), t_mode(), and t_start() Definitions                     */
/*---------------------------------------------------------------------*/
#define T_ISR           0x00000000   /* Interrupts enabled */

//hpna required start
/*---------------------------------------------------------------------*/
/* Structure of pna functions passed to NI during ni_init              */
/*---------------------------------------------------------------------*/
struct free_rtn
    {
    void   (*free_func)(void *);
    void   *free_arg;
    };
typedef struct free_rtn frtn_t;

struct msgb
    {
    struct msgb    *b_next;    /* next msg on queue */
    struct msgb    *b_prev;    /* previous msg on queue */
    struct msgb    *b_cont;    /* next msg block of msg */
    unsigned char  *b_rptr;    /* first unread data byte in buffer */
    unsigned char  *b_wptr;    /* first unwritten data byte in buffer */
    struct datab   *b_datap;   /* data block */
    short           whichp;    /* used internally */
    short           reserved;  /* Reserved for future use */
    };
typedef struct msgb mblk_t;

struct ni_funcs
    {
    mblk_t *(*allocb)(long, long);  /* Allocate pNA buffer */
    void   (*freeb)(mblk_t *);      /* Free Message block */
    void   (*freemsg)(mblk_t *);    /* Free message */
//    mblk_t *(*esballoc)(UCHAR *, int, int, frtn_t *);
                                    /* Attach message block */
    void   (*intf_cb)(int, char*, int, long);
                                    /* pNA callback for interface operations */
    };

//hpna required end

//---------------------------- extracted from pSOS --------------------------

//***************Special section end ************************************




#if __cplusplus
}
#endif

#endif

