/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: cmmtaint.h
*    Creation Date: 17 Aug 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:20 $
*
****************************************************************************
*    Description:
*
*    Broadcom Header file for CM - MTA interface (Call-In/Out Interface Group)
*    used in CableXchange to interact between the embedded MTA and CM.
*
*    Call-In interface constists of the following interface groups:
*       - E-MTA Startup Interface Group
*
*    Call-Out interface constists of the following interface groups:
*       - Memory Management
*       - Service Flow
*       - NVRAM Services
*       - SNMP Agent common MIB services (IF, TCP, SNMP, etc. MIB implementation
*       - ASN.1 and BER common library
*       - DQOS Services
*
****************************************************************************/
#ifndef _CMMTAINT_H
#define _CMMTAINT_H

#ifdef __cplusplus
extern "C" {
#endif

#include <hausenv.h>
#include <globalCfg.h>  /* MTA IPSEC Support CFG Flag */

/*
*****************************************************************************
** TYPE DECLARATIONS
*****************************************************************************
*/

#define  MAX_SUBOPT_SIZE      256
#define  MAX_AUTH_BLOCK_SIZE  14


typedef enum CM_STATE
{
   CM_STATE_NULL,
   CM_STATE_MTA_INIT_STARTED,
   CM_STATE_MTA_INIT_COMPLETE,
   CM_STATE_MTA_SHUTDOWN_STARTED,
   CM_STATE_MTA_SHUTDOWN_COMPLETE
} CM_STATE;




/*==================================================================================*/
/* Classifier settings */
typedef struct _ClassifierSettings
{

   UINT32  destinationIPAddress;
   UINT8   destinationIPAddressSpecified;
   UINT32  sourceIPAddress;
   UINT8   sourceIPAddressSpecified;

   UINT16  destinationPort;
   UINT8   destinationPortSpecified;
   UINT16  sourcePort;
   UINT8   sourcePortSpecified;
} ClassifierSettings;

typedef enum _UsFlowSchedulingType
{
   SchedulingType_UGS,                 /* unsolicited grant service */
   SchedulingType_UGS_AD               /* unsolicited grant service with VAD */
} UsFlowSchedulingType;

/*  settings common to both US and DS flows */
typedef struct _CommonFlowSettings
{
   UINT32   packetSize;                       /* packet size starting with RTP header */
   UINT32   packetPeriodMS;                   /* period of packet generation */
   UINT32   timeoutForActiveQosParametersSeconds;    /* timeout for active/admitted QOS params */
   UINT32   timeoutForAdmittedQosParametersSeconds;  /* timeout for active/admitted QOS params */
} CommonFlowSettings;

/* service flow settings for upstream service flows */
typedef struct _UpstreamServiceFlowSettings
{
   CommonFlowSettings commonSettings;        /* settings common to both US and DS flows */
   UINT32 schedulingType;                    /* UsFlowSchedulingType (UGS / BE / UGS-AD) */
   UINT32 toleratedGrantJitterMicroseconds;  /* grant jitter */
   UINT32 toleratedPollJitterMicroseconds;   /* poll jitter */
   UINT32 nominalPollIntervalMicroseconds;   /* nominal polling interval for UGS-AD service */
} UpstreamServiceFlowSettings;

/* service flow settings for DS service flows */
typedef struct _DownstreamServiceFlowSettings
{
   CommonFlowSettings commonSettings;
} DownstreamServiceFlowSettings;

typedef struct _DocsisQosSettings
{
   UINT8                         admittedUsSettingsSpecified;
   UpstreamServiceFlowSettings   admittedUsSettings; /* Admitted US service flow settings */
   UINT8                         dscRefreshAdmittedUsSettingsEnabled;/* should CM periodically refresh US settings with DSC */

   UINT8                         activeUsSettingsSpecified;
   UpstreamServiceFlowSettings   activeUsSettings;   /* Active US service flow settings */
   ClassifierSettings            usClassifier;       /* Classifier for US flow */

   UINT8                         admittedDsSettingsSpecified;
   DownstreamServiceFlowSettings admittedDsSettings;/* Admitted DS service flow settings */
   UINT8                         dscRefreshAdmittedDsSettingsEnabled; /* should CM periodically refresh US settings with DSC */
   UINT8                         activeDsSettingsSpecified;
   DownstreamServiceFlowSettings activeDsSettings;  /* Active DS service flow settings */
   ClassifierSettings            dsClassifier;       /* Classifier for DS flow */

   UINT8                         ipTOSbyte;          /* 8 bit type of service for Tx IP header */
   UINT8                         authorizationBlock[MAX_AUTH_BLOCK_SIZE]; /* authorization block code */
   UINT32                        authBlockSize;       /* number of bytes in the auth block*/
} DocsisQosSettings;

/* This is in case we decide to change the core type of the handle, so the
** API declarations can be the same. */
typedef UINT32 CmConnectionHandle;
typedef UINT32 EmtaConnectionHandle;

/* SNMP_VIEW_STATUS values control OID inclusion/exclusion */
typedef unsigned long SNMP_VIEW_STATUS;
#define VIEW_STATUS_EXCLUDE           0
#define VIEW_STATUS_INCLUDE           1
/* SNMP_SEC_LEVEL values - note these may be bitwise or-ed together */
typedef unsigned long SNMP_SEC_LEVEL;
#define SNMP_SEC_NONE                 0
#define SNMP_SEC_AUTH                 1
#define SNMP_SEC_PRIV                 2

/* SNMP_SEC_Model values  */
typedef unsigned long SNMP_SEC_MODEL;
#define SNMP_SEC_MDL_V1               1
#define SNMP_SEC_MDL_V2               2
#define SNMP_SEC_MDL_USM              3 /* Only MODE used by the PacketCable! */



/* Values for the result parameter passed in various callback functions. */
typedef enum
{
   kFailure,                   /* Negotiations with the headend failed. */
   kSuccessful,                /* It worked! */
   kTransactionErred           /* Fatal error - serviceflow deleted */
} ServiceFlowResult;



/*
** Date types for NVRAM Blind Area Handling
*/
typedef struct DQOS_NONVOL_CONFIG
{
   UINT8    enforceGateId;       /* If set, serviceflows will allocate dedicated QoS
                                 ** only if a gate-id is provided by the CMS. If the
                                 ** gate-id is absent, serviceflows will use
                                 ** best-effort. If not set, service flows will use
                                 ** dedicated QoS independent of whether a gate-id
                                 ** is provided by the CMS. */
   UINT8    enableUgsAd;         /* Enable/disable UGS-AD support */
   UINT8    enableAuthBlockTlv;  /* Enable/disable new authorization block TLV
                                 ** format. Legacy format is a 32-bit gate-id */
   UINT8    reserved[13];        /* Reserved for future use */
} DQOS_NONVOL_CONFIG;

typedef struct tagBLIND_MTA_NONVOL
{
   int                  magicNumber;
   int                  version;
   char                 snmpV3SessKeyAndTktTimes[40];
   char                 mtaHwrVersion[64];
   char                 mtaSerialNum[130];
   DQOS_NONVOL_CONFIG   dqosConfig;
   int                  mtaKerberosFlag;
   int                  mtaMediaSecDisabled;
   int                  mtaIpsecDisabled;       /* IPSEC Will/Will_Not be established */
   int                  mtaSnmpInformDisabled;  /* == TRAP Disabled */
   char                 mtaSnmpMgrUserName[64]; /* SNMP User Name of the Prov Server */
   char                 mtaSnmpMgrEngineID[64]; /* SNMP Engine ID of the Prov Server */
   UINT32               mtaSnmpMgrEngineBoot;   /* SNMP Engine Boot of the Prov Server */
   UINT32               mtaSnmpMgrEngineTime;   /* SNMP Engine Time of the Prov Server */
   int                  mtaKrbContProv;    /* Continue provisioning if kerberos failes */
   int                  mtaFQDNCheckDisabled;  /*Validate FQDNs in Provisioning. FQDNs with invalid syntax rejected*/
} BLIND_MTA_NONVOL;

#define  NV_BLIND_MTA_VER        1
#define  NV_BLIND_MAGIC_NUMBER   0xacecafe

/* Values for the result parameter passed in various callback functions. */


/*****************************************************************************
** EXTERNAL FUNCTION PROTOTYPES
******************************************************************************/



/*****************************************************************************
** E-MTA Startup Interface Group
**
**    - mtaInit
**    - mtaConsole
**
******************************************************************************/

/*****************************************************************************
** FUNCTION:   mtaInit
**
** PURPOSE:    to start and initialize the MTA CALLCLIENT object
**
** PARAMETERS: isStandAlone - indicates whether or not the CM should run
**                            in stand-alone ethernet mode, or full
**                            DOCSIS mode
**             cmIface (V2 only)  - Pointer to CM-INTERFACE
**
** RETURNS:    SUCCESS (0)
**             FAILURE (!=0)
**
*****************************************************************************/
typedef struct CM_INTERFACE*  PCM_INTERFACE;
int mtaInit( BOOL isStandAlone, PCM_INTERFACE pcmIface);

/*
** This is for storing the the PacketCable Servers option values (which are
** returned by the DHCP server).  Each value is a DISPLAYSTRING, and the maximum
** length is 255 bytes; I have added one byte for a \0 string terminator.
** This structure is similar to what the V1 had with the extensions for the
** new Sub-Option (secondary DHCP)
*/

typedef struct _PacketCableServers
{
      char dhcpServerAddress[MAX_SUBOPT_SIZE];
      char snmpEntityAddress[MAX_SUBOPT_SIZE];
      char primaryDNSAddress[MAX_SUBOPT_SIZE];
      char secondaryDNSAddress[MAX_SUBOPT_SIZE];
      char dhcpServerAddress2[MAX_SUBOPT_SIZE];
      char provKerbRealmName[MAX_SUBOPT_SIZE];
      UINT32 tgtKerberosExec;
      UINT32 provMtaProvTimer;

      BOOL foundDHCPServerAddress;
      BOOL foundSNMPEntityAddress;
      BOOL foundPrimaryDNSAddress;
      BOOL foundSecondaryDNSAddress;
      BOOL foundDHCPServerAddress2;
      BOOL foundProvKerbRealmName;
      BOOL foundTgtKerberosExec;
      BOOL foundProvMtaProvTimer;
} PacketCableServers;


/*
*****************************************************************************
** FUNCTION:   cmInit
**
** PURPOSE:    CM-MTA interface module initialization function. Initializes
**             the interface and allocates required resources.
**
** PARAMETERS: state - CM_STATE_MTA_INIT_STARTED, or CM_STATE_MTA_INIT_COMPLETE
**
** RETURNS:    None
**
*****************************************************************************
*/
void cmInit( CM_STATE state );


/*
*****************************************************************************
** FUNCTION:   cmDeInit
**
** PURPOSE:    CM-MTA interface module de-initialization function. Frees
**             module resources.
**
** PARAMETERS: state - CM_STATE_MTA_SHUTDOWN_STARTED, or CM_STATE_MTA_SHUTDOWN_COMPLETE
**
** RETURNS:    None
**
*****************************************************************************
*/
void cmDeInit( CM_STATE state );


/*
*****************************************************************************
** FUNCTION:   mtaConsole
**
** PURPOSE:    Enters the debug console for the MTA.
**
** PARAMETERS: none
**
** RETURNS:    none
**
*****************************************************************************
*/
void mtaConsole( void );

/*****************************************************************************
** Memory Management Interface Group
**
**    - cmMemAlloc
**    - cmMemFree
**
******************************************************************************/

/*****************************************************************************
** FUNCTION:   cmMemAlloc
**
** PURPOSE:    Allocates a memory block of size "bufferSize" used by the
**             ServiceFlow API calls to send the packet data over the DOCSIS.
**             Buffer allocations should be on a 32-bit boundary.
**
** PARAMETERS: bufferSize - size of the memory block to be allocated (bytes)
**             connHandle - for V1 is 0, for V2 SF Connection Handle
**
** RETURNS:    If success, returns an address of the memory block.
**             NULL, otherwise.
**
*****************************************************************************/
void* cmMemAlloc (CmConnectionHandle connHandle, UINT32 bufferSize);

/*****************************************************************************
** FUNCTION:   cmMemFree
**
** PURPOSE:    Frees a memory block pointed to by the "pBuffer" and previously
**             allocated by the "cmMemAlloc" call.
**
** PARAMETERS: pBuffer - pointer to the memory block to be freed.
**             connHandle - for V1 is 0, for V2 - SF Connection Handle
**
** RETURNS:    If success, returns 1 (true)
**             otherwise, returns 0 (false)
**
** NOTES:      The pointer to the memory block may not necessarily be at the
**             start of the memory buffer but may be indise it.  For example,
**             when an RTP packet is received the IP/UDP headers are parsed
**             off first, and the remaining portion of the buffer passed on
**             the DSP. Therefore the lower layer software may have no knowledge
**             of the amount of preamble present.
**
*****************************************************************************/
BOOL cmMemFree (CmConnectionHandle connHandle, void* pBuffer);


/*****************************************************************************
** FUNCTION:   cmReleasePacket
**
** Releases packets transmitted to the EMTA with ServiceFlowReceiveCallback
**
** PARAMETERS:
**             connHandle  - handle used to identify the connection
**             bufDesc     - used by the CM to identify the packet being released.
**                           This was passed in as an argument to
**                           ServiceFlowReceiveCallback.
**
** RETURNS:    None
**
****************************************************************************/
void cmReleasePacket ( CmConnectionHandle connHandle, UINT32 bufDesc);


/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   mtaServiceFlowAddChangeResult
//
// PURPOSE:    result of a previous cmSfAdd or cmSfChange request.
//
// PARAMETERS: emtaConnectionHandle - connection handle to correlate the
//                  result to the request.
//             result - enumerated result
//             cmConnectionHandle - connection handle provide by the CM
//                  to be used in all future requests for this flow.
//
// RETURNS:    true if successful
//             false otherwise
/////////////////////////////////////////////////////////////////////////////////
typedef void (*PFN_EMTA_IF_mtaServiceFlowAddChangeResult)(EmtaConnectionHandle emtaConnectionHandle,
                                                          ServiceFlowResult result,
                                                          CmConnectionHandle cmConnectionHandle,
                                                          const DocsisQosSettings * pSettings
                                                         );

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   mtaServiceFlowDeleted
//
// PURPOSE:    invoked when at least 1 of 2 flows in this connection
//               are deleted for any reason
//
// PARAMETERS: emtaConnectionHandle - connection handle to correlate the
//                  result to the request.
//
// RETURNS:    true if successful
//             false otherwise
/////////////////////////////////////////////////////////////////////////////////
*/
typedef void (*PFN_EMTA_IF_mtaServiceFlowDeleted)(EmtaConnectionHandle emtaConnectionHandle);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   mtaServiceFlowChanged
//
// PURPOSE:    invoked when at a CMTS initiated DSC transaction has completed
//               on either of the 2 flows in this connection.
//
// PARAMETERS: emtaConnectionHandle - connection handle to correlate the
//                  result to the request.
//             pSettings - settings that we changed to.  you MUST copy the settings
//                  over if you want to use them after you return.
//
// RETURNS:    true if successful
//             false otherwise
/////////////////////////////////////////////////////////////////////////////////
*/
typedef void (*PFN_EMTA_IF_mtaServiceFlowChanged)(EmtaConnectionHandle emtaConnectionHandle,
                                                  const DocsisQosSettings * pSettings
                                                 );



/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   mtaServiceFlowDeleteResult
//
// PURPOSE:    result of a previous cmSfDelete request.
//
// PARAMETERS: emtaConnectionHandle - connection handle to correlate the
//                  result to the request.
//             result - enumerated result
//
// RETURNS:    none
/////////////////////////////////////////////////////////////////////////////////
*/
typedef void (*PFN_EMTA_IF_mtaServiceFlowDeleteResult)(EmtaConnectionHandle emtaConnectionHandle,
                                                       ServiceFlowResult result
                                                      );


/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   PFN_EMTA_IF_mtaPacketReceiveCallback
//
// PURPOSE:    used to transmit voice RTP packets to the EMTA.
//
// PARAMETERS: emtaConnectionHandle - connection handle to identify the
//                  connection
//             cmConnectionHandle - handle to be used when the packet is released
//             bufDesc - buffer descriptor to be used when packet is released
//             pRtpPacket - points to first byte after UDP header
//             size - number of bytes in packet after UDP header
//
// RETURNS:    true if successful
//             false if you want an Error message to print with the
//               entire contents of the packet including IP headers.
//               useful for debugging.
/////////////////////////////////////////////////////////////////////////////////
typedef BOOL (*PFN_EMTA_IF_mtaPacketReceiveCallback)(EmtaConnectionHandle emtaConnectionHandle,
                                                     CmConnectionHandle cmConnectionHandle,
                                                     UINT32 bufDesc,
                                                     void * pRtpPacket,
                                                     UINT32 size
                                                    );

/*****************************************************************************
** FUNCTION:   cmSfSend
**
** PURPOSE:    Adds the UDP/IP and Ethernet LLC headers to the specified
**             packet, and queues it for transmission on the connection
**             associated with the specified handle.
**
** PARAMETERS: connHandle      - handle of the connection to be used
**             pSRAMBuffer     - pointer to the SRAM buffer containing the audio
**                               packet.
**             bufferLength    - buffer length (bytes)
**
** RETURNS:    If success, return 1.
**             Otherwise, 0.
**
****************************************************************************/
BOOL cmSfSend (
                 CmConnectionHandle   connHandle,
                 void               *pSRAMBuffer,
                 UINT32             bufferLength
              );


/*****************************************************************************
** Service Flow Interface Group
**
**    - cmSfAdd
**    - cmSfDelete
**    - cmSfChange
**
******************************************************************************/

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSfAdd
//
// Creates a new bidirectional ServiceFlow with the headend with the
// specified parameters. When the ServiceFlow is created, the ServiceFlowCreateCallback
// callback function is called with the result.
//
// Parameters:
//      pSettings                   - QOS settings for this connection
//      emtaConnectionHandle        - a user defined valued to be passed to the receive callback function
//      pResultCallback             - callback to be invoked with the status of this request
//      pDeletedCallback            - callback to be invoked if and flow created was deleted
//      pChangedCallback            - callback to be invoked if any flow was changed by a CMTS initiated transaction
//
// Returns:  true/false indicating failure or success
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmSfAdd (const DocsisQosSettings * pSettings,
               EmtaConnectionHandle emtaConnectionHandle,
               PFN_EMTA_IF_mtaServiceFlowAddChangeResult pResultCallback,
               PFN_EMTA_IF_mtaServiceFlowDeleted pDeletedCallback,
               PFN_EMTA_IF_mtaServiceFlowChanged pChangedCallback,
               PFN_EMTA_IF_mtaPacketReceiveCallback pPacketReceive
               );


/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSfDelete
//
// Deletes an existing service flow
//
// Parameters:
//      connHandle                  - handle of the connection to be deleted
//
// Returns:  true/false indicating failure or success
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmSfDelete(CmConnectionHandle connHandle,
               PFN_EMTA_IF_mtaServiceFlowDeleteResult pDeleteResultCallback);


/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSfChange
//
// Modifies an existing service flow
//
// Parameters:
//      connHandle                  - CM connection handle this request refers to
//      pSettings                   - QOS settings for this connection
//      pResultCallback             - callback to be invoked with the status of this request
//
// Returns:  true/false indicating failure or success
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmSfChange( CmConnectionHandle connHandle,
                  const DocsisQosSettings * pSettings,
                  PFN_EMTA_IF_mtaServiceFlowAddChangeResult pResultCallback
                 );

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSfSetUgsAdGrantsActive
//
// When the EMTA detects silence and want to drop from
// UGS to real-time-poll, he should call this with and argument of false.
// To go back to UGS mode, call it with true.
// This method should only be called if SchedulingType_UGS_AD scheduling
// was specified in the schedulingType field of UpstreamServiceFlowSettings.
// If UGS or Best Effort scheduling is specified, the request will be ignored and
// false returned.
// If any packets are transmitted while not active, they will be rejected.
// By default, after a service flow is set up, the grants are active.
//
// Parameters:
//      connHandle                  - CM connection handle this request refers to
//      grantsActive                - true grants should be active
//
// Returns:
//      BOOL                        - true if request was successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmSfSetUgsAdGrantsActive( CmConnectionHandle connHandle,
                                 BOOL grantsActive
                              );


/*****************************************************************************
** Misc, Net Interface Group
**
**    - cmMiscGetMtaMacAddr
**    - cmGetIpStackIndex
**    - cmResetBoard
******************************************************************************/

/*****************************************************************************
** FUNCTION:   cmMiscGetMtaMacAddr
**
** PURPOSE:    Retrieves the MTA's Ethernet address
**
** PARAMETERS: mtaEthAddr - array will contain the 6-bytes of MTA's MAC
**                          address upon return.
**
** RETURNS:    none
**
*****************************************************************************/
void cmMiscGetMtaMacAddr (unsigned char* mtaEthAddr);

/*****************************************************************************
** FUNCTION:   cmGetIpStackIndex()
**
** PURPOSE:    Called by EMTA to retrieve the index used for referencing the
**             EMTAs IP stack.
**
** PARAMETERS: none
**
** RETURNS:    index
*****************************************************************************/
UINT32 cmGetIpStackIndex( void );

/*****************************************************************************
** FUNCTION:   cmResetBoard
**
** PURPOSE:    Perform a board reset
**
** PARAMETERS: none
**
** RETURNS:    none
****************************************************************************/
void cmResetBoard( void );

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmMiscGetUTCTimeOffset()
//
// PURPOSE:    Called by EMTA to retrieve the time offset from
//             Coordinated Universal Time.  Units in seconds.
//             Obtained from the CM DHCP process.
//
// PARAMETERS: none
//
// RETURNS:    offset in seconds
/////////////////////////////////////////////////////////////////////////////////
*/
int cmMiscGetUTCTimeOffset(void);

/*****************************************************************************
** FUNCTION:   cmNetGetDhcpSuboptions()
**
** PURPOSE:    Retrieves the DHCP IP Address for MTA received in sub-option 1
**             and sub-option 2 of option #177 during the CM's DHCP process.
**
** PARAMETERS: dhcpSubOpts - pointer to the "PacketCableServers" structure
**             which is V1-V2 compatible.
**
** RETURNS:    If success, return 1.
**             Otherwise, 0.
**
*****************************************************************************/
BOOL cmNetGetDhcpSuboptions(PacketCableServers *dhcpSubOpts);


/*****************************************************************************
** Non-Vol Group
**
******************************************************************************/
/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvGetTelephonyRootCertificate
//
// The Telephony Root Certificate value.
//
// Parameters:
//      pBuf - buffer to write to
//      buf_len - length of the target buffer
// Returns:
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmNvGetTelephonyRootCertificate(UINT8 *pBuf, unsigned int *buf_len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvSetTelephonyRootCertificate
//
// The Telephony Root Certificate value.
//
// Parameters:
//      pBuf - buffer to read from
//      len  - length of the value to set
// Returns:
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvSetTelephonyRootCertificate(const UINT8  *pBuf, unsigned int len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvGetMtaManufacturerCertificate
//
// The number of valid bytes of data in the MTA Manufacturer Certificate field.
//
// Parameters:
//          pBuf - buffer to write to
//          buf_len - length of buffer
// Returns:
//        BOOL - true if successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvGetMtaManufacturerCertificate(UINT8 *pBuf, unsigned int *buf_len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvSetMtaManufacturerCertificate
//
// The number of valid bytes of data in the MTA Manufacturer Certificate field.
//
// Parameters:
//      pBuf - buffer to read from
//      len  - length of the value to set
// Returns:
//        BOOL - true if successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmNvSetMtaManufacturerCertificate(const UINT8 *pBuf, unsigned int len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvGetMtaDeviceCertificate
//
// The MTA Device Certificate value.
//
// Parameters:
//          pBuf - buffer to write to
//          buf_len - length of buffer
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvGetMtaDeviceCertificate(UINT8 *pBuf, unsigned int *buf_len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvSetMtaDeviceCertificate
//
// The MTA Device Certificate value.
//
// Parameters:
//      pBuf - buffer to read from
//      len  - length of the value to set
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmNvSetMtaDeviceCertificate (const UINT8 *pBuf, unsigned int len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvGetMtaDevPrivKeyModulus
//
// The MTA Device Private Key Modulus value
//
// Parameters:
//          pBuf - buffer to write to
//          buf_len - length of buffer
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvGetMtaDevPrivKeyModulus(UINT8 *pBuf, unsigned int *buf_len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvSetMtaDevPrivKeyModulus
//
// The MTA Device Private Key Modulus value
//
// Parameters:
//      pBuf - buffer to read from
//      len  - length of the value to set
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvSetMtaDevPrivKeyModulus(const UINT8 *pBuf, unsigned int len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvGetMtaDevPrivKeyExponent
//
// The MTA Device Private Key Exponent value.
//
// Parameters:
//          pBuf - buffer to write to
//          buf_len - length of buffer
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmNvGetMtaDevPrivKeyExponent(UINT8  *pBuf, unsigned int *buf_len);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvSetMtaDevPrivKeyExponent
//
// The MTA Device Private Key Exponent value.
//
// Parameters:
//      pBuf - buffer to read from
//      len  - length of the value to set
// Returns:
//        BOOL - true if write successful
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL  cmNvSetMtaDevPrivKeyExponent(const UINT8 *pBuf, unsigned int len);


/*****************************************************************************
** FUNCTION:   cmNvGetPacketCableKerberosRealm
**
** This is the "sub-option-6" of Option 177 - "Kerberos Realm"
** string that the MTA will use.  The string will have a \0 terminator.
**
** Parameters:
**        pBuf - destination buffer
**        buf_len - length of destination buffer
** Returns:
**        BOOL - 0 if fucntion was successful
*****************************************************************************/
BOOL   cmNvGetPacketCableKerberosRealm(char * pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvGetTGTActive
**
** PURPOSE:    Function returns the BOOL value of teh Flag stored in the
**             NV memory corresponding to the Suboption-7 (TGT Flag).
**
** Parameters:
**        pTgtActive - pointer to the output BOOL variable
** Returns:
**        BOOL - 0 if function was successful, <> 0 - otherwise
**
*****************************************************************************/
BOOL   cmNvGetTGTActive(BOOL *pTgtActive);

/*****************************************************************************
** FUNCTION:   cmNvIsMtaDhcpEnabled
**
** PURPOSE:    Returns TRUE if MTA DHCP is set to TRUE in NV-ram
**
** PARAMETERS: none.
**
** RETURNS:    TRUE   if MTA DHCP is set to TRUE in NV-ram
**             FALSE  otherwise.
**
*****************************************************************************/
BOOL  cmNvIsMtaDhcpEnabled(void);

/*****************************************************************************
** FUNCTION:   cmNvIsStandaloneEnabled
**
** PURPOSE:    Checks if MTA is running in "standalone" mode (no HFC)
**
** PARAMETERS: none.
**
** RETURNS:    TRUE   if MTA is running in "standalone" mode (set in NV ram)
**             FALSE  otherwise.
**
*****************************************************************************/
BOOL  cmNvIsStandaloneEnabled(void);

/*****************************************************************************
** FUNCTION:   cmMiscGetMtaCfgFileName
**
** PURPOSE:    Retrieves the MTA's Proprietory Configuration String.
**
** PARAMETERS: mtaCfgFile - array will contain the MTA's Proprietory
**                          Configuration String upon return.
** NOTES:      Proprietory Configuration String should contain CA name and
**             MTA name in the format:
**             <CA_Name>/<MTA_Name>
**
** RETURNS:    0 if OK, <>0 - otherwise
**
*****************************************************************************/
int cmMiscGetMtaCfgFileName (unsigned char* mtaCfgFile);


/*****************************************************************************
**
** SNMP Agent common MIB services Interface Group
**
******************************************************************************/
/*****************************************************************************
** FUNCTION:   cmSnmpMtaAgentRegister
**
** PURPOSE:    Registers MTA SNMP Agent's IP Addr to SNMP Engine on the
**             CM's side.
**
** PARAMETERS: agntIpAdde  - IP Address of the MTA's SNMP Agent.
**             ipIfaceNum  - IP Iface Number on which the SNMP Agent is working
**
** RETURNS:    0 if OK, or <> 0 - otherwise
**
*****************************************************************************/
BOOL  cmSnmpMtaAgentRegister(IPADDR agntIpAddr, int ipIfaceNum);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpSendNotify
//
// PURPOSE:    Send a RFC 2573 SNMPv3 Notify message with the specified OID and variable
//             bindings.  The snmpNotify, snmpTargetAddr, and snmpTargetParams tables
//             determine how and where this notify is sent.  These tables must be initialized
//             prior to this call.
//
// PARAMETERS: void *trapOid: The OID used to identify this trap.
//                   This must be of type OBJ_ID_T, per Epilogue SNMP
//                   agent specifications.
//             void *vbList: A list of variable bindings to be attached to the
//                   trap messsage.  This must be of type VBL_T, per Epilogue
//                   SNMP agent specifications.  Note that this argument must
//                   be NULL if no variable bindings are to be attached.
//             unsigned char *ElementName:  A name used as the index into the
//                    snmpNotifyTable. This should be the same name as used in
//                    the cmSnmpInitNotifys function.
//             int ElementNameLength: The length of the ElementName.
//
// RETURNS:    0 on success.
//
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpSendNotify (void *trapOid, void* vbList,
                     unsigned char * ElementName, int ElementNameLen);


/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpPktcSnmpV3Init
//
// PURPOSE:    Initialize the EMTA's SNMPv3 VACM data with the supplied user
//             name and view name.
//
// PARAMETERS: char *pSecName: The SNMPv3 security user name to provision
//             int SecLevel: The security level associated with the security name.
//                           SNMP_SEC_NONE (no auth no priv)
//                           SNMP_SEC_AUTH (auth no priv)
//                           SNMP_SEC_PRIV (auth and priv)
//             unsigned char* pAuthKey: The password for authorization. NULL if not used.
//             int AuthKeyLen: The length of the pAuthKey, zero if not used.
//             int AuthType: The authorization protocol: EMTA_AUTH_MD5/EMTA_AUTH_SHA,
//                           zero if not used.
//             unsigned char *pPrivKey: The password for privacy.  NULL if not used.
//             int PrivKeyLen: The length of the pPrivKey, zero if not used.
//             char* pViewName: The name of the view into the MIB.  The view will be
//                              provisioned to allow access to the entire .1.3.6 subtree.
//             int LocalizePwd: 0= don't localize the password keys, 1 do localize keys
//
// RETURNS:    0 on success, -1 on failure.
//
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpPktcSnmpV3Init (char *pSecName, int SecLevel,
                         unsigned char *pAuthKey, int AuthKeyLen,
                         int AuthType, unsigned char *pPrivKey,
                         int PrivKeyLen, char * pViewName, int LocalizePwd
                         );

/*
////////////////////////////////////////////////////////////////////////////////
//
// FUNCTION:   cmSnmpInitV3UserComplete
//
// PURPOSE:    This function installs an SNMPv3 user given the input
//             parameters.  The view will show the entire MIB.
//             ** This function obsoletes EmtaSnmpPktcSnmpV3Init(). **
//
// PARAMETERS: pSecName:  The SNMPv3 security user name.
//             SecLevel:  SNMP_SEC_NONE (no auth no priv)
//                        SNMP_SEC_AUTH (auth no priv)
//                        SNMP_SEC_PRIV (auth and priv)
//             pEngineId:  The engine id for this user.
//                         If NULL then the local agent engine id will be used.
//             EngineIdLen: The length of the engine id.
//             pAuthKey:   The password for authorization. NULL if not used.
//             AuthKeyLen: Length of the pAuthKey
//             AuthType:   The authorization protocol: EMTA_AUTH_MD5/EMTA_AUTH_SHA,
//                           zero if not used.
//             pPrivKey:   The password for privacy.  NULL if not used.
//             PrivKeyLen: Length of the pPrivKey.
//             pViewName:  The name of the Mib view for this user.
//             Access:     Sets the access for this user.  Access is entire mib for:
//                          0x4 = Read Access
//                          0x2 = Write Access
//                          0x1 = Notify Access
//             Password2Key: If true, the func performs a password-to-key translation
//                           on the values given for pAuthKey and pPrivKey. This parameter
//                           is ignored if LocalizePwd is set to 1.
//                           0- do not convert passwords to keys.
//                           1- change passwords to keys.
//             LocalizePwd:  0- do not localize keys, 1- localize keys
//
//  RETURNS:   Zero if successful.
//
////////////////////////////////////////////////////////////////////////////////
*/
int   cmSnmpInitV3UserComplete (char *pSecName, int SecLevel,
                                unsigned char *pEngineId, int EngineIdLen,
                                unsigned char *pAuthKey, int AuthKeyLen, int AuthType,
                                unsigned char *pPrivKey, int PrivKeyLen,
                                char *pViewName, int Access, int Password2Key, int LocalizePwd
                                );


/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpGetEngineParams
//
// PURPOSE:    This function gets the SNMPv3 engine parameters for this agent.
//
// PARAMETERS: unsigned char *pEngineId: The engine id.  This buffer should be at
//                  least 12 bytes to get the current implementation of engine id.
//                  Success returns the engine Id in the pEngineId buffer.
//             int MaxEngineLen: The max size of the buffer pointed to by pEngineId.
//                  This routine copies the Engine Id into the suppied
//                  pEngineId buffer or upto the Max size of the buffer.
//             int* EngineIdLen: Success returns the number of bytes of the pEngineId.
//             int* EngineBoots: Success returns the engineBoots variable. (The number
//                  of times this SNMP engine has rebooted)
//             int* EngineTime: Success returns the number of seconds since this
//                  SNMP engine has rebooted.
//
// RETURNS:    0 on success.
//
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpGetEngineParams (unsigned char * pEngineId, int MaxEngineLen,
                           int *pEngineIdLen, int *pEngineBoots, int *pEngineTime);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpAddEngineAddr
//
// PURPOSE:    This function is called to add a foreign engine id.
//             The engine is installed and associated with an Ip address.
//             This function is typically called to add an SNMP Manager
//             engine as the authoratative engine for Informs.  To
//             enable Informs a user must be created (through separate
//             calls) using the engine id set here.
//
// PARAMETERS: unsigned char * pEngineId:   The engine id to install.
//             int             EngineIdLen: Length of the engine id to install.
//             unsigned int    EngineBoots: The engineBoots var for the engine.
//             int             EngineTime:  The engineTime var for the engine.
//             unsigned int    MgrIpAddress: The Ip address associated with the engine id.
//
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpAddEngineAddr(unsigned char * pEngineId, int EngineIdLen,
                       unsigned int EngineBoots, int EngineTime,
                       unsigned int MgrIpAddress);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpDestroyUser
//
// PURPOSE:    This function removes an SNMPv3 user and cleans up any allocated
//             memory (Deinstall and Destroy).
//
// PARAMETERS: char * pUserName:   The name of the user to destroy.
//
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpDestroyUser (char * pUserName);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpConfigView
//
// PURPOSE:    This function is called to add SNMPv3 views.
//             It may be called any number of times to either create a new
//             view or to include/exclude elements of an exiting view.
//
// PARAMETERS: char *pViewName:         The name of the view.
//             unsigned long *pSubtree: The OID to exclude/include.
//             int SubtreeSize:         The size of pSubtree.
//             SNMP_VIEW_STATUS Status: VIEW_STATUS_INCLUDE, VIEW_STATUS_EXCLUDE
//
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpConfigView (char *pViewName, unsigned long *pSubtree,
                      int SubtreeSize, SNMP_VIEW_STATUS Status);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpInstallAccess
//
// PURPOSE:    This function installs an access.  It associates views with a user name.
//
// PARAMETERS: char *pSecName:    The user name.
//             char *pReadView:   The view this user will use for read access.
//             char *pWriteView:  The view this user will use for write access.
//             char *pNotifyView: The view this user will use for notify access.
//             SNMP_SEC_LEVEL SecLevel:  SNMP_SEC_NONE, SNMP_SEC_AUTH, SNMP_SEC_PRIV.
//             SNMP_SEC_MODEL SecModel:  SNMP_SEC_MDL_USM.
//
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpInstallAccess (char *pSecName, char *pReadView,
                        char *pWriteView, char *pNotifyView,
                        SNMP_SEC_LEVEL SecLevel, SNMP_SEC_MODEL SecModel);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpInstallGroup
//
// PURPOSE:    This function installs an SNMPv3 group and associates it with
//             a user.
//
// PARAMETERS: char *pGroupName:         The name of the group to create.
//             char *pSecName:           The user name.
//             SNMP_SEC_MODEL SecModel:  The security model - SNMP_SEC_MDL_USM.
//
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpInstallGroup(char *pGroupName, char *pSecName,
                       SNMP_SEC_MODEL SecModel);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpInstallUserComplete
//
// PURPOSE:    This function installs an SNMPv3 user given the input
//             parameters.
//
// PARAMETERS: pSecName:  The SNMPv3 security user name.
//             SecLevel:  SNMP_SEC_NONE (no auth no priv)
//                        SNMP_SEC_AUTH (auth no priv)
//                        SNMP_SEC_PRIV (auth and priv)
//            pEngineId:  The engine id for this user.
//                        If NULL then the local agent engine id will be used.  In
//                        general this param will be NULL unless the user is for informs.
//            EngineIdLen:The length of the engine id.
//            pAuthKey:   The password for authorization. NULL if not used.
//            AuthKeyLen: Length of the pAuthKey
//            AuthType:   The type of authorization to use.
//                        EMTA_AUTH_MD5,  EMTA_AUTH_SHA
//            pPrivKey:   The password for privacy.  NULL if not used.
//            PrivKeyLen: Length of the pPrivKey.
//            Password2Key: If true, the func performs a password-to-key translation
//                          on the values given for pAuthKey and pPrivKey. This parameter
//                          is ignored if LocalizePwd is set to 1.
//                          0- do not convert passwords to keys.
//                          1- change passwords to keys.
//            LocalizePwd: 0- do not localize keys, 1- localize keys  //
// RETURNS:    Zero on success
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpInstallUserComplete (char *pSecName, int SecLevel,
                                unsigned char *pEngineId, int EngineIdLen,
                                unsigned char *pAuthKey, int AuthKeyLen, int AuthType,
                                unsigned char *pPrivKey, int PrivKeyLen,
                                int Password2Key, int LocalizePwd);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpInitNotifys
//
// PURPOSE:    This function enables RFC2573 notifys by adding a row to
//             the snmpNotifyTable, snmpTargetAddrTable and snmpTargetParamsTable.
//             These tables determine how and where a notify is sent.  This function
//             may be called multiple times to add any number of rows; however,
//             it should not be called with duplicate ElementNames.
//
// PARAMETERS: unsigned char * ElementName: A name used as the index and tag
//                   name parameters in the tables.
//             int ElementNameLength: The length of the ElementName.
//             UINT32 TargetIpAddr: The Ip Address where notifys
//                   should be sent.
//             char * pSecName: The security name to use when sending the notify.
//                   This name/user must be installed and active before the notify
//                   can be sent.
//             int SecLevel:  The security level associated with the security name
//                   used for sending the notify.
//             int NotifyType: The notify type.  EMTA_SNMP_NOTIFY_TRAP,
//                   EMTA_SNMP_NOTIFY_INFORM.
//
// RETURNS:    0 on success.
//
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpInitNotifys (unsigned char * ElementName, int ElementNameLen,
                      UINT32 TargetIpAddr, char * pSecName,
                      int SecLevel, int NotifyType);

/*
////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmSnmpRemoveNotifys
//
// PURPOSE:    This function removes rows (added by cmSnmpInitNotifys) in the
//             the snmpNotifyTable, snmpTargetAddrTable and snmpTargetParamsTable.
//
// PARAMETERS: unsigned char * ElementName: A name used as the index and tag
//                   name parameters in the tables.
//             int ElementNameLength: The length of the ElementName.
//
// RETURNS:    0 on success.
//
////////////////////////////////////////////////////////////////////////////////
*/
int cmSnmpRemoveNotifys (unsigned char * ElementName, int ElementNameLen);



/*****************************************************************************
** FUNCTION:   cmSnmpSetIfIFaceRow
**
** PURPOSE:    Create a new row in the ifTable to represent a voice line.
**
** PARAMETERS:  ifType  -  The IANA-assigned type for such a voice line
**              pIfDesc -  The textual description of the interface
**              ifStat  -  The ifAdminStatus value of the interface
**              pIfIndex-  On success, the ifIndex of the new ifEntry
**                         will be loaded here
**
** RETURNS:    0 on success, <> 0 - otherwise
**
*****************************************************************************/
int cmSnmpSetIfIFaceRow (int ifType, const char *pIfDesc, int ifStat, UINT32 *pIfIndex);

/*****************************************************************************
** FUNCTION:   cmSnmpSetIfIFaceAdminStat
**
** PURPOSE:    Set the ifAdminStatus of the specified ifTable entry.
**
** PARAMETERS: int ifIndex: the ifIndex value of the entry to set
**             int ifAdminStatus: the ifAdminStatus value to set.  Must be
**                  one of the following values (per RFC-2233):
**                  1=up, 2=down, 3=testing
**
** RETURNS:    0 on success, -1 on failure.
**
*****************************************************************************/
int cmSnmpSetIfIFaceAdminStat (int ifIndex, int ifAdminStat);


/*****************************************************************************
** FUNCTION:   cmSnmpGetCmSyslog
**
** PURPOSE:    Returns the IP address of the syslog server encoded as a 32 bit
**             unsigned integer
**
** PARAMETERS: none
**
** RETURNS:    The IP address of docsDevEvSyslog, encoded as a 32 bit unsigned
**             integer.  On failure, a value of 0 is returned.
**
*****************************************************************************/
UINT32 cmSnmpGetCmSyslog (void);


/*****************************************************************************
** FUNCTION:   cmSnmpGetDocsDevSwCurrentVers
**
** PURPOSE:    Returns the current value of docsDevSwCurrentVers.
**
** PARAMETERS: None
**
** RETURNS:    A static char * that contains a NULL terminated string with the
**             current value of docsDevSwCurrentVers.  The pointer returned
**             is owned by this function and must not be freed by the caller.
**
*****************************************************************************/
const char* cmSnmpGetDocsDevSwCurrentVers(void);


/********************************************************************************/
/******************                                            ******************/
/******************   Permanent SNMP Non-Vol Settings          ******************/
/******************                                            ******************/
/********************************************************************************/

/*****************************************************************************
** FUNCTION:   cmNvGetMtaDeviceSerialNumber
**
** PURPOSE:    A string containing the hardware version of the MTA device,
**             including the \0.
**
** PARAMETERS:
**             pBuf     - destination buffer
**             buf_len  - length of destination buffer
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvGetMtaDeviceSerialNumber(char * pBuf, UINT32 buf_len);


/*****************************************************************************
** FUNCTION:   cmNvGetMtaDeviceHardwareVersion
**
** PURPOSE:    A string containing the hardware version of the MTA device,
**             including the \0.
**
** PARAMETERS:
**        pBuf    - destination buffer
**        buf_len - length of destination buffer
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvGetMtaDeviceHardwareVersion(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvGetSysDescr
**
** PURPOSE:    A string containing the sysDescr of the MTA device,
**             including the \0.
**
** PARAMETERS:
**        pBuf    - destination buffer
**        buf_len - length of destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvGetSysDescr(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSysDescr
**
** PURPOSE:    A string containing the sysDescr of the MTA device,
**             including the \0.
**
** PARAMETERS:
**        pBuf    - destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvSetSysDescr(char * pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetSysObjectId
**
** PURPOSE:    Gets a string containing the sysObjectId of the MTA device,
**             buf_len - length of destination buffer
**
** PARAMETERS:
**        pBuf    - destination buffer
**        buf_len - length of destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvGetSysObjectId(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSysObjectId
**
** PURPOSE:    Sets sysObjectId of the MTA device in NVRAM
**
** PARAMETERS:
**        pBuf    - destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvSetSysObjectId(char * pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetSysContact
**
** PURPOSE:    The MTA's sysContact string, including the \0 terminator.
**
** PARAMETERS:
**        pBuf    - destination buffer
**        buf_len - length of destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL  cmNvGetSysContact(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSysContact
**
** PURPOSE:    The MTA's sysContact string, including the \0 terminator.
**
** PARAMETERS:
**        pBuf - source buffer
** RETURNS: None
*****************************************************************************/
void   cmNvSetSysContact(const char * pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetSysName
**
** PURPOSE:    The MTA's sysName string, including the \0 terminator.
**
** PARAMETERS:
**        pBuf    - destination buffer
**        buf_len - length of destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL   cmNvGetSysName(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSysName
**
** PURPOSE:    The MTA's sysName string, including the \0 terminator.
**             The length of this field is specified above.
**
** PARAMETERS:
**        pBuf - source buffer
**
** RETURNS: None
*****************************************************************************/
void   cmNvSetSysName(const char * pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetSysLocation
**
** PURPOSE:    The MTA's sysLocation string, including the \0 terminator.
**
** PARAMETERS:
**        pBuf - destination buffer
**        buf_len - length of destination buffer
**
** RETURNS:
**        0 - if OK, <> 0 - otherwise
*****************************************************************************/
BOOL   cmNvGetSysLocation(char * pBuf, UINT32 buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSysLocation
**
** PURPOSE:    The MTA's sysLocation string, including the \0 terminator.
**
** PARAMETERS:
**        pBuf - source buffer
**
** RETURNS: None
*****************************************************************************/
void   cmNvSetSysLocation(const char * pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetProvKdcSkewTime
**
** PURPOSE:    Get the time skew for AS req provisioning
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
UINT32   cmNvGetProvKdcSkewTime(void);

/*****************************************************************************
** FUNCTION:   cmNvSetProvKdcSkewTime
**
** PURPOSE:    Get the time skew for AS req provisioning
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvSetProvKdcSkewTime(UINT32 skew_value);

/*****************************************************************************
** FUNCTION:  cmNvGetProvApSkewTime
**
** PURPOSE:    Get the time skew for AP req provisioning
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
UINT32   cmNvGetProvApSkewTime(void);

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvCommitPermanentNonVolSettings
//
// Commit non-vol settings changes (set functions) to the non-vol device.
// This is only necessary if you did not specify true in the commit argument
// of the set function.  It is desireable to write to non vol only when
// necessary to preserve device lifespan.  If you have multiple items to change,
// commit only at the end.
//
// It is very undesireable to commit permanent settings in an environment that
// you cannot tolerate then being set back to default values in the event
// of a corruption of these settings due to a power glitch for example.
// A warning message will print any time you do this.
//
// Parameters:
//      synchronous - if true, commit is complete upon return.  This
//                      could take several seconds.
// Returns:
/////////////////////////////////////////////////////////////////////////////////
*/
void cmNvCommitPermanentNonVolSettings ( BOOL synchronous );

/*
/////////////////////////////////////////////////////////////////////////////////
// FUNCTION:   cmNvCommitDynamicNonVolSettings
//
// Commit non-vol settings changes (set functions) to the non-vol device.
// It is desireable to write to non vol only when necessary to preserve device
// lifespan.  If you have multiple items to change, commit only once at the end.
//
// Parameters:
//      synchronous - if true, commit is complete upon return.  This
//                      could take several seconds.
// Returns:
/////////////////////////////////////////////////////////////////////////////////
*/
void cmNvCommitDynamicNonVolSettings( BOOL synchronous );


/*****************************************************************************
** FUNCTION:  cmNvSetProvApSkewTime
**
** PURPOSE:    Set the time skew for AP req provisioning
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvSetProvApSkewTime(UINT32 skew_value);

/*****************************************************************************
** FUNCTION:  cmNvSetSnmpV3KerberosTicket
**
** PURPOSE:   Store the prov ticket in NVRAM
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvSetSnmpV3KerberosTicket(const UINT8 * pBuf, unsigned int len);

/*****************************************************************************
** FUNCTION:  cmNvGetSnmpV3KerberosTicket
**
** PURPOSE:   Retrieve the prov ticket in NVRAM
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvGetSnmpV3KerberosTicket(UINT8 * pBuf, unsigned int buf_len);


typedef struct
{
    char    *pRealmName;
    UINT32  asSkewTime;
} MTA_KdcEntry;


/*****************************************************************************
** FUNCTION:  cmNvSetKdcEntry
**
** PURPOSE:   Store the AS time skew for realm
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvSetKdcEntry(int index, const MTA_KdcEntry *pEntry);

/*****************************************************************************
** FUNCTION:  cmNvGetKdcEntry
**
** PURPOSE:   Retrieve the AS time skew for realm
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvGetKdcEntry(int index, const MTA_KdcEntry *pEntry);

typedef struct MTA_CmsEntry_
{
   char           *pCmsFqdn;
   UINT8          *pKerberosTicket;
   unsigned int   kerberosTicketLength;
   UINT8          *pKerberosSessionKey;
   unsigned int   kerberosSessionKeyLength;
   UINT32         ticketTime[4];
   UINT32         apSkewTime;
} MTA_CmsEntry;

/*****************************************************************************
** FUNCTION:  cmNvSetCmsEntry
**
** PURPOSE:   Store the NVRAM CMS info
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvSetCmsEntry(int index, const MTA_CmsEntry *pEntry);

/*****************************************************************************
** FUNCTION:  cmNvGetCmsEntry
**
** PURPOSE:   Retrieve the NVRAM CMS info
**
** PARAMETERS:
**
** RETURNS: None
*****************************************************************************/
BOOL   cmNvGetCmsEntry(int index, const MTA_CmsEntry *pEntry);


/****************************************************************************/
/***********  BLIND NVRAM ACCESSOR FUNCTIONS DEFINITIONS    *****************/
/****************************************************************************/

/*
*****************************************************************************
** FUNCTION:   cmNvResetDefaultBlindEmtaData
**
** PURPOSE:    Resets the blind EMTA non-vol data to it's default settings
**
** PARAMETERS: None
**
** RETURNS:    None
**
*****************************************************************************
*/
void cmNvResetDefaultBlindEmtaData( void );

/*****************************************************************************
** FUNCTION:   cmNvWriteAndCommitBlindEmtaData
**
** Commits MTA's NV "Blind" settings to the non-vol device.
** It is desireable to write to non vol only when necessary to preserve device
** lifespan.  If you have multiple items to change, commit only once at the end.
**
** Parameters:
**             bSynch - if TRUE, commit is complete upon return.  This
**                      could take several seconds.
** Returns:    void
*****************************************************************************/
void cmNvWriteAndCommitBlindEmtaData( BOOL bSynch );


/*****************************************************************************
** FUNCTION:   cmNvGetSnmpV3SessionKeyAndTktTimes
**
** PURPOSE:    Retrieve the prov session key and ticket times
**
** PARAMETERS: pBuf - contains a concatenation of SNMPv3 session key and
**                    4 integers which are the 4 ticket times
**
** RETURNS:    TRUE - if OK, <>FALSE - otherwise.
*****************************************************************************/
UINT32 cmNvGetSnmpV3SessionKeyAndTktTimes( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:  cmNvSetSnmpV3SessionKeyAndTktTimes
**
** PURPOSE:   Store the prov session key and ticket times
**
** PARAMETERS:  pBuf - contains a concatenation of session key and 4 integers
**                     which are the 4 ticket times
**
** RETURNS:   none
*****************************************************************************/
void   cmNvSetSnmpV3SessionKeyAndTktTimes( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvGetMtaHardwareVersion
**
** PURPOSE:    Retrieves the MTA Hardware Version data from NVRAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvGetMtaHardwareVersion( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetMtaHardwareVersion
**
** PURPOSE:    Sets the MTA Hardware Version data in NV RAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the data in the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvSetMtaHardwareVersion( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvGetMtaSerialNumber
**
** PURPOSE:    Retrieves the MTA Serial number from NVRAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvGetMtaSerialNumber( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetMtaSerialNumber
**
** PURPOSE:    Sets the MTA Serial Number in NVRAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the data in the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvSetMtaSerialNumber( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvGetDqosConfig
**
** PURPOSE:    Retrieve the DQoS non-vol settings
**
** PARAMETERS: dqos [OUT] - DQoS configuration structure
**
** RETURNS:    TRUE - if OK, <>FALSE - otherwise.
*****************************************************************************/
UINT32 cmNvGetDqosConfig( DQOS_NONVOL_CONFIG *dqos );

/*****************************************************************************
** FUNCTION:   cmNvSetDqosConfig
**
** PARAMETERS: dqos [IN] - DQoS configuration structure
**
** PARAMETERS: pBuf - contains a concatenation of session key and 4 integers
**
** RETURNS:    none
*****************************************************************************/
void cmNvSetDqosConfig( DQOS_NONVOL_CONFIG *dqos );

/*****************************************************************************
** FUNCTION:   cmNvGetMtaKerberosFlag
**
** PURPOSE:    Retrieve the kerberos flag
**
** PARAMETERS: krbflag [OUT] - value of the kerberos flag
**
** RETURNS:    TRUE - if OK, <>FALSE - otherwise.
*****************************************************************************/
UINT32 cmNvGetMtaKerberosFlag( int *krbFlag );

/*****************************************************************************
** FUNCTION:   cmNvSetMtaKerberosFlag
**
** PARAMETERS: krbFlag [IN] - kerberos flag value
**
** RETURNS:    none
*****************************************************************************/
void cmNvSetMtaKerberosFlag( int krbFlag );

/*****************************************************************************
** FUNCTION:   cmNvGetMtaKrbContProv
**
** PURPOSE:    Retrieve the kerberos flag
**
** PARAMETERS: krbContProv [OUT] - value of the kerberos flag
**
** RETURNS:    TRUE - if OK, <>FALSE - otherwise.
*****************************************************************************/
UINT32 cmNvGetMtaKrbContProv( int *krbContProv );

/*****************************************************************************
** FUNCTION:   cmNvSetMtaKrbContProv
**
** PARAMETERS: krbContProv [IN] - kerberos flag value
**
** RETURNS:    none
*****************************************************************************/
void cmNvSetMtaKrbContProv( int krbContProv );

/*****************************************************************************
** FUNCTION:   cmNvSetMediaSecDisabled
**
** PURPOSE:    Sets the mtaMediaSecDisabled flag in NVRAM
**
** PARAMETERS: flagval - value to set.  0 - if not disabled (enabled), 1 - if disabled
**
** RETURNS:   nothing
 * *****************************************************************************/
void   cmNvSetMtaMediaSecDisabled(int flagval ) ;

/*****************************************************************************
** FUNCTION:   cmNvGetMtaSecDisabled
**
** PURPOSE:    Retrieves the mtaMediaSecDisabled flag from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    0 - if not disabled (enabled), 1 - if disabled
*****************************************************************************/
int   cmNvGetMtaMediaSecDisabled( void );

/*****************************************************************************
** FUNCTION:   cmNvGetMtaFQDNCheckDisabled
**
** PURPOSE:    Retrieves the mtaFQDNCheckDisabled flag from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    0 - if not disabled (enabled), 1 - if disabled
*****************************************************************************/
int   cmNvGetMtaFQDNCheckDisabled( void ) ;
/*****************************************************************************
** FUNCTION:   cmNvSetMtaFQDNCheckDisabled
**
** PURPOSE:    Sets the mtaFQDNCheckDisabled flag in NVRAM
**
** PARAMETERS: flagval - value to set.  0 - if not disabled (enabled), 1 - if disabled
**
** RETURNS:   nothing
 * *****************************************************************************/
void   cmNvSetMtaFQDNCheckDisabled(int flagval )  ;

/*****************************************************************************
** FUNCTION:   cmNvSetMtaIpsecDisabled
**
** PURPOSE:    Sets the "mtaIpsecDisabled" flag in NVRAM.
**
** PARAMETERS: flagval - value to set:
**                      0 - if not disabled (enabled),
**                      1 - if disabled
**
** RETURNS:   void
******************************************************************************/
void   cmNvSetMtaIpsecDisabled(int flagval );

/*****************************************************************************
** FUNCTION:   cmNvGetMtaIpsecDisabled
**
** PURPOSE:    Retrieves the "mtaIpsecDisabled" flag from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    0 - if not disabled (enabled), 1 - if disabled
*****************************************************************************/
int   cmNvGetMtaIpsecDisabled( void );

/*****************************************************************************
** FUNCTION:   cmNvSetMtaInformDisabled
**
** PURPOSE:    Sets the "mtaSnmpInformDisabled" flag in NVRAM for SNMP
**             INFORM message.
**
** PARAMETERS: flagval - value to set:
**                      0 - if not disabled (enabled),
**                      1 - if disabled
**
** RETURNS:   void
 * *****************************************************************************/
void   cmNvSetMtaInformDisabled(int flagval ) ;


/*****************************************************************************
** FUNCTION:   cmNvGetMtaInformDisabled
**
** PURPOSE:    Retrieves the "mtaSnmpInformDisabled" flag from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    0 - if not disabled (enabled), 1 - if disabled
*****************************************************************************/
int   cmNvGetMtaInformDisabled( void );

/*****************************************************************************
** FUNCTION:   cmNvGetSnmpMgrUserName
**
** PURPOSE:    Retrieves the SNMP Mgr's User Name from NVRAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvGetSnmpMgrUserName( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSnmpMgrUserName
**
** PURPOSE:    Sets the SNMP Mgr's User Name from NVRAM
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the data in the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvSetSnmpMgrUserName( UINT8 *pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetSnmpMgrEngineID
**
** PURPOSE:    Retrieves the SNMP Mgr's EngineID from NVRAM
**             Engine ID is represented as an ASCII string
**             of Hexadecimal string of bytes.
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvGetSnmpMgrEngineID( UINT8 *pBuf, unsigned int buf_len);

/*****************************************************************************
** FUNCTION:   cmNvSetSnmpMgrEngineID
**
** PURPOSE:    Sets the SNMP Mgr's EngineID to NVRAM
**             Engine ID is represented as an ASCII string
**             of Hexadecimal string of bytes.
**
** PARAMETERS: pBuf     - pointer to the buffer to retrieve the data
**             buf_len  - size of the data in the buffer (including '\0' terminator)
**
** RETURNS:    0 - if OK, <>0 - otherwise
*****************************************************************************/
int   cmNvSetSnmpMgrEngineID( UINT8 *pBuf);

/*****************************************************************************
** FUNCTION:   cmNvGetMtaEngineBoot
**
** PURPOSE:    Retrieves the "mtaSnmpMgrEngineBoot" value from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    Engine Boot value
*****************************************************************************/
int   cmNvGetMtaEngineBoot( void );

/*****************************************************************************
** FUNCTION:   cmNvSetMtaEngineBoot
**
** PURPOSE:    Sets the "mtaSnmpMgrEngineBoot" in NVRAM for SNMP
**             INFORM message.
**
** PARAMETERS: engBoot  - value to set.
**
** RETURNS:   void
******************************************************************************/
void   cmNvSetMtaEngineBoot(int engBoot);

/*****************************************************************************
** FUNCTION:   cmNvGetMtaEngineTime
**
** PURPOSE:    Retrieves the "mtaSnmpMgrEngineTime" value from NVRAM
**
** PARAMETERS: none
**
** RETURNS:    Engine Boot value
*****************************************************************************/
int   cmNvGetMtaEngineTime( void );

/*****************************************************************************
** FUNCTION:   cmNvSetMtaEngineTime
**
** PURPOSE:    Sets the "mtaSnmpMgrEngineTime" in NVRAM for SNMP
**             INFORM message.
**
** PARAMETERS: engTime  - value to set.
**
** RETURNS:   void
******************************************************************************/
void   cmNvSetMtaEngineTime(int engTime);




#if (GLOBAL_CFG_IPSEC_SUPPORT)

/********************************************************************************/
/******************                                            ******************/
/******************         IPSEC   Interface   Group          ******************/
/******************                                            ******************/
/********************************************************************************/

/*
   First, some data type defs. All these must compatible with those
   in IpsecApi.h file
*/
typedef UINT32 IPSEC_ID;      /* IPSEC Entry */
typedef UINT32 IPSEC_POLICY;  /* IPSEC Spec */
typedef UINT32 IPSEC_EXPTIME; /* IPSEC LifiTime */

typedef enum IPSEC_CALLBACK_REASON /* IPSEC_CALLBACK_TYPE */
{
    IPSEC_TIMEOUT_DELETED = 1 /* IPSEC_TIMEOUT_BUNDLE_DELETED */
} IPSEC_CALLBACK_REASON;

typedef void(*PFN_IPSEC_CALLBACK) (IPSEC_CALLBACK_REASON type, void *);

/* IPSEC protocol used */
typedef enum IPSEC_IPPROTOCOL
{
    IPSEC_IPPROTO_ANY = 0,
    IPSEC_IPPROTO_ICMP = 1,
    IPSEC_IPPROTO_TCP  = 6,
    IPSEC_IPPROTO_UDP  = 17
} IPSEC_IPPROTOCOL;

/* Encryption (Privacy) Algorithm used */
typedef enum IPSEC_PRIV  /* IPSEC_ENCRYPT_ALGO */
{
    IPSEC_PRIV_NONE = 0,
    IPSEC_PRIV_DES  = 2,
    IPSEC_PRIV_3DES = 3,
    IPSEC_PRIV_NULL = 11
} IPSEC_PRIV;

/* Authienication mode */
typedef enum IPSEC_AUTH  /* IPSEC_AUTH_ALGO */
{
   IPSEC_AUTH_NONE  = 0,
   IPSEC_AUTH_HMAC_MD5 = 1,
   IPSEC_AUTH_HMAC_SHA = 2
} IPSEC_AUTH;

/* Encapsulation Type */
typedef enum IPSEC_ENCAP_TYPE /* IPSEC_ENCAP_MODE */
{
    IPSEC_ENCAPS_NONE = 0,
    IPSEC_ENCAPS_TUNNEL = 1,
    IPSEC_ENCAPS_TRANSPORT = 2
} IPSEC_ENCAP_TYPE;

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to setup the callback function. This function gets passed various
// events which occur during Ipsec processing.
//
// Parameters:
//      callback_function  - the callback function
//
// Returns:
//      none
//
/////////////////////////////////////////////////////////////////////////////////
*/
void cmIpsecSetupCallback(PFN_IPSEC_CALLBACK callback_function);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Outbound IPSEC bundle. If a packet fits in the
// requirements defined by this bundle, then the Specification defined by ipsecCreateSpec
// will occur.
//
// Parameters:
//      source_start_addr - the source start address of the packet
//      source_end_addr   - the source end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//      dest_start_addr   - the destination start address of the packet
//      dest_end_addr     - the destination end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//                          (or the same start and end specifies one address
//      protocol          - the protocol which must match
//      source_port       - the source port (TCP or UDP only)
//      dest_port         - the destination port (TCP or UDP only)
//
// Returns:
//      IPSEC_ID          - an IPSEC entry which was just created
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_ID cmIpsecCreateOutboundEntry(UINT32 source_start_addr, UINT32 source_end_addr,
                                   UINT32 dest_start_addr,   UINT32 dest_end_addr,
                                   IPSEC_IPPROTOCOL protocol, UINT16 source_port,
                                   UINT16 dest_port);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to modify an existing Outbound IPSEC bundle.
//
// Parameters:
//      existing_entry    - the existing ipsec_entry
//      source_start_addr - the source start address of the packet
//      source_end_addr   - the source end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//      dest_start_addr   - the destination start address of the packet
//      dest_end_addr     - the destination end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//                          (or the same start and end specifies one address
//      protocol          - the protocol which must match
//      source_port       - the source port (TCP or UDP only)
//      dest_port         - the destination port (TCP or UDP only)
//
// Returns:
//      IPSEC_ENTRY  - an IPSEC entry which was created
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_ID cmIpsecModifyOutboundEntry(IPSEC_ID existing_entry,  UINT32 source_start_addr, UINT32 source_end_addr,
                                   UINT32 dest_start_addr,   UINT32 dest_end_addr,
                                   IPSEC_IPPROTOCOL protocol, UINT16 source_port,
                                   UINT16 dest_port);


/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Inbound IPSEC bundle. If a packet fits in the
// requirments defined by this bundle, then the Specification defined by ipsecCreateSpec
// will occur.
//
//
// Parameters:
//      source_start_addr - the source start address of the packet
//      source_end_addr   - the source end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//      dest_start_addr   - the destination start address of the packet
//      dest_end_addr     - the destination end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//                          (or the same start and end specifies one address
//      protocol          - the protocol which must match
//      spi                - the Security parameter index
//      source_port       - the source port (TCP or UDP only)
//      dest_port         - the destination port (TCP or UDP only)
//      esp_enabled            - whether encryption is enabled
//
// Returns:
//      IPSEC_ENTRY  - an IPSEC entry used for ipsecCreateInboundSpec, and ipsecDeleteInboundSpec
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_ID cmIpsecCreateInboundEntry(UINT32 source_start_addr, UINT32 source_end_addr,
                                 UINT32 dest_start_addr,   UINT32 dest_end_addr,
                                 IPSEC_IPPROTOCOL protocol, UINT16 source_port,
                                 UINT16 dest_port, BOOL esp_enabled, UINT32 spi);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to modify an existing Inbound IPSEC bundle.
// Parameters:
//      existing_entry    - the existing inbound entry
//      source_start_addr - the source start address of the packet
//      source_end_addr   - the source end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//      dest_start_addr   - the destination start address of the packet
//      dest_end_addr     - the destination end address of the packet
//                          (0 - start_addr, 0xFFFFFFFF - end_addr, specifies all)
//                          (or the same start and end specifies one address
//      protocol          - the protocol which must match
//      source_port       - the source port (TCP or UDP only)
//      dest_port         - the destination port (TCP or UDP only)
//      esp_enabled            - whether encryption is enabled
//      spi                - the Security parameter index
//
// Returns:
//      IPSEC_ENTRY  - an IPSEC entry used for ipsecCreateInboundSpec, and ipsecDeleteInboundSpec
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_ID cmIpsecModifyInboundEntry(IPSEC_ID ipsec_entry, UINT32 source_start_addr, UINT32 source_end_addr,
                              UINT32 dest_start_addr,    UINT32 dest_end_addr,
                              IPSEC_IPPROTOCOL protocol, UINT16 source_port,
                              UINT16 dest_port, BOOL esp_enabled, UINT32 spi);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Outbound specification which occurs when a packet
// meets the requirements defined by the bundle.
//
// Parameters:
//      ipsec_entry        - returned by the ipsecCreateOutboundEntry call
//      encrypt_algo       - the encryption algorithm
//      auth_algo          - the authenication algorithm
//      ekey               - the encoded des or 3des key (0 if none)
//                           must be the correct size depending on encrypt algo.
//      akey               - the encoded authenication key
//      decode_dkey        - the decoded encrypt key
//      decode_akey        - the decoded authenication key
//      spi                - the Security parameter index
//
// Returns:
//      true inbound bundle.
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_POLICY cmIpsecCreateOutboundSpec(IPSEC_ID ipsec_entry, IPSEC_PRIV encrypt_algo,
                            IPSEC_AUTH auth_algo, char *dkey,
                            char *akey,  UINT32 spi);


/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Inbound specification which occurs when a packet
// meets the requirements defined by the bundle.
//
// Parameters:
//      ipsec_entry        - returned by the ipsecCreateInboundEntry call
//      encrypt_algo       - the encryption algorithm
//      auth_algo          - the authenication algorithm
//      decode_dkey        - the decoded encrypt key
//      decode_akey        - the decoded authenication key
//      spi                - the spi
//
// Returns:
//      true inbound bundle.
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_POLICY  cmIpsecCreateInboundSpec(IPSEC_ID ipsec_entry, IPSEC_PRIV encrypt_algo,
                            IPSEC_AUTH auth_algo, char * decode_dkey, char *decode_akey,
                            unsigned long spi);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Outbound Expiration Time Policy.
//
// Parameters:
//      outbound_sa_bundle - the outbound IPSEC_ID
//      time_expiration    - expiration time in seconds (0 if not used)
//
// Returns:
//      New IPSEC Spec outbound bundle.
//      NULL if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_EXPTIME cmIpsecCreateOutboundExptimeSpec(IPSEC_ID outbound_sa_bundle, UINT32 time_expiration);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to create an Inbound Lifetime specification.
//
// Parameters:
//      inbound_sa_bundle - the outbound bundle
//      time_expiration    - expiration in seconds (0 if not used)
//
// Returns:
//      New IPSEC Spec inbound bundle.
//      NULL if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
IPSEC_EXPTIME cmIpsecCreateInboundExptimeSpec(IPSEC_ID inbound_sa_bundle, UINT32 time_expiration);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to delete an IPSEC Outbound bundle.
//
// Parameters:
//      oubound_entry       - the entry to be deleted
//
// Returns:
//      true if Entry was deleted.
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmIpsecDeleteOutboundEntry(IPSEC_ID oubound_entry);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to delete an Inbound IPSEC bundle.
//
// Parameters:
//      inbound_entry       - the entry to be deleted
//
// Returns:
//      true if Entry was deleted.
//      false if there was a problem.
//
/////////////////////////////////////////////////////////////////////////////////
*/
BOOL cmIpsecDeleteInboundEntry(IPSEC_ID inbound_entry);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to delete an Outbound IPSEC Spec.
//
// Parameters:
//      outbound_entry     - the entry which contains the spec
//      outbound_spec      - the spec to be deleted
//
// Returns:
//      none
//
/////////////////////////////////////////////////////////////////////////////////
*/
void cmIpsecDeleteOutboundSpec(IPSEC_ID outbound_entry, IPSEC_POLICY oubound_spec);

/*
/////////////////////////////////////////////////////////////////////////////////
// This function is used to delete an Outbound IPSEC bundle.
// Parameters:
//      inbound_entry     - the inbound entry which contains the spec
//      inbound_spec      - the spec to be deleted
//
// Returns:
//      none
//
/////////////////////////////////////////////////////////////////////////////////
*/
void cmIpsecDeleteInboundSpec(IPSEC_ID inbound_entry, IPSEC_POLICY inbound_spec);

#endif /* #if (GLOBAL_CFG_IPSEC_SUPPORT) */

/*
*******************************************************
*******************************************************
****           O B S O L E T E ! ! !               ****
*******************************************************
*******************************************************
*/
BOOL cmMiscIsCableModem(void);

/*****************************************************************************
** ASN.1 and BER Library Interface Group
**
**    TBD
**
******************************************************************************/


/*****************************************************************************
** DQoS Services Interface Group
**
**    TBD
**
******************************************************************************/
#ifdef __cplusplus
    }
#endif

#endif
