/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: asreq.c
*    Creation Date: 27 Oct 2000 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:30 $
*
****************************************************************************
*    Description:
*
*      Header file for Implementation of packetCable PKINIT Request and reply
*
****************************************************************************/



#ifndef __PKINIT_H
#define __PKINIT_H

#include "seckrb.h"

#define OAKLEY1_BASE        2
#define MTA_OAKLEY1_PRIME_LEN   96    /* 768 bits = 96 bytes */
#define OAKLEY1_Q_LEN       96    /* 768 bits = 96 bytes */
#define OAKLEY2_BASE        2
#define MTA_OAKLEY2_PRIME_LEN   128
#define OAKLEY2_Q_LEN       128

/* Diffie hellman oakley group 1 prime and generator */
extern const unsigned char MTA_OAKLEY1_PRIME[];

/* Diffie hellman oakley group 1 prime and generator */
extern const unsigned char OAKLEY1_Q[];

/* Diffie hellman oakley group 1 prime and generator */
extern const unsigned char MTA_OAKLEY2_PRIME[];

/* Diffie hellman oakley group 1 prime and generator */
extern const unsigned char OAKLEY2_Q[];

/*
*****************************************************************************
** FUNCTION:     prep_pkinit_req
**
** PURPOSE:      put a pkinit request structure together
**
** PARAMETERS:   request - in - kdc-req-body to be used for checksum
**               rpk - the pkinit request structure
**               pkinit_nonce - out: nonce generated
**               mtaprivkey - our: provate ket generated for mta
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code prep_pkinit_req(
      krb5_kdc_req *request,
      pkinit_request *rpk,
      int *pkinit_nonce,
      priv_key_t *mtaprivkey,
	  int  app_flag);

/*
*****************************************************************************
** FUNCTION:     make_krb5_pkinit_req
**
** PURPOSE:      encodes a pkinit request (ASN.1)
**
** PARAMETERS:   request - in - AS request to be encoded for checksum
**               out_padata - out: encoded pkinit request
**               pkinit_nonce - out: nonce generated
**               mtaprivkey - out: private key generated
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code make_krb5_pkinit_req(krb5_kdc_req *request,
      int *pkinit_nonce, priv_key_t *mtaprivkey);


/*
*****************************************************************************
** FUNCTION:     get_krb5_pkinit_rep
**
** PURPOSE:      decode a pkinit reply (ASN.1)
**
** PARAMETERS:   in_padata - encoded pkinit reply
**               out_pk - out: pkinit reply structure
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code get_krb5_pkinit_rep(krb5_pa_data *in_padata,
      pkinit_reply **out_pk);


/*
*****************************************************************************
** FUNCTION:     get_krb5_auth_pack_reply
**
** PURPOSE:      decode econtent of as_reply (ASN.1)
**
** PARAMETERS:   io_authpack - out: auth_pack structure decoded
**
** RETURNS:      success or failure
**
** NOTE:
**
**
*****************************************************************************
*/
krb5_error_code get_krb5_auth_pack(encapsulated_content_info *io_authpack);

#endif
