/***************************************************************************
*    Copyright 2000  Broadcom Corporation
*    All Rights Reserved
*    No portions of this material may be reproduced in any form without the
*    written permission of:
*             Broadcom Corporation
*             16251 Laguna Canyon Road
*             Irvine, California  92618
*    All information contained in this document is Broadcom Corporation
*    company private, proprietary, and trade secret.
*
****************************************************************************
*
*    Filename: vxwrkdef.h
*    Creation Date: 28 June 2001 (v0.00)
*    VSS Info:
*        $Revision: 1.1 $
*        $Date: 2002/08/21 17:54:35 $
*
****************************************************************************
*    Description:
*
*      Private VxWorks related prototypes and type defines.
*
****************************************************************************/

#ifndef _VXWRKDEF_H
#define _VXWRKDEF_H

#ifdef __cplusplus
extern "C" {
#endif

#include "bcmosHelper.h"
#include "osdef.h"

/*
*****************************************************************************
** CONSTANTS
*****************************************************************************
*/

/* Definitions for OS specific tick values */
#define VXWORKS_MSEC_PER_TICK   10

/* Scaling factor used to convert generic tick value(OS_MSEC_PER_TICK) into
 * OS specific tick value */
#define OS_TICK_SCALE_FACTOR  ( VXWORKS_MSEC_PER_TICK / OS_MSEC_PER_TICK )


/* Definitions for the maximum value that may be returned by bcmOsGetTime().
 * This is used to handle wrapping conditions when calculating delta system
 * times. */
#define OS_MAX_TIME   0xffffffff


/*
*****************************************************************************
** TYPE DEFINITIONS
*****************************************************************************
*/

/* OS-specific status types. This is the data type returned by (most) VxWorks
 * API calls that relate to tasks, mutexes, semaphores, events, message queues... */
typedef STATUS    NATIVE_STATUS;


/*
*****************************************************************************
** FUNCTION PROTOTYPES
*****************************************************************************
*/

/*****************************************************************************
** FUNCTION:   checkStatus
**
** PURPOSE:    This function takes the status returned by an OS-specific API
**             function call. It maps this to a generic BcmOS return status, and
**             logs any errors.
**
** PARAMETERS: status   - operating system specific status returned from system call
**             fnc      - name of calling function, used for error LOG statements
**
** RETURNS:    Operating system independent status value of type OS_STATUS
**             (that corresponds to 'status') . BcmOS public interface functions
**             should return OS independent status values. Native OS status
**             values should be encapsulated; however, they may be LOGed for
**             debug purposes.
**
** NOTES:      This function is used to validate all OS-specific API calls invoked
**             by the BcmOS layer that deal with tasks, mutexes, semaphores, events,
**             message queues... However, not all OS API calls return consistent
**             status values/error codes. Therefore, the following assumptions have
**             been made:
**
**             PSOS:    an error is indicated by a non-zero 'status' value
**             VXWORKS: an error is indicated by a 'status' value of ERROR
**             WIN32:   an error is indicated by a 'status' value of 0
**
**             These assumptions should handle most OS-specific API calls. However, it
**             is the responsibility of the caller to massage the input parameters to conform
**             to these assumptions for any OS API calls that do not follow these rules.
**
*****************************************************************************/
OS_STATUS checkStatus( NATIVE_STATUS status, char *fnc );



#ifdef __cplusplus
    }
#endif

#endif  /* _VXWRKDEF_H  */


